/*
 *    gnirc - Another IRC client for X 
 *    Copyright (C) 1998  david Ramboz <dramboz@imaginet.fr>
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 */

#include <stdio.h>
#include <string.h>
#include <gtk/gtk.h>

#include "gnirc.h"
#include "enums.h"

extern	 UVariable gb_uvariables[];
extern   GList     *gb_gui_windows_list;
extern   char      *gb_tmp_buff;

Channel*
channel_new( Connection *connection, Channel *parent, char *name) {
        Channel *channel;
                	
        channel = g_malloc( sizeof(Channel));
        channel->connection = connection;
        channel->name = g_strdup(name);
        channel->flags= CHANNEL_VISIBLE;
        channel->data = history_new( VAR_INT( VAR_HISTORY_SIZE), (FreeFunc) irc_message_delete);
        channel->users = NULL;
        channel->topic = NULL;
        channel->logfile = NULL;
        
        lower_string( channel->name);
        
        if ( parent) {
                channel->window = parent->window;
       		gui_window_new_channel( parent->window, channel);
        } else {
               	channel->window = gui_window_new();
       		gui_window_new_channel( channel->window, channel);
		gtk_widget_show( channel->window->window);
        }
        
        return channel;
}

void
channel_free( Channel *channel) {
        GList *user;
        
        g_return_if_fail( channel != NULL);

	if ( channel->window)
            	gui_window_delete_channel( channel->window, channel);
		
        if ( channel->name)
             	g_free( channel->name);
                
        user = channel->users;
        while (user) {
                g_free( user->data);
                user = user->next;
        }
        if (channel->users)
            	g_list_free( channel->users);
                
        if ( channel->data)
        	history_free( channel->data);
                
        if ( channel->logfile)
              	fclose( channel->logfile);
                	
        g_free( channel);
}

Channel*
channel_find( GList *channel_list, char *name) {
        Channel *channel;
        GList	*channel_item;
        
        g_return_val_if_fail( channel_list != NULL, NULL);
        g_return_val_if_fail( name != NULL, NULL);
        
        lower_string( name);
        channel_item = channel_list;
        while (channel_item) {
                channel = (Channel*) channel_item->data;
                if ( !strcmp( channel->name, name))
                	return channel;
                channel_item = channel_item->next;
        }
        return NULL;
}


gint
channel_find_user( Channel *channel, char *user) {
        GList *item;
        char   *s;
        char   *nick;
        gint	nth = 0;
              
        g_return_val_if_fail( channel != NULL, FALSE);
        g_return_val_if_fail( user != NULL, FALSE);
        
        nick = g_strdup( user);
        lower_string( nick);
        
        item = channel->users;
        while (item) {
              s = (char*) item->data;
              if ( !compare_nicks( s, nick)) {
                        g_free(nick);
                      	return nth;
              }
              item = item->next;
              nth++;
        }
        g_free(nick);
        return -1;
}

void
channel_add_users( Channel *channel, char *users) {
        char buff[512];
        char *tmp;
        char *nick;
        
        g_return_if_fail( channel != NULL);
        g_return_if_fail( users != NULL);
        
        tmp = users;
       	while ( *tmp && sscanf( tmp, "%s", buff)) {
       	       nick = g_strdup(buff);
       	       lower_string( nick);
       	       channel->users = g_list_insert_sorted( channel->users, nick, (GCompareFunc) compare_users);
       	       if ( channel == channel->window->current_channel)
       	       		gui_window_add_user( channel->window, nick, g_list_find_pos( channel->users, nick));
       	       tmp += strlen(buff);
       	       while ( *tmp == ' ') tmp++;
      	}
}

gint
channel_del_user( Channel *channel, char *user) {
        gint nth;
        
        g_return_val_if_fail( channel != NULL, FALSE);
        g_return_val_if_fail( user != NULL, FALSE);
        
        nth = channel_find_user( channel, user);
        if (nth >= 0) {
             	channel->users = g_list_remove_link( channel->users, g_list_nth( channel->users, nth ));
                if ( channel == channel->window->current_channel)
                        gui_window_delete_user( channel->window, nth);
                        
                return TRUE;
        }
        return FALSE;
}

void
channel_change_user_mode( Channel *channel, char *nick, int mode, int set) {
        int nick_pos;
        
        g_return_if_fail( channel != NULL);
        g_return_if_fail( nick != NULL);
        
        if ( (nick_pos = channel_find_user( channel, nick)) == -1)
        	return;
        
        nick = g_list_nth_data( channel->users, nick_pos);	
        if ( set) {
                char *new_nick = g_malloc( strlen( nick) + 2);
                switch ( mode) {
                        case MODE_O:
                        	sprintf( new_nick, "@%s", nick);
                        	break;
                        case MODE_V:
                        	if ( *nick != '@')
                        		sprintf( new_nick, "+%s", nick);
                        	break;
                        default:
                        	g_free( new_nick);
                        	return;
               }
       	       channel_del_user( channel, nick);
       	       channel_add_users( channel, new_nick);
               g_free( new_nick);
       } else {
                 switch ( mode) {
                        case MODE_V:
                        	if ( *nick == '@')
                        		break;
                        case MODE_O:
       	       			channel_del_user( channel, nick);
       	       			channel_add_users( channel, nick);
       	       			break;
       	       		default:
       	       			break;
       	         }
       }               
       
}

void
channel_set_topic( Channel *channel, char *topic) {
        g_return_if_fail( channel != NULL);
        g_return_if_fail( topic != NULL);
        
        if ( channel->topic)
        	g_free( channel->topic);
        
        channel->topic = g_strdup( topic);
}

void
channel_append_message( Channel *channel, void *param) {
        char *message = NULL;
        int  property;
        
        g_return_if_fail( param != NULL);
	g_return_if_fail( channel != NULL);
                		
        if ( FLAG_IS_SET(channel->flags, CHANNEL_VISIBLE) || channel->logfile) {
                format_message( param, channel, &message, &property);
                if ( message) {
                        if ( FLAG_IS_SET(channel->flags, CHANNEL_VISIBLE))
        			gui_window_display_message( channel->window, message, property);
        		if ( channel->logfile) {
        			if ( fwrite( message, strlen( message), 1, channel->logfile) != 1) {
        			     g_warning( "Closing the log file of %s : %s", channel->name,
        			     		strerror( ferror( channel->logfile)));
        			     fclose( channel->logfile);
        			     channel->logfile = NULL;
        			}
        		}
        		g_free( message);
        	}
        }
        
        if ( channel->data)	
        	history_append( channel->data, irc_message_copy(  param));
}

void
channel_set_logfile( Channel *channel, FILE *logfile) {
         g_return_if_fail( channel != NULL);
         g_return_if_fail( logfile != NULL);
         
         if ( channel->logfile)
         	fclose( channel->logfile);
         
         channel->logfile = logfile;
} 
       
void
channel_redisplay_all( Channel *channel) {
        IRCMessage *last;
        IRCMessage *param;
        int i;
        
        g_return_if_fail( channel != NULL);
        
        if ( channel->data == NULL)
        	return;
        	
	gtk_text_freeze( GTK_TEXT( channel->window->text));   
	
       	i = gtk_text_get_length( GTK_TEXT( channel->window->text));
       	gtk_text_set_point( GTK_TEXT( channel->window->text), i);
       	gtk_text_backward_delete( GTK_TEXT( channel->window->text), i);
       	
        last = (IRCMessage*) history_last( channel->data);
        if ( last) { /* if there's something in the history then redisplay it */
        	history_rewind( channel->data);
        
        	channel->window->dont_freeze = TRUE;
        	while ( (param = (IRCMessage*) history_next( channel->data)) != last)
                	gui_window_display_ircmessage( channel->window, param);
                	
       		gui_window_display_ircmessage( channel->window, last);
              	channel->window->dont_freeze = FALSE;
        }
        
       	gtk_text_thaw( GTK_TEXT( channel->window->text));
       	
       	gtk_text_scroll_down_to_end( GTK_TEXT( channel->window->text));
}

void
channel_popup_window( Channel *channel) {
        g_return_if_fail( channel != NULL);
        
        gui_window_delete_channel( channel->window, channel);
        
        channel->window = gui_window_new();
       	gtk_widget_realize( channel->window->text);
      	gui_window_new_channel( channel->window, channel);
	gtk_widget_show( channel->window->window);
	
	channel_redisplay_all( channel);
}

void 
channel_popdown_window( Channel *parent, Channel *channel) {
        g_return_if_fail( parent != NULL);
        g_return_if_fail( channel != NULL);
        
        gui_window_delete_channel( channel->window, channel);
        channel->window = parent->window;
        gui_window_new_channel( parent->window, channel);
}

  /************/
 /*** Misc ***/
/************/

void
user_nickname_changed( GList *channel_list, char *old_nick, char *new_nick) {
        Channel *channel;
        GList   *channel_item;
        
        g_return_if_fail( channel_list != NULL);
        g_return_if_fail( old_nick != NULL);
        g_return_if_fail( new_nick != NULL);
                
        channel_item = channel_list;
        while( channel_item) {
                channel = (Channel*) channel_item->data;
		channel_item = channel_item->next;
                if ( channel_del_user( channel, old_nick))
             	         channel_add_users( channel, new_nick);
		if ( !IS_IRC_CHANNEL( channel->name))
                         gui_window_update_status( channel->window);
        }
}

void
user_has_quit_irc( GList *channel_list, char *nick) {
        Channel *channel;
        GList	*channel_item;
        
        g_return_if_fail( channel_list != NULL);
        g_return_if_fail( nick != NULL);
        
        channel_item = channel_list;
        while ( channel_item) {
                channel = (Channel*) channel_item->data;
                channel_item = channel_item->next;
		if ( IS_IRC_CHANNEL( channel->name))
                         channel_del_user( channel, nick);
                else if (!strcmp( nick, channel->name)) {
                         sprintf( gb_tmp_buff, "Closing chat connection with %s (gone)\n", channel->name);
			 gui_window_display_message( channel->window, gb_tmp_buff, PROPERTY_NUMERIC_REPLY);
		         cmd_leave( channel->connection, channel);
                }
        }
}

/* BUG : the message is displayed in each channel where "nick" is found.
	 And as there could be more than on channel in a window the message
	 could be displayed more than 1 time.
	 

void		
wallop_nick_message( GList *channel_list, void *msg, char *nick) {
        Channel *channel;
        GList	*channel_item;
        
        g_return_if_fail( channel_list != NULL);
        g_return_if_fail( msg != NULL);
        g_return_if_fail( nick != NULL);
        
        channel_item = channel_list;
        while ( channel_item) {
                channel = (Channel*) channel_item->data;
		if ( channel_find_user( channel, nick) >= 0)
			channel_append_message( channel, msg);
                channel_item = channel_item->next;
        }
}
		
*/
