/*
 *    gnirc - Another IRC client for X 
 *    Copyright (C) 1998  david Ramboz <dramboz@imaginet.fr>
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 */

#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <gtk/gtk.h>

#include "gnirc.h"
#include "enums.h"

extern	GHashTable	*gb_ualias_hash;
extern	GHashTable	*gb_uvariable_hash;
extern	char		*gb_tmp_buff;
extern	UCmd		gb_ucmd_messages[];
extern	UVariable	gb_uvariables[];

GList   	*gb_ignore_list;

/* 
   build in aliases :				 
  	$0 = first parameter			 
 	$i = i th parameter			 
 	$+ = all remaining parameters		
*/

UAlias 
gb_ualiases[] = 
{
        { "MSG"		, "PRIVMSG $0 $+"		},
        { "ME"		, "CTCP $c ACTION $+"		},
        { "DCC"		, "CTCP $1 DCC $+"		},
        { NULL		, NULL				}
};


gint		get_type_position		( int type, int *types);
void		show_alias			( char *key, UAlias *alias, Channel *channel);
void		show_variable			( char *key, UVariable *variable, Channel *channel);

gint
get_type_position( int type, int *types) {
        int i = 0;
        while ( *types) {
                if ( *types++ == type)
                	return i;
                i++;
        }
        return -1;
}


char*
expand_message( IRCMessage *param) {
        GList *item;
        char  *src;
        char  *dst;
        char  *arg;
        int expand, merge, size;
        
        
        src	 = param->message_type == SERVER_NUMERIC_REPLY ? 
        		param->def.urply->format :
        		gb_ucmd_messages[ param->token].format;
        if ( src == NULL || *src == '\0')
        	return NULL;
        
        expand = 0; merge = 0;
        dst = gb_tmp_buff; size = TMP_BUFF_SIZE - 2;
        do {
                if ( *src == '$') {
                	expand = 1;
                } else if ( expand) {
                        if ( isdigit( *src)) 
                                arg = g_list_nth_data( param->arg_list, *src - '0');
                        else switch ( *src) {
                                case '+':
                        		arg = g_list_last_data( param->arg_list);
                        		break;
                                case '*':
                                	merge = 1;
                                	arg = NULL;
                                	break;
                                case 'n':
                                	arg = param->nick;
                                	break;
                                case 'u':
                                	arg = param->user;
                                	break;
                                case 'h':
                                	arg = param->address;
                                	break;
                                default:
                                	arg = NULL;
                                	g_warning( "%c : unrecognized expansion character !", *src);
                                	break;
                        }
                        if ( arg) 
	                        while ( *arg && size) { *dst++ = *arg++; size--;}
			expand = 0;
                } else { 
                	*dst++ = *src; 
                	size--;
                }
                
                if ( merge) {
                        item = param->arg_list;
                        while ( item) {
                                arg = (char*) item->data;
                                item = item->next;
                                if ( arg) {
                                        size--;
                                	while ( *arg && size) { *dst++ = *arg++; size--;}
                                        *dst++ = ' ';
                                }
                        }
                        dst--;
                        merge = 0;
               }
               
    	} while( *++src && size);
    	
    	*dst++ = '\n';
   	*dst++ = '\0';
	return g_strdup(gb_tmp_buff);                	
}

/* returns the malloced formated message in *message and its "level" in *property */
void
format_message( IRCMessage *param, Channel *channel, char **message, int *property) {
        int levels;
        char *target;
        char *trailing;
        
        g_return_if_fail( param != NULL);
	g_return_if_fail( message && property);
	
	*message = NULL;
        target = g_list_nth_data( param->arg_list, 0);
        trailing = g_list_last_data(  param->arg_list);
        levels = channel && channel->window ? channel->window->levels : LEVEL_ALL;
        
        if ( param->message_type == SERVER_NUMERIC_REPLY) {
       		*property =  param->token < ERR_BEGIN ? 
       			PROPERTY_NUMERIC_REPLY : PROPERTY_ERROR;
                if ( !FLAG_IS_SET( levels, LEVEL_REPLIES) && *property == PROPERTY_NUMERIC_REPLY /* always show errors */)
                	return;
       		*message = expand_message( param);
        } else {
             char *msg = NULL;
                          
             switch ( param->token) {
                case IRC_PRIVMSG:
                	if ( IS_IRC_CHANNEL( target)) {
                		if ( !FLAG_IS_SET( levels, LEVEL_PUBLIC))
                			return;
                	} else if ( !FLAG_IS_SET( levels, LEVEL_MSGS))
                		return;
                		
             		msg = g_malloc( 512*sizeof(char));
             		if ( FLAG_IS_SET( param->connection->flags, CON_IS_DCC)) {
                		if ( !param->nick) 
                			sprintf( msg, "=> =%s= %s\n", target, trailing);
             			else
             				sprintf( msg, "=%s= %s\n", param->nick, trailing);
                	} else if ( channel == NULL || channel == channel->window->current_channel) {
                		if ( !param->nick) /* privmsg from me */
                			sprintf( msg, "> %s\n", trailing);
                		else if ( IS_IRC_CHANNEL( target)) /* privmsg to a channel */
                			sprintf( msg, "<%s> %s\n", param->nick, trailing);
                		else
                			sprintf( msg, "*%s* %s\n", param->nick, trailing);
                	} else { 
                		if ( !param->nick)
                			sprintf( msg, "%s> %s\n", channel->name, trailing);
                		else if ( IS_IRC_CHANNEL( target)) 
                			sprintf( msg, "<%s %s> %s\n", param->nick, channel->name, trailing);
                		else
                			sprintf( msg, "*%s* %s\n", param->nick, trailing);
                	}
                	*property = PROPERTY_PRIVMSG;
                	break;
                case IRC_NOTICE:
                	if ( !FLAG_IS_SET( levels, LEVEL_NOTICES))
                		return;
             		msg = g_malloc( 512*sizeof(char));
                	if ( !param->nick)
                		sprintf( msg, "-> -%s-  %s\n", target, trailing);
                	else 
                		sprintf( msg, "[%s] %s\n", param->nick, trailing);
                	*property =  PROPERTY_NOTICE;
                	break;
                case CTCP_ACTION_Q:
                	if ( !FLAG_IS_SET( levels, LEVEL_ACTIONS))
                		return;
             		msg = g_malloc( 512*sizeof(char));
                	if ( !param->nick)
                		sprintf( msg, "*** %s %s\n", 
                			       param->connection ? param->connection->data.irc->nick : NULL, 
                			       trailing);
                	else
                		sprintf( msg, "*** %s %s\n", param->nick, trailing);
                	*property = PROPERTY_ACTION;
                	break;
                case IRC_WALLOPS:
                	if ( !FLAG_IS_SET( levels, LEVEL_WALLOPS))
                		return;
        		msg = expand_message( param);
        		break;
        	case IRC_ERROR:
        		/* always shows errors */
        		msg = expand_message( param);
        		break;
                default:
        		msg = expand_message( param);
        		if ( msg) {
                		if ( param->token < IRC_END && param->token > CTCP_BEGIN) {
                			*property = param->def.ucmd->token_num == IRC_PRIVMSG ?
                				PROPERTY_CTCP_QUERY : PROPERTY_CTCP_REPLY;
               			} else {
                 			if ( !FLAG_IS_SET( levels, LEVEL_JOINS))
                				return;
               	       			*property =  PROPERTY_NUMERIC_REPLY;
                	       	}
                	} else {
                		/* for debugging purpose */
                		g_warning( "message not displayed");
                	}
                	break;
             }
             *message = msg;
        }
}
        
  /***************/
 /*** Aliases ***/
/***************/

/* 
    Bug : don't make an alias like the following :	
	alias <alias 1> <command 1>			
	....						
  	alias <alias i> <command i>			
	....						
	alias <alias n-1> <command n-1>			
	alias <command n> <alias 1>			
   ie don't make an alias alias to itself		
*/

int
new_alias( Channel *channel, IRCMessage *param) {
        UAlias	*ualias;
        char	*name;
        char	*format;
        
        name 	= g_list_nth_data( param->arg_list, 0);
        format 	= g_list_nth_data( param->arg_list, 1);
        if ( name == NULL || format == NULL)
        	return 1;
        
        upper_string( name);
        if ( (ualias = (UAlias*) g_hash_table_lookup( gb_ualias_hash, name))) {
                g_warning( "There's already an alias named %s ( %s ~ %s)", name, name, format);
                return 1;
        }
        
        ualias 	= g_malloc( sizeof( ualias));
        ualias->name 	= g_strdup( name);
        ualias->format 	= g_strdup( format);
        
	g_hash_table_insert( gb_ualias_hash, ualias->name, ualias);
	
	if ( channel) {
	       snprintf( gb_tmp_buff, TMP_BUFF_SIZE, "%s is now an alias to %s\n", name, format);
	       gui_window_display_message( channel->window, gb_tmp_buff, PROPERTY_OTHER);
	}
	
	return 0;
}

void
show_alias( char *key, UAlias *alias, Channel *channel) {
         snprintf( gb_tmp_buff, TMP_BUFF_SIZE, "  %-20s~ %s\n", alias->name, alias->format);
         gui_window_display_message( channel->window, gb_tmp_buff, PROPERTY_OTHER);
}

void
show_aliases( Channel *channel) {
        gui_window_display_message( channel->window, "Aliases :\n", PROPERTY_OTHER);
        g_hash_table_foreach( gb_ualias_hash, (GHFunc) show_alias, channel);
}

char*
expand_alias( Channel *channel, char *command, char *message) {
        UAlias *alias;
        GList *arg_list = NULL;
        GList *arg_item;
        char  *src;
        char  *dst;
        char  *arg;
        int    expand, merge, size, newcmd, i;
        
        
        alias = (UAlias*) g_hash_table_lookup( gb_ualias_hash, command);
        if ( !alias)
        	return NULL;
        
        while ( *message) {
                get_next_item( &message, &arg, ' ');
                arg_list = g_list_append( arg_list, arg);
        }
                
                
        src = alias->format;
        dst = gb_tmp_buff; size = TMP_BUFF_SIZE - 1;
        if ( src == NULL || *src == '\0')
        	return NULL;
        	
        expand = merge  = 0;
        newcmd = 1;
        
        do {
                if ( *src == ';') 
                        newcmd = 1;
                else if ( newcmd && *src != VAR_CHAR( VAR_ESCAPE_CHAR)) {
                        *dst++ = VAR_CHAR( VAR_ESCAPE_CHAR);
                        size--;
                        newcmd = 0;
                }
                
		if ( *src == '$') {
                	expand = 1;
                } else if ( expand) {
			if ( isdigit( *src)) {
				i = *src - 48;
				arg_item = g_list_nth( arg_list, i);
				if ( arg_item) {
					arg = (char*) arg_item->data;
					arg_item->data = NULL;
				} else arg = NULL;
			} else switch ( *src) {
				    case 'c': 	arg = channel->name; 
				    	break;
				    case 'n': 	arg = channel->connection ? 
				    			channel->connection->data.irc->nick :
				    			VAR_STR( VAR_NICKNAME); 
				    	break;
				    case '+': 	merge = 1; arg = NULL;
				    	break;
				    default :
				    		g_warning( "Unrecognized expansion character : %c", *src);
				    		return NULL;
				}
			if ( arg)  		
				while ( *arg && size) { *dst++ = *arg++; size--;}
			expand = 0;
		} else { *dst++ = *src; size--;}
                
                if ( merge) {
                        arg_item = arg_list;
                        while ( arg_item) {
                                arg = (char*) arg_item->data;
                                arg_item = arg_item->next;
                                if ( arg) {
                                        size--;
                                	while ( *arg && size) { *dst++ = *arg++; size--;}
                                        *dst++ = ' ';
                                }
                        }
                        dst--;
                        merge = 0;
		}
		
	} while ( *++src && size);
        
        *dst = '\0';
        
        return g_strdup( gb_tmp_buff);
                
}
  /*******************/
 /*** Ignore List ***/
/*******************/

  /*********************/
 /*** User commands ***/
/*********************/

void
pong_server( IRCMessage *param) {
        IRCMessage reply;
        char *last;
        
        last       = g_list_last_data( param->arg_list);
        reply.connection   = param->connection;
        reply.message_type = USER_COMMAND;
        reply.def.ucmd = &gb_ucmd_messages[ IRC_PONG];
        reply.token    = reply.def.ucmd->token_num;
        reply.arg_list = last ? g_list_append( NULL, last) : NULL;
        reply.arg_trailing = -1;
	reply.nick = param->nick;
	reply.user = param->user;
	reply.address = param->address;
        parse_user_parameters( NULL, &reply);
}

/*** Variables ***/


UVariable*
new_variable( char *name, int type, void *value) {
        UVariable *var;
        
        g_return_val_if_fail( name != NULL, FALSE);
        
        var = g_malloc( sizeof( UVariable));
        var->name  = g_strdup( name);
        var->value = NULL;
        var->arg_type = type;
        g_hash_table_insert( gb_uvariable_hash, var->name, var);
        
        set_variable( var, value);
        return var;
}
        
UVariable*
new_variable_from_string( char *name, char *value) {
        UVariable *var;
        
        g_return_val_if_fail( name != NULL, FALSE);
        
	var = g_malloc( sizeof( UVariable));
	var->name  = g_strdup( name);
	       
	if ( value) {
		char *tmp;
	       		
	       	tmp = value;
	       	while ( *tmp) 
	       		if ( !isdigit(*tmp)) 
	       			break;
	       		else
	       			tmp++;
	       	if ( *tmp) {
	       	      if ( (tmp = strchr( value, ' '))) {
	       	              var->arg_type = ARG_STR;
	       	              (char*) var->value = g_strdup( tmp);
	       	      } else {
	       	              var->arg_type = ARG_TRAILING;
	       	              (char*) var->value = g_strdup( value);
	       	      }
	       	} else {
	       	      var->arg_type = ARG_INT;
	       	      var->value = g_malloc( sizeof( int));
	       	      *((int*) var->value) = strtol( value, (char**) NULL, 10);
	       	}
       	}
	       	
       	g_hash_table_insert( gb_uvariable_hash, var->name, var);
       	return var;
}
                
void
set_variable( UVariable *var, void *value) {
        g_return_if_fail( var != NULL);

        if ( var->value)
              	g_free( var->value);
	if ( value) {
		switch( var->arg_type) {
                        case ARG_INT:
                        case ARG_UINT:
                        	var->value = g_malloc( sizeof( int));
                        	*((int*) var->value) = *((int*) value);
                        	break;
                        case ARG_CHAR:
                        	var->value = g_malloc( sizeof( char));
                        	*((char*) var->value) = *((char*) value);
                        	break;
                        case ARG_STR:
                        	if ( !strchr( value, ' '))
                        		(char*) var->value = g_strdup( (char*) value);
                        	break;
                        case ARG_TRAILING:
                        	(char*) var->value = g_strdup( (char*) value);
                        	break;
                        default:
                        	var->value = NULL;
                        	g_warning( "Unknown argument type %d", var->arg_type);
                        	break;
              }
       } else 
       		var->value = NULL;
}

void
set_variable_from_string( UVariable *var, char *value) {
        g_return_if_fail( var != NULL);

        if ( var->value)
              	g_free( var->value);
	if ( value) {
		switch( var->arg_type) {
                        case ARG_INT:
                        case ARG_UINT:
                         	var->value = g_malloc( sizeof( int));
                       		*((int*) var->value) = atoi( value);
                        	break;
                        case ARG_CHAR:
                        	var->value = g_malloc( sizeof( char));
                        	*((char*) var->value) = *((char*) value);
                        	break;
                        case ARG_STR:
                        	if ( !strchr( value, ' '))
                        		(char*) var->value = g_strdup( value);
                        	break;
                        case ARG_TRAILING:
                        	(char*) var->value = g_strdup( value);
                        	break;
                        default:
                        	var->value = NULL;
                        	g_warning( "Unknown argument type %d", var->arg_type);
                        	break;
              }
       } else 
       		var->value = NULL;

}
        
UVariable*
cmd_set_variable( IRCMessage *param, Channel *channel) {
        char *name, *value;
        UVariable *var;
        
        name  = g_list_nth_data( param->arg_list, 0);
        value = g_list_nth_data( param->arg_list, 1);
        
        if ( name == NULL) {
                if ( channel)
        		show_variables( channel);
        	return FALSE;
        }
        
        var = g_hash_table_lookup( gb_uvariable_hash, name);
        if ( var)
                set_variable_from_string( var, value);
	else
		var = new_variable_from_string( name, value);
		
	if ( channel) {
	       snprintf( gb_tmp_buff, TMP_BUFF_SIZE, "%s value set to %s\n", name, value);
	       gui_window_display_message( channel->window, gb_tmp_buff, PROPERTY_OTHER);
	}
	 
	return var;
}

void
show_variable( char *key, UVariable *var, Channel *channel) {
         switch ( var->arg_type) {
                case ARG_STR:
                case ARG_TRAILING:
                	sprintf( gb_tmp_buff, "  %-20s= %s\n", (char*) var->name, (char*) var->value);
                	break;
                case ARG_INT:
                case ARG_UINT:
                	sprintf( gb_tmp_buff, "  %-20s= %d\n", (char*) var->name, *((int*) var->value));
                	break;
                case ARG_CHAR:
                	sprintf( gb_tmp_buff, "  %-20s= %c\n", (char*) var->name, *((char*) var->value));
                	break;
                default:
                	return;
         }
         gui_window_display_message( channel->window, gb_tmp_buff, PROPERTY_OTHER);
}

void
show_variables( Channel *channel) {
        gui_window_display_message( channel->window, "Variables :\n", PROPERTY_OTHER);
        g_hash_table_foreach( gb_uvariable_hash, (GHFunc) show_variable, channel);
}

void*
check_variable_type( UVariable *var, gint arg_type) {
        
        if ( var == NULL) {
                g_warning( "Trying to cast an empty variable !!!");
                return NULL;
        }
        
        if ( var->arg_type != arg_type) {
                g_warning( "Invalid cast of variable %s from %d to %d", var->name, var->arg_type, arg_type);
                return NULL;
        }
        
        return var->value;
}

/*** log files ***/

void
cmd_log( IRCMessage *param, Channel *channel) {
        char *target, *logfile, *dir, *wd;
        FILE *file;
        
        g_return_if_fail( param != NULL);
        g_return_if_fail( channel != NULL);
        
        target  = g_list_nth_data( param->arg_list, 0);
        logfile = g_list_nth_data( param->arg_list, 1);
        
        if ( target) {
        	channel = channel_find( param->connection->channels, target);
        	if ( !channel) {
        		logfile = target;
        		target  = channel->name;
        	}
        	if ( !logfile)
        		logfile = channel->name;
        } else
                target = logfile = channel->name;
        
        wd = getcwd( NULL, 0);
        if ( (dir = VAR_STR( VAR_LOGFILES_DIR)))
        	chdir( dir);
        	
        file = fopen( logfile, "a");
        if ( file) {
                fprintf( file, "# IRC log of %s started %s\n", target, get_local_time());
		channel_set_logfile( channel, file);
        } else
                g_warning( "Can't open %s : %s", logfile, strerror( errno));
        
        if ( wd) {
                chdir( wd);
                g_free( wd);
        }
}

/*** window levels ***/

 
void
show_levels( Channel *channel) {
        int   *levels;
        
        levels = &channel->window->levels;
        
        sprintf( gb_tmp_buff, "Window's level : ");
        if ( *levels == LEVEL_ALL) 
                strcat( gb_tmp_buff, "ALL");
        else {
		if ( FLAG_IS_SET( *levels, LEVEL_PUBLIC))
                	strcat( gb_tmp_buff, "PUBLIC ");
		if ( FLAG_IS_SET( *levels, LEVEL_MSGS))
                	strcat( gb_tmp_buff, "MSGS ");
		if ( FLAG_IS_SET( *levels, LEVEL_NOTICES))
                	strcat( gb_tmp_buff, "NOTICES ");
		if ( FLAG_IS_SET( *levels, LEVEL_WALLOPS))
                	strcat( gb_tmp_buff, "WALLOPS ");
		if ( FLAG_IS_SET( *levels, LEVEL_ACTIONS))
                	strcat( gb_tmp_buff, "ACTIONS ");
		if ( FLAG_IS_SET( *levels, LEVEL_REPLIES))
                	strcat( gb_tmp_buff, "REPLIES ");
		if ( FLAG_IS_SET( *levels, LEVEL_JOINS))
                	strcat( gb_tmp_buff, "JOINS ");
       }
       
       strcat( gb_tmp_buff, "\n");
       gui_window_display_message( channel->window, gb_tmp_buff, PROPERTY_OTHER);
			
}

void
cmd_level( IRCMessage *param, Channel *channel) {
        char  *level, *args;
        int   *levels;
        int   set;
        
        g_return_if_fail( param != NULL);
        g_return_if_fail( channel != NULL);
        
        if ( param->arg_trailing == -1) {
                show_levels( channel);
        	return;
        }
        	
        levels = &channel->window->levels;
        args = g_list_last_data( param->arg_list);
        get_next_item( &args, &level, ' ');
        while ( level && *level) {                
                switch ( *level) {
                        case '+':
                        	set = 1; level++; 
                        	break;
                        case '-':
                        	set = 0; level++; 
                        	break;
                        default:
                        	set = 1; 
                        	break;
                }
                upper_string( level);
                if ( !strcmp( level, "ALL"))
                	SWITCH_FLAG( *levels, LEVEL_ALL, set);
                else if ( !strcmp( level, "PUBLIC"))
                	SWITCH_FLAG( *levels, LEVEL_PUBLIC, set);
                else if ( !strcmp( level, "MSGS"))
                	SWITCH_FLAG( *levels, LEVEL_MSGS, set);
                else if ( !strcmp( level, "NOTICE"))
                	SWITCH_FLAG( *levels, LEVEL_NOTICES, set);
                else if ( !strcmp( level, "WALLOPS"))
                	SWITCH_FLAG( *levels, LEVEL_WALLOPS, set);
                else if ( !strcmp( level, "ACTIONS"))
                	SWITCH_FLAG( *levels, LEVEL_ACTIONS, set);
                else if ( !strcmp( level, "REPLIES"))
                	SWITCH_FLAG( *levels, LEVEL_REPLIES, set);
                else if ( !strcmp( level, "JOINS"))
                	SWITCH_FLAG( *levels, LEVEL_JOINS, set);
                else {
                        g_warning( "%s : unrecognized level", level);
                        break;
                }
                
        	get_next_item( &args, &level, ' ');
       }
       show_levels( channel);
}
                
void		
cmd_server( Channel *channel, char *server, char *port) {
        int   portn;
                
        portn = port ? atoi( port) : VAR_UINT( VAR_DEFAULT_PORT);
        
        if ( channel->connection == NULL) {
        	if ( (channel->connection = server_connection_new( server, portn))) {
        	        connection_append_channel( channel->connection, channel);
			//gui_window_update_status( channel->window);
        		gui_window_display_message( channel->window, 
        			"Connection in progress ...\n", PROPERTY_NUMERIC_REPLY);
        	} else
        		gui_window_display_message( channel->window, 
        			"Can't connect to server ...\n", PROPERTY_ERROR);
        } else {
        	if ( server_connection_reuse( channel->connection, server, portn)) {
        		gui_window_display_message( channel->window, 
        			"Connection in progress ...\n", PROPERTY_NUMERIC_REPLY);
        	} else
        		gui_window_display_message( channel->window, 
        			"Can't connect to server ...\n", PROPERTY_ERROR);
        		
       }
}	      
        
void
cmd_leave( Connection *connection, Channel *channel) {
        g_return_if_fail( channel != NULL);
       
	if ( !connection) {
	       if ( !FLAG_IS_SET( channel->flags, CHANNEL_MASTER))
		         channel_free( channel);
	} else if ( FLAG_IS_SET(connection->flags, CON_IS_DCC)) {
		gui_window_display_message( connection->default_channel->window,
		    	"closing dcc chat ...\n", PROPERTY_NUMERIC_REPLY);
		connection_free( connection);
	} else if ( IS_IRC_CHANNEL( channel->name) &&
		    FLAG_IS_SET( channel->flags, CHANNEL_JOINED) && 
		    !FLAG_IS_SET( channel->flags, CHANNEL_LEAVED)) {
		      	/* Send a PART <channel> message to the server */
	        	IRCMessage param;
	        	
	        	param.connection = connection;
	        	param.message_type = USER_COMMAND;
	        	param.arg_list = g_list_append( NULL, channel->name);
	        	param.arg_trailing = 0;
	        	param.def.ucmd = &gb_ucmd_messages[IRC_PART];
	        	param.token = param.def.ucmd->token_num;
	        	param.nick = param.user = param.address = NULL;
	        	
	        	parse_user_parameters( connection->default_channel, &param);
	} else {
	        connection_remove_channel( connection, channel);
		if ( connection->channels == NULL) {
		      //gui_window_display_message( connection->default_channel->window,
		      //	"closing connection ...\n", PROPERTY_NUMERIC_REPLY);
		      connection_free( connection);
		}
		
	}
}
        
