/*
 *    gnirc - Another IRC client for X 
 *    Copyright (C) 1998  david Ramboz <dramboz@imaginet.fr>
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>

#include <gtk/gtk.h>
#include "gnirc.h"
#include "enums.h"

int		dcc_send_init			( DCCParam *dcc, char *path);

extern  char		*gb_tmp_buff;
extern	UCmd 		gb_ucmd_messages[];
extern	UVariable	gb_uvariables[];
extern  unsigned long int gb_address;

char   *chat_argument  = "chat"; 

GList  *gb_dcc_requests = NULL;
 
Connection*
dcc_answer_new( IRCMessage *param) {
	static char my_address[32] = "";
        Connection *connection;
        DCCParam *dcc;
        char	 *arg;
        char	 *nick;
        int	  port;
        
        g_return_val_if_fail( param != NULL, FALSE);
                       
        connection = connection_new();
        if ( !((connection->sock = socket_new()) &&
               (port = socket_listen( connection->sock, 0)) &&
               connection_register( connection)) ) {
             	connection_free( connection);
             	return FALSE;
        }
        
        if ( *my_address == '\0') 
                sprintf( my_address, "%ld", gb_address);
                
        dcc = g_malloc( sizeof( DCCParam));
        connection->data = dcc;
        connection->input_func = dcc_input;
        connection->flags = CON_LISTEN | CON_IS_DCC;
        connection->default_channel = param->connection->default_channel;
        connection->port = g_malloc( 8);
        sprintf( connection->port, "%d", htons(port));
        
        arg  = g_list_last_data( param->arg_list);
        nick = g_list_nth_data( param->arg_list, 0);
        dcc->token = param->token;
        dcc->nick  = g_strdup( nick);
        dcc->user  = NULL;
        dcc->send  = NULL;
        
        switch ( dcc->token) {
                case CTCP_DCC_CHAT_A:
                	param->arg_list = g_list_append( param->arg_list, 
                				chat_argument);
                	param->arg_list = g_list_append( param->arg_list, 
                				my_address);
                	param->arg_list = g_list_append( param->arg_list, 
                				connection->port);
                	break;
                case CTCP_DCC_SEND_A:
                	if ( dcc_send_init( dcc, arg)) {
                	       param->arg_list = g_list_append( param->arg_list, 
                	       				my_address);
                	       param->arg_list = g_list_append( param->arg_list, 
                	       				connection->port);
                	       param->arg_list = g_list_append( param->arg_list, 
                	       				dcc->send->size);
                	} else {
                	       dcc_free( dcc);
                	       return FALSE;
                	}
                	break;
                default: /* this shouldn't happen ( the message will be discarded earlier) */
                	dcc_free( dcc);
                	return FALSE;
                	break;
        }
        
        
        return connection;
}

Connection*
dcc_request_new( IRCMessage *param) {
        Connection *connection;
        DCCParam *dcc;
        char	 *last;
        char	 *arg;
        char	 *host;
        char	 *port;
        
        g_return_val_if_fail( param != NULL, FALSE);
        
        last = g_list_last_data( param->arg_list);
        get_next_item( &last, &arg , ' ');
        get_next_item( &last, &host, ' ');
        get_next_item( &last, &port, ' ');
        
        if ( arg == NULL || host == NULL || port == NULL) {
                g_warning( "Bad dcc request from %s", param->nick);
                return FALSE;
        }
                
        connection = connection_new();
        
        dcc = g_malloc( sizeof( DCCParam));
        connection->data = dcc;
        connection->input_func = dcc_input;
        connection->flags = CON_IS_DCC;
	connection->default_channel = param->connection->default_channel;
        connection->address = g_strdup( host);
        connection->port = g_strdup( port);
        
        dcc->token = param->token;
        dcc->nick = g_strdup( param->nick);
        dcc->user = g_strdup( param->user);
        dcc->send = NULL;
        
        switch ( dcc->token) {
        	case CTCP_DCC_CHAT_Q:
        		break;
        	case CTCP_DCC_SEND_Q:
        		if ( !dcc_send_init( dcc, arg)) {
        		      dcc_free( dcc);
        		      return FALSE;
        		}
        		break;
        	default:
        		break;
        }
        
        gb_dcc_requests = g_list_append( gb_dcc_requests, connection);
        
        return connection;
}
        			
void
dcc_request_accept( Connection *connection) {
        DCCParam *dcc;
        Channel  *channel;
        GList	 *item;
        
        g_return_if_fail( connection != NULL);
        
        if ( !(item = g_list_find( gb_dcc_requests, connection))) {
                
                return;
        }
        
        dcc = connection->data.dcc;
        if ( (connection->sock = socket_new()) &&
             socket_connect_by_name(connection->sock, connection->address, atoi(connection->port)) &&
             connection_register( connection) ) 
		switch ( dcc->token) {
        		case CTCP_DCC_CHAT_Q:
				channel = channel_new( connection, 
					       connection->default_channel, 
			  		       dcc->nick);
			 	connection->default_channel = channel;
				connection_append_channel( connection, channel);
				sprintf( gb_tmp_buff, "DCC chat connection with %s (%s:%s) established\n",
					dcc->nick, 
					long_ip_to_str( ntohl(atol( connection->address))), 
					connection->port);
				gui_window_display_message( channel->window, gb_tmp_buff, PROPERTY_NUMERIC_REPLY);
        			break;
        		case CTCP_DCC_SEND_Q:
				sprintf( gb_tmp_buff, "DCC send connection with %s (%s:%s) established\n",
					dcc->nick, 
					long_ip_to_str( ntohl(atol( connection->address))), 
					connection->port);
				gui_window_display_message( connection->default_channel->window, 
					gb_tmp_buff, PROPERTY_NUMERIC_REPLY);
        			break;
        		default:
        			break;
        	}
        else
        	connection_free( connection);
        	
        gb_dcc_requests = g_list_remove_link( gb_dcc_requests, item);
}

Connection*
dcc_request_find( int token, char *arg) {
        GList *connection_item;
        Connection *connection;
        
        connection_item = gb_dcc_requests;
        while( connection_item) {
                connection = connection_item->data;
                connection_item = connection_item->next;
                switch ( token) { 
                	case CTCP_DCC_SEND_Q:
                	case CTCP_DCC_SEND_A:
                		if ( !strcmp( arg, connection->data.dcc->send->filename))
                			return connection;
                	case CTCP_DCC_CHAT_A:
                	case CTCP_DCC_CHAT_Q:
                		if ( !strcmp( arg, connection->data.dcc->nick))
                			return connection;
                		break;
                	default:
                		break;
               }
        }
        return FALSE;
}

void
dcc_free( DCCParam *dcc) {
        g_return_if_fail( dcc != NULL);
                        	
        if ( dcc->nick)
        	g_free( dcc->nick);
        	
        if ( dcc->send) {
                if ( dcc->send->file)
                        fclose( dcc->send->file);
                g_free( dcc->send->filename);
        	g_free( dcc->send);
        }
        
        g_free( dcc);
}

int
dcc_send_init( DCCParam *dcc, char *path) {
        DCCSendParam *send;
        FILE *file;
        struct stat stat_buff;
        char *filename, *wd, *dir;
        
        filename = strrchr( path, '/');
        if ( !filename)
        	filename = path;
        
        wd = getcwd( NULL, 0);
        dir = VAR_STR( VAR_DCC_SEND_DIRECTORY);
        if ( dir)
        	chdir( dir);
        
        if ( dcc->token == CTCP_DCC_SEND_A)
        	file = fopen( path, "r");
        else /* Todo : check if the file exists !!! */
                file = fopen( filename, "w");
        
        if ( wd) {
                chdir( wd);
                g_free( wd);
        }
        
        if ( !file) {
                g_warning( "Can't open %s : %s", filename, strerror( errno));
        	return FALSE;
        }
        
        if ( fstat( fileno( file), &stat_buff)) {
                g_warning( "Can't stat file %s : %s", filename, strerror( errno));
        	return FALSE;
        }
                	
        send = g_malloc( sizeof( DCCSendParam));
        send->filename = g_strdup( filename);
        send->file = file;
        send->pos  = 0;
        sprintf( send->size, "%d", (unsigned int) stat_buff.st_size);
        
        dcc->send = send;
        return TRUE;
}       
        
void
dcc_chat_parse_line( Connection *connection, char *line) {
	IRCMessage param;
        DCCParam *dcc;
        
        g_return_if_fail( connection != NULL);
        g_return_if_fail( line != NULL);
        
        dcc = connection->data.dcc;

       	param.connection   = connection;
        param.message_type = SERVER_COMMAND;
        param.token = IRC_PRIVMSG;
        param.def.ucmd = &gb_ucmd_messages[IRC_PRIVMSG];
        param.arg_list = g_list_append( NULL, line);
        param.arg_trailing = -1;
        param.nick = dcc->nick;
        param.user = dcc->user;
        param.address = connection->address;
        
        channel_append_message( connection->default_channel, &param);
        
        g_list_free( param.arg_list);
}
        
unsigned int
dcc_send_read_packet( Connection *connection, unsigned int ack) {
        DCCParam *dcc;
        unsigned int byte_read = 0;
        
        dcc = connection->data.dcc;
        
	if ( ack == dcc->send->pos) {
		/* read the next packet */
	       
		byte_read = fread( connection->buffer.buff, 1, INPUT_BUFFER_SIZE, dcc->send->file);
		if ( byte_read != INPUT_BUFFER_SIZE && ferror( dcc->send->file)) {
			/* an error occured */
 			g_warning( "fread : %s", strerror(ferror( dcc->send->file)));
			return 0;
       		} 
       		dcc->send->pos += byte_read;
       		connection->buffer.next_byte = connection->buffer.buff + byte_read;
       	} else {
       	        /* the previous packet hasn't been received */
       	        byte_read = connection->buffer.next_byte - connection->buffer.buff;
       	}
       	
       	return byte_read;
}

unsigned int
dcc_send_write_packet( Connection *connection, unsigned int byte_read) {
        DCCParam *dcc;
        
        dcc = connection->data.dcc;
       	
        if ( fwrite( connection->buffer.buff, byte_read, 1, dcc->send->file) != 1) {
		/* an error occured */
		g_warning( "fwrite : %s", strerror(ferror( dcc->send->file)));
		return 0;
      	} else
      		dcc->send->pos += byte_read;
      	return dcc->send->pos;
}

	        
