
#ifndef __GNIRC_H__
#define __GNIRC_H__

#include <gtk/gtk.h>

#ifdef __cplusplus
extern "C"
{
#endif /* __cplusplus */

/*
#undef g_return_if_fail
#undef g_return_val_if_fail
#define g_return_if_fail(c) if ( c) return 
#define g_return_val_if_fail(c, v) if ( c) return v
*/

#define PROGRAM_NAME 		"gnirc"
#define PROGRAM_VERSION 	"0.1a"
#define PROGRAM_SOURCE_HOST  	"none"
#define PROGRAM_SOURCE_DIR   	"none"
#define PROGRAM_SOURCE_FILES 	"none"
#define TEMP_SIZE 1024
#define TMP_BUFF_SIZE 1024

typedef struct _InputBuffer     InputBuffer;
typedef struct _IRCServer	IRCServer;
typedef struct _UserInfo     	UserInfo;
typedef struct _Connection  	Connection;
typedef struct _IRCParam	IRCParam;
typedef struct _UCmd		UCmd;
typedef struct _URply		URply;
typedef struct _IRCMessage	IRCMessage;
typedef struct _UAlias		UAlias;
typedef struct _UVariable	UVariable;
typedef struct _DCCParam	DCCParam;
typedef struct _DCCSendParam	DCCSendParam;
typedef struct _Channel	    	Channel;
typedef struct _GUIWindow 	GUIWindow;
typedef struct _GUICList 	GUICList;
typedef struct _GUIServerList	GUIServerList;
typedef struct _Icon		Icon;
typedef struct _TextProperty TextProperty;
typedef struct _IRCMenu	     IRCMenu;
typedef struct _IRCToolbar   IRCToolbar;
typedef struct _IRCNetwork   IRCNetwork;
typedef struct _History History;

typedef void  (*FreeFunc) (void *data);
typedef void  (*UFunc) ( UCmd *command, char **arguments);
typedef void  (*InputFunc) (Connection *connection);

/* how often do we need to check for input from the server ? */
/* 0.2s seem to be fast enough 	*/
#define DEFAULT_INPUT_TIMEOUT 200

void	exit_program		( int err);

  /***********************/
 /*** ircio.c headers ***/
/***********************/

/* An irc message is no more than 512 bytes. So each time we put 512 bytes in the
   buffer and parse it there will remain no more than 511 bytes in the buffer. 
   To read the next message in a buffer we will need 511+512= 1023 bytes 	
   the 1024'th  byte is used to hold an extra '\0' so that we can manipulate the 
   message as a string.( null char are illegal in messages)  				
*/
#define INPUT_BUFFER_SIZE 1024

struct _InputBuffer {
        char buff[INPUT_BUFFER_SIZE];
        char  *next_byte; /* next byte to be read from server */
        char  *next_line; /* next line to be parsed */
};


struct _Connection {
        int		sock;
        char		*address;
        char		*port;
        InputFunc	input_func;
        InputBuffer 	buffer;
        int		flags;
        GList		*channels;  
        Channel		*default_channel;      
        char		*target;	 
	union {
                DCCParam *dcc;
                IRCParam *irc;
        } data;
};


Connection*	server_connection_new	( char *server, int port);
int		server_connection_reuse	( Connection *connection, char *server, int port);
int		socket_new	();
int		socket_listen	( int sock, int port);
int		socket_connect	( int sock, unsigned long int host, int port);
int		socket_connect_by_name( int sock, char *host, int port);    

Connection*	connection_new		();
void		connection_free		( Connection *connection);
void		connection_free_all	();
Channel*	connection_new_channel	( Connection *connection, char *name);
void		connection_append_channel( Connection *connection, Channel *channel);
void		connection_remove_channel( Connection *connection, Channel *channel);
int		connection_register	( Connection *connection);
void		connection_unregister	( Connection *connection);
Connection*	connection_find		( int sock);
gint		connection_input_timeout();
void		connection_error	( Connection *connection, int error);
char*		connection_get_line	( Connection *connection, char *separator);

void		irc_connect	( Connection *connection, char *nick, char *realname, char *password);
void		irc_disconnect 	( Connection *param, char *reason);
void		irc_input	( Connection *param);
void		irc_send	( Connection *param, void *data, unsigned int size);
void		dcc_input	( Connection *connection);

  /*********************/
 /*** irc.c headers ***/
/*********************/

#define FLAG_IS_SET( obj, flag)		( (obj) & (flag))
#define SET_FLAG( obj, flag)		( (obj) |= (flag))
#define UNSET_FLAG( obj, flag)		( (obj) &= ~(flag))
#define SWITCH_FLAG( obj, flag, set)	( (set) ? SET_FLAG( (obj), (flag)) : \
						  UNSET_FLAG( (obj), (flag))) 
#define IS_IRC_CHANNEL( obj)		( (*(obj) == '#' || *(obj) == '&') )

struct _IRCParam {
        char	*nick;
        int	user_mode;
        int	connected : 1;
};

struct _UCmd {
        char	*token;		
        unsigned short int token_num;
        unsigned short int min_arg;
	unsigned short int *args;
	char	*format;   /* format used to display the message */
};

struct _URplyFormat {
        char	*label;
        unsigned short int arg_type;
};

struct _URply {
        unsigned short int token;
        unsigned short int *args;
       	char    *format;
};


/* message_type : SERVER_COMMAND, SERVER_NUMERIC_REPLY, USER_COMMAND 	
   token : the real token number which is different from the on found 	
 	in the 'def' field when the message is a ctcp message	     	
   arg_trailing : only used when message_type is equal to  USER_COMMAND
   	to keep the "trailing parameter" ( cf RFC)
   arg_list	: the list of arguments 				
*/
struct _IRCMessage {
        Connection   *connection;
        unsigned short int message_type;
        unsigned short int token;
        union {
       		UCmd	*ucmd;
       		URply	*urply;
       	} def;
        GList	*arg_list;
        int	arg_trailing;
        
        char	*nick;
        char	*user;
        char	*address;
};

void	initialize_parser	( ); 
int	parse_server_message	( Connection *connection, char *server_command);
int	parse_server_parameters	( IRCMessage *param);
int	parse_user_message	( Channel *channel, char *user_message);
int	parse_user_command	( Channel *channel, char *user_command);
int	parse_user_parameters	( Channel *channel, IRCMessage *param);

IRCMessage* irc_message_copy	( IRCMessage *param);
void	    irc_message_delete	( IRCMessage *param);

  /**************************/
 /*** commands.c headers ***/
/**************************/

#define GET_VAR( var) ( gb_uvariables[(var)] )
#define SET_VAR( var, value) ( set_variable( &gb_uvariables[(var)], (value)) )
#define VAR_VALUE( var) ( gb_uvariables[(var)].value )
#define VAR_INT( var) ( *((gint*) check_variable_type( &GET_VAR(var), ARG_INT)) )
#define VAR_UINT( var) ( *((guint*) check_variable_type( &GET_VAR(var), ARG_UINT)) )
#define VAR_CHAR( var) ( *((char*) check_variable_type( &GET_VAR(var), ARG_CHAR)) )
#define VAR_STR( var) ( (char*) check_variable_type( &GET_VAR(var), ARG_STR) )
#define VAR_TRAILING( var) ( (char*) check_variable_type( &GET_VAR(var), ARG_TRAILING) )
#define SET_IF_NULL( var, value) if ( !VAR_VALUE(var))  SET_VAR( (var), (value))

struct _UAlias {
        char *name;
        char *format;
};

struct _UVariable {
        char 	 *name;
        unsigned short int arg_type;
        void	 *value;
};

char*	    	expand_message		( IRCMessage *param);
void		format_message		( IRCMessage *param, Channel *channel, char **message, int *property);

int		new_alias		( Channel *channel, IRCMessage *param);
void		show_aliases		( Channel *channel);
char*		expand_alias		( Channel *channel, char *command, char *message);

void		pong_server		( IRCMessage *param);
UVariable*	new_variable		( char *name, int type, void *value);
UVariable*	new_variable_from_string( char *name, char *value);
void		set_variable		( UVariable *var, void *value);
void		set_variable_from_string( UVariable *var, char *value);
UVariable*	cmd_set_variable	( IRCMessage *param, Channel *channel);
void		show_variables		( Channel *channel);
void*		check_variable_type	( UVariable *var, gint arg_type);
void		cmd_log			( IRCMessage *param, Channel *channel);
void		cmd_level		( IRCMessage *param, Channel *channel);
void		cmd_server		( Channel *channel, char *server, char *port);
void		cmd_leave		( Connection *connection, Channel *channel);

  /*********************/
 /*** dcc.c headers ***/
/*********************/


struct _DCCSendParam {
        char *filename;
        FILE *file;
        char size[16];
        unsigned int pos;
};

struct _DCCParam {
        int	token;
        
        /* nickname and address of the "other side" of the connection */ 
        char	*nick;
        char	*user;
        int	flags;
        
        DCCSendParam *send;
};

Connection*	dcc_answer_new		( IRCMessage *param);
Connection*	dcc_request_new		( IRCMessage *param);
void		dcc_request_accept	( Connection *connection);
Connection*	dcc_request_find	( int token, char *arg);
void		dcc_free		( DCCParam *dcc);
void		dcc_chat_parse_line	( Connection *connection, char *line);
unsigned int	dcc_send_read_packet	( Connection *connection, unsigned int ack);
unsigned int	dcc_send_write_packet	( Connection *connection, unsigned int byte_read);

  /*************************/
 /*** channel.c headers ***/
/*************************/

/* Channels : to keep track of who is on each channel 			
	this is usefull to deliver some kind of server messages 		
	as NICK, QUIT which doesn't contains the channel name in which 	
	they occured 							
*/
struct _Channel {
        Connection *connection;
        char 	  *name;
        GList	  *users;
        GUIWindow *window;
        
        /* history of messages for that channel 		*/
        History	*data;
        char	*topic;
        int	mode;
        int	user_mode;
        
        FILE	*logfile;
        int	flags;
};

Channel*	channel_new		( Connection *connection, Channel *parent, char *name);
void		channel_free		( Channel *channel);
Channel*	channel_find		( GList *channel_list, char *name);
gint		channel_find_user	( Channel *channel, char *user);
void		channel_add_users	( Channel *channel, char *users);
gint		channel_del_user	( Channel *channel, char *user);
void		channel_change_user_mode( Channel *channel, char *nick, int mode, int set);
void		channel_set_topic	( Channel *channel, char *topic);
void		channel_set_logfile	( Channel *channel, FILE *logfile);
void		channel_append_message	( Channel *channel, void *param);
void		channel_redisplay_all	( Channel *channel);
void		channel_popup_window	( Channel *channel);
void		channel_popdown_window	( Channel *parent, Channel *channel);

void		user_nickname_changed	( GList *channel_list, char *old_nick, char *new_nick);
void		user_has_quit_irc	( GList *channel_list, char *nick);
void		wallop_nick_message	( GList *channel_list, void *msg, char *nick);

  /************************/
 /*** window.c headers ***/
/************************/


#define GUI_CLIST_UPDATE_TIMEOUT 1500

enum {
        PROPERTY_PRIVMSG,
        PROPERTY_NOTICE,
        PROPERTY_ACTION,
        PROPERTY_NUMERIC_REPLY,
        PROPERTY_CTCP_QUERY,
        PROPERTY_CTCP_REPLY,
        PROPERTY_ERROR,
        PROPERTY_OTHER,
        PROPERTY_LAST
        
} Properties;

struct _GUIWindow {
        GtkWidget  *window;
        GtkWidget  *text;
        GtkWidget  *status;
        GtkWidget  *cmdline;
        GtkWidget  *channel_list;
        GtkWidget  *user_list;
        
        Channel	   *current_channel;
        GList	   *channels;
        
        History	   *history; /* command history */
        guint	   max_lines;
        guint	   num_lines;
        gint	   dont_freeze : 1;
        gint	   levels;
};

struct _GUICList {
        GtkWidget *window;
        GtkWidget *clist;
        GList	  *items;
        int	  timeout_tag;
        GtkWidget *filesel;
};

struct _GUIServerList {
        GtkWidget *net_desc;
        GtkWidget *server;
        GtkWidget *ports;
        GtkWidget *list;
        IRCNetwork *current_network;
};

struct _Icon {
        GdkPixmap *pixmap;
        GdkBitmap *mask;
};

GUIWindow*	gui_window_new( );
void		gui_window_new_channel		( GUIWindow *window,
					  	  Channel   *channel);
void		gui_window_delete_channel	( GUIWindow *window,
					   	  Channel   *channel);
void		gui_window_select_channel	( GUIWindow *window, 
						  Channel *channel);
void		gui_window_display_message 	( GUIWindow *window, 
					     	  char *text, 
					     	  int property);		
void		gui_window_display_ircmessage 	( GUIWindow *window, 
					          IRCMessage *param);
void		gui_window_update_status	( GUIWindow *window);
void		gui_window_build_channel_list	( GUIWindow *window);
void		gui_window_build_user_list	( GUIWindow *window);
void		gui_window_add_user		( GUIWindow *window, gchar *nickname, gint pos);
void 		gui_window_delete_user		( GUIWindow *window, gint pos);
						  
GUICList*	gui_clist_new	( char *label, char **titles);
void		gui_clist_free	( GUICList *list);
void		gui_clist_append( GUICList *list, GList *items);
gint		gui_clist_update( GUICList *list);
void		gui_clist_start	( GUICList *list);
void		gui_clist_end	( GUICList *list);



GUIServerList*		gui_server_list_new		();

  /***********************/
 /*** prefs.c headers ***/
/***********************/

struct _TextProperty {
        GdkFont		*font;
        GdkColor	color;
};


struct _IRCMenu {
        char	*label;
        char	*command;
};

struct _IRCToolbar {
        char	*label;
        char	*command;
        char	*icon;
        char	*help;
};

struct _IRCServer {
        char	*label;
        char	*address; /* server address */
        int	family;   /* 'kind' of server ( eg dalnet ...) */
        int	port;
        GList	*ports;   /* port on which the server runs */
};

struct _IRCNetwork {
        char	*label;
        char 	*description;
        GList	*servers;
};

void	irc_prefs_init		();
int	parse_menus		( char *file);
int 	parse_ircrc		( char *file);
int	parse_server_list	( char *file);
int	parse_gtkrc		( char *file);        

  /**********************/
 /*** misc.c headers ***/
/**********************/

#define g_list_last_data( list) ( (list) ? g_list_last(list)->data : NULL )

struct _History {
        gint	pos_current;
        gint	pos_max;
        gint	vpos;	/* virtual position : used to walk through the history */
        void	**data;
        FreeFunc free;
};

/* Strings functions */
int		get_next_item		( char **src, char **dst, char sep);
int		get_next_string		( char **src, char **dst);
int		get_next_line		( char **src, char **dst);
void		lower_string		( char *src);
void		upper_string		( char *src);
int		compare_users		( char *user1, char *user2);
int 		compare_nicks		( char *nick1, char *nick2);

/* misc functions */
char*		get_local_time		();
guint		g_num_hash		( const gpointer pointer);
gint		g_num_compare		( const gpointer v1, const gpointer v2);
gint		g_list_find_pos		( GList *list, gpointer data);
gpointer	g_list_find_data	( GList *list, gpointer data);
gpointer	g_list_nth_data		( GList *list, gint pos);
char*		long_ip_to_str		( unsigned long int ip);

/* Simple history */
History*	history_new		( gint size, FreeFunc func);
void		history_free		( History *history);
void*		history_next		( History *history);
void*		history_previous	( History *history);
void*		history_first		( History *history);
void*		history_last		( History *history);
void		history_append		( History *history, void *data);
void		history_rewind		( History *history);
/* for debugging purpose */
void		history_dump_strings	( History *history);

GtkWidget*	new_pixmap 		( char      *filename,
	    				  GdkWindow *window,
	    				  GdkColor  *background);
void		destroy_widget_cb	( GtkWidget *w, gpointer *data);
void		gtk_text_scroll_down_to_end	( GtkText *text);
void		gtk_text_delete_first_line	( GtkText *text);
  
  /***********************/
 /*** Utils.c headers ***/
/***********************/

void		init_messages_handlers		();

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif
