/*
 *    gnirc - Another IRC client for X 
 *    Copyright (C) 1998  david Ramboz <dramboz@imaginet.fr>
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 */
 
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <gtk/gtk.h>

#include "gnirc.h"
#include "enums.h"
#include "ircdefs.h"

extern	UAlias		gb_ualiases[];
extern  char		gb_default_channel_name[16];
extern	char		*gb_tmp_buff;
        
/* hash tables are so easy to use under gtk+ :-)*/
GHashTable *gb_ucmd_hash;
GHashTable *gb_urply_hash;
GHashTable *gb_ualias_hash;
GHashTable *gb_uvariable_hash;

char		gb_small_buff[64];
GUICList	*gb_gui_clist	= NULL;

char gb_rpl_list_label[16] = "Channel list";
char*
gb_rpl_list_titles[] = {
        "Channel name",
        "Visible",
        "Topic",
        NULL,
};

int		get_message_def		( IRCMessage *param, char *command);
void		get_ctcp_token_num	( IRCMessage *param, char *command);
void		ctcp_dequote		( IRCMessage *param);
void		ctcp_handle_query	( IRCMessage *param);


/* ctcp_handle_query : respond to ctcp queries */
void
ctcp_handle_query( IRCMessage *param) {
        IRCMessage reply;
        Connection *dcc_con;
        char	*arg;
   	char	buff[512];
   	int	answer = TRUE;
   	
        arg = g_list_nth_data( param->arg_list, 5); 
        
        if ( arg == NULL)
        	arg = "";
        
        switch ( param->token) {
                case CTCP_FINGER_Q:
                	sprintf( buff, "FINGER :%s (%s) has been idled %d seconds", 
				VAR_TRAILING( VAR_REALNAME), VAR_STR( VAR_EMAIL) , 0);
			break;
		case CTCP_VERSION_Q:
			sprintf( buff, "VERSION %s:%s:%s(%s)", VAR_STR( VAR_PROGRAM_NAME), 
				VAR_STR( VAR_PROGRAM_VERSION), VAR_STR( VAR_SYSTEM_TYPE), 
				VAR_STR( VAR_SYSTEM_VERSI0N));
			break;
		case CTCP_SOURCE_Q:
			sprintf( buff, "SOURCE  %s:%s:%s\r\n", VAR_STR( VAR_PROGRAM_SOURCE_HOST), 
				VAR_STR( VAR_PROGRAM_SOURCE_DIR), VAR_STR( VAR_PROGRAM_SOURCE_FILES));
			break;
		case CTCP_USERINFO_Q:
			sprintf( buff, "USERINFO :%s", VAR_STR( VAR_USERINFO));
			break;
		case CTCP_CLIENTINFO_Q:
			sprintf( buff, "CLIENTINFO :FINGER VERSION SOURCE USERINFO CLIENTINFO ERRMSG TIME PING");
			break;
		case CTCP_ERRMSG_Q:
			sprintf( buff, "ERRMSG %s:no error has occured", arg);
			break;
		case CTCP_TIME_Q:
			arg = get_local_time();
			sprintf( buff, "TIME :%s %s", arg,  *tzname);
			break;
		case CTCP_PING_Q:
			sprintf( buff, "PING %s", arg);
			break;
		case CTCP_ACTION_Q:
			answer = FALSE;
			break;
		case CTCP_DCC_CHAT_Q:
		case CTCP_DCC_SEND_Q:
			dcc_con = dcc_request_new( param);
			answer = FALSE;
			break;
		default:
			/* I don't know what is the right thing to do here 
			   sending an error message to all unknown queries or just 
			   discard them.
			   I've choosed to discard all queries that this program don't understand
			 */
			answer = FALSE;
			//sprintf( buff, "ERRMSG %s:Unknown query", arg);
			break;
	}
	
	if ( answer) {
		reply.connection   = param->connection;  	
   		reply.message_type = USER_COMMAND;
        	reply.def.ucmd = &gb_ucmd_messages[ IRC_NOTICE];
        	reply.token    = param->token;
        	reply.arg_list = g_list_append( NULL, param->nick);
        	reply.arg_trailing = g_list_length( reply.arg_list);
        	reply.arg_list = g_list_append( reply.arg_list, buff);
        	reply.nick     = NULL;
        	reply.user     = NULL;
        	reply.address  = NULL;
       		parse_user_parameters( NULL, &reply);
        
       		g_list_free( reply.arg_list);
        }
}

void
initialize_parser() {
        UCmd	*ucmd;
        URply	*urply;
        UAlias	*ualias;
        UVariable *uvar;
        
        gb_ucmd_hash = g_hash_table_new( g_str_hash, g_str_equal);
        ucmd = gb_ucmd_messages;
        while ( ucmd->token) {
                g_hash_table_insert( gb_ucmd_hash, ucmd->token, ucmd);
                ucmd++;
        }
        g_hash_table_thaw( gb_ucmd_hash);
        
        gb_urply_hash = g_hash_table_new( g_num_hash, g_num_compare);
        urply = gb_urply_messages;
        while( urply->token) {
                g_hash_table_insert( gb_urply_hash, &urply->token, urply);
                urply++;
        }
        g_hash_table_thaw( gb_urply_hash);
        
        gb_ualias_hash = g_hash_table_new( g_str_hash, g_str_equal);
        ualias = gb_ualiases;
        while ( ualias->name) {
                g_hash_table_insert( gb_ualias_hash, ualias->name, ualias);
                ualias++;
	}                
        
        gb_uvariable_hash = g_hash_table_new( g_str_hash, g_str_equal);
        uvar = gb_uvariables;
        while ( uvar->name) {
                g_hash_table_insert( gb_uvariable_hash, uvar->name, uvar);
                uvar++;
        }
}

void
get_ctcp_token_num( IRCMessage *param, char *command) {
        char	*tmp;
	int	request;
	
	get_next_item( &command, &tmp, ' ');
	if ( *command == ':')
		command++;
	upper_string( tmp);
	param->arg_list = g_list_append( param->arg_list, command);
	
	request = param->def.ucmd->token_num == IRC_PRIVMSG;
	if ( !strcmp( tmp, "FINGER"))
		param->token =  request ? CTCP_FINGER_Q    : CTCP_FINGER_A;
	else if ( !strcmp( tmp, "ACTION"))
		param->token =  request ? CTCP_ACTION_Q     : CTCP_ACTION_A;
	else if ( !strcmp( tmp, "PING"))
		param->token =  request ? CTCP_PING_Q       : CTCP_PING_A;
	else if ( !strcmp( tmp, "VERSION"))
		param->token =  request ? CTCP_VERSION_Q    : CTCP_VERSION_A;
	else if ( !strcmp( tmp, "SOURCE"))
		param->token =  request ? CTCP_SOURCE_Q     : CTCP_SOURCE_A;
	else if ( !strcmp( tmp, "USERINFO"))
		param->token =  request ? CTCP_USERINFO_Q   : CTCP_USERINFO_A;
	else if ( !strcmp( tmp, "CLIENTINFO"))
		param->token =  request ? CTCP_CLIENTINFO_Q : CTCP_CLIENTINFO_A;
	else if ( !strcmp( tmp, "ERRMSG"))
		param->token =  request ? CTCP_ERRMSG_Q     : CTCP_ERRMSG_A;
	else if ( !strcmp( tmp, "TIME"))
		param->token =  request ? CTCP_TIME_Q       : CTCP_TIME_A;
	else if ( !strcmp( tmp, "DCC")) {
		get_next_item( &command, &tmp, ' ');
		if ( *command == ':')
			command++;
		upper_string( tmp);
		if ( *command)
			param->arg_list = g_list_append( param->arg_list, command);
		
	        if ( !strcmp( tmp, "CHAT"))
	        	param->token =  param->message_type == USER_COMMAND ? 
	        		CTCP_DCC_CHAT_A : CTCP_DCC_CHAT_Q;
	        else if ( !strcmp( tmp, "SEND"))
	        	param->token =  param->message_type == USER_COMMAND ? 
	        		CTCP_DCC_SEND_A : CTCP_DCC_SEND_Q;
	} else
		param->token =  request ? CTCP_UNKNOWN_Q : CTCP_UNKNOWN_A;
}
        
/* ctcp_dequote : remove the '\001' quoting characters and notify if a ctcp	
	message was found by changing the param->token value			
*/
void
ctcp_dequote( IRCMessage *param) {
	char *src;
	char *dst;
	int isquoted;
			
	isquoted = 0;
	src  = dst = g_list_last_data( param->arg_list);
	
	if ( src == NULL)
		return;
		
	while ( *src) 
		if ( *src == '\001') {
			isquoted = 1;
			src++;
		} else *dst++ = *src++;
	*dst = '\0';
	
	if ( isquoted) {
	        src = g_list_last_data( param->arg_list);
		get_ctcp_token_num( param, src);			      		
	}
}

/* get_message_def : get the message "definition" by looking in 2 hash tables:
	gb_ucmd_hash  : contains rfc1459, ctcp and some user commands		
	gb_urply_hash : contains an entry for each numeric reply. 		
*/
int
get_message_def(  IRCMessage *param, char *command) {
        if ( isalpha( *command)) {
                if ( param->message_type != USER_COMMAND)
                	param->message_type = SERVER_COMMAND;
        	upper_string( command);
        	param->def.ucmd  = g_hash_table_lookup( gb_ucmd_hash, command);
        	if ( !param->def.ucmd) 
        		return 1;
        	param->token = param->def.ucmd->token_num;
        	
        } else {
                if ( param->message_type == USER_COMMAND) {
                        g_warning( "Unknown command :%s", command);
                        return 1;
                }
                param->message_type = SERVER_NUMERIC_REPLY;
                param->token	    = atoi( command);
                param->def.urply    = (URply*) g_hash_table_lookup( gb_urply_hash, &param->token);
                
                /* If the numeric reply is unknwon to us then display it without any change */
                if ( !param->def.urply) 
                        param->def.urply = &gb_urply_messages[0];
                
       }
       return 0;
}

/* get_user_message_arguments : get the message arguments. The distinction between	
	trailing and middle parameters is done by looking in the arguments type.	
	there's only two types of parameters usefull					
	for the moment ( ARG_TRAILING and ARG_STR)					
*/
int
get_user_message_arguments( IRCMessage *param, char *message) {
        unsigned short int  *arg_type;
        char *arg_value;
        int  arg_count;
        
        arg_type = param->message_type == SERVER_NUMERIC_REPLY ? 
        		param->def.urply->args :param->def.ucmd->args;
        if ( !arg_type)
        	return 0;
        	
        arg_count = 0;
        while ( *arg_type && *message) {
        	switch ( *arg_type) {
        	        /* one word */
              		case ARG_CHAR:
               		case ARG_INT:
               		case ARG_STR:
               		case ARG_NICK:
             		case ARG_USER:
              		case ARG_CHANNEL:
               		case ARG_SERVER:
             		case ARG_TIME:
             		case ARG_MODE_FLAG:
            		case ARG_STR_LIST_COMMA:
              		case ARG_NICK_LIST:
               		case ARG_CHANNEL_LIST:
            		case ARG_STR_LIST_SPACE:
              			/* all of those tokens are equivalent for the moment */
             			arg_type++;
         	     	 	get_next_item( &message, &arg_value, ' ');
      	     	 		param->arg_list = g_list_append( param->arg_list, arg_value);
      	     	 		arg_count++;
               	     		break;
               	     	/* a trailing parameter ( cf RFC1459) */ 
               		case ARG_TRAILING:
               		case ARG_REALNAME:
               		case ARG_TOPIC:
               		case ARG_TIME_STR:
               			param->arg_trailing = g_list_length( param->arg_list);
              	 	/* one string */
               	     	case ARG_QUOTE:
               	     		param->arg_list = g_list_append( param->arg_list, message);
             			arg_type++;
               	     		arg_count++;
               	      		break;
               	      	default:
               	      		/* this shouldn't happen */
                     		arg_count--;
             			arg_type++;
               	      		g_warning( "%d :Unknown argument code", *arg_type);
               	      		break;
                }
               	
       	}        
       	
       	if ( arg_count < param->def.ucmd->min_arg) {
       	        g_warning( "%s requires %d parameters at least", param->def.ucmd->token, param->def.ucmd->min_arg);
       	        if ( param->arg_list) {
       	        	g_list_free( param->arg_list);
       	        	param->arg_list = NULL;
       	        }
       	        return 1;
       	}
       	
        return 0;
}

/* get_server_message_arguments : find the arguments of the message which are of two kinds:	
	- middle parameter : equal to one word							
	- trailing parameter : one or more words				
	  trailing parameters are put after a ':' character					
*/
int
get_server_message_arguments( IRCMessage *param, char *message) {
        char *command;
        char *arg_value;
        
        get_next_item( &message, &command, ' ');
        if ( get_message_def( param, command))
        	return 1;
        while ( *message) {
                if ( *message ==  ':') {
                	param->arg_list = g_list_append( param->arg_list, ++message);
                	break;
                }
                get_next_item( &message, &arg_value, ' ');
                param->arg_list = g_list_append( param->arg_list, arg_value);
	}
	
	/* remove the first parameter of numeric replies since it seems (is?) 
	   to be always our nickname 
	*/
	if ( param->message_type == SERVER_NUMERIC_REPLY) {
	       if ( param->arg_list)
	       		param->arg_list = g_list_remove_link( param->arg_list, param->arg_list);
	}
	
	return 0;
}        

/* parse_server_parameters : dequote the message, display it to the right channel,	
	maintains user list... The real job is done here				
*/
int
parse_server_parameters( IRCMessage *param) {
        Channel *channel = NULL;
        char *target;
        char *last;
        
        int   show_msg = 1;
        int   is_channel;
        
	g_return_val_if_fail( param != NULL, FALSE);
	
	if ( param->token == IRC_PRIVMSG || param->token == IRC_NOTICE)
		ctcp_dequote( param);
	        
        last = g_list_last_data( param->arg_list);
        
        target = g_list_nth_data( param->arg_list, 0);
        if ( target) {
	        is_channel = IS_IRC_CHANNEL( target);
	        if ( !is_channel && param->nick)
 	              	target = param->nick;
       		channel = param->connection->channels ?
       				channel_find( param->connection->channels, target) : NULL;
	}    
       	if ( !channel && !param->connection->default_channel) 
        	return FALSE;
        	    	
	if ( param->message_type == SERVER_COMMAND) {
	       
	       switch( param->def.ucmd->token_num) {
	               case IRC_PING:
	               		pong_server( param);
	               		show_msg = 0;
	               		break;
	               case IRC_NOTICE:
	               		if ( param->token == CTCP_PING_A && last) {
	               		      gint ping_delay;
	               		      
	               		      ping_delay = atoi( last);
	               		      ping_delay = time(NULL) - ping_delay;
	               		      sprintf( gb_small_buff, "%d", ping_delay);
	               		      
				      param->arg_list = g_list_remove_link( param->arg_list, g_list_last( param->arg_list));
	               		      param->arg_list = g_list_append( param->arg_list, gb_small_buff);
	               		}
	               		break;
	               case IRC_PRIVMSG:
	               		if ( param->token != IRC_PRIVMSG && param->token != CTCP_ACTION_Q) {
	               		        /* It's a ctcp query */
	               			ctcp_handle_query( param);
	               		}
	               		break;
	               case IRC_JOIN:
	               		if ( channel && FLAG_IS_SET(channel->flags, CHANNEL_JOINED))
	               			channel_add_users( channel, param->nick);
	               		break;
	               case IRC_KICK:
	               		target = g_list_nth_data( param->arg_list, 1);
               		        if ( strcmp( target, param->connection->data.irc->nick))
               		      		channel_del_user( channel, target);
               		        else if ( channel) {
               		                SET_FLAG( channel->flags, CHANNEL_LEAVED);
               		                cmd_leave( param->connection, channel);
               		      		channel = NULL;
               		      	}
	               		break;
	               case IRC_PART:
               		        if ( strcmp( param->nick, param->connection->data.irc->nick))
               		      		channel_del_user( channel, param->nick);
               		        else if ( channel) {
               		                SET_FLAG( channel->flags, CHANNEL_LEAVED);
               		                cmd_leave( channel->connection, channel);
					channel = NULL;
               		      	}
	               		break;
	               case IRC_TOPIC:
	               		channel_set_topic( channel, last);
	               		break;
	               case IRC_NICK:
	               		//wallop_nick_message( param->connection->channels, param, param->nick);
	               		target = g_list_nth_data( param->arg_list, 0);
	               		if ( !strcmp( param->nick, param->connection->data.irc->nick)) {
	               			g_free( param->connection->data.irc->nick);
	               			param->connection->data.irc->nick = g_strdup( target);
	               			if ( channel)
	               				gui_window_update_status( channel->window);
	               		} else 
	               			user_nickname_changed( param->connection->channels, param->nick, target);
	               		//show_msg = 0;
	               		break;
	               case IRC_QUIT:
	               		//wallop_nick_message( param->connection->channels, param, param->nick);
	               		user_has_quit_irc( param->connection->channels, param->nick);
	               		//show_msg = 0;
	               		break;
	               case IRC_MODE:
	               		if ( is_channel) {
					char *mode = g_list_nth_data( param->arg_list, 1);
					char *arg1 = g_list_nth_data( param->arg_list, 2);
					int  set = *mode++ == '+';
					do {
					  switch ( *mode) {
					   case 'o': /* give/take channel operator privileges */
					   	if ( !strcmp( param->connection->data.irc->nick, arg1))
					   		SWITCH_FLAG( channel->user_mode, MODE_O, set);
					   	channel_change_user_mode( channel, arg1, MODE_O, set);
					   	break;
					   case 'v': /* give/take the ability to speak on a moderated channel */
					   	if ( !strcmp( param->connection->data.irc->nick, arg1))
					   		SWITCH_FLAG( channel->user_mode, MODE_V, set);
					   	channel_change_user_mode( channel, arg1, MODE_V, set);
					   	break;
					   case 'p': /* private channel flag */
					   	SWITCH_FLAG( channel->mode, MODE_P, set);
					   	break;
					   case 's': /* secret channel flag */
					   	SWITCH_FLAG( channel->mode, MODE_S, set);
					   	break;
					   case 'i': /* invite-only channel flag */
					   	SWITCH_FLAG( channel->mode, MODE_I, set);
					   	break;
					   case 't': /* topic settable by channel operator only flag */
					   	SWITCH_FLAG( channel->mode, MODE_T, set);
					   	break;
					   case 'n': /* no messages to channel from clients on the outside */
					   	SWITCH_FLAG( channel->mode, MODE_N, set);
					   	break;
					   case 'm': /* moderated channel */
					   	SWITCH_FLAG( channel->mode, MODE_M, set);
					   	break;
					   case 'l': /* set the user limit to channel */
					   	/* SWITCH_FLAG( channel->mode, MODE_L, set); */
					   	break;
					   case 'b': /* set a ban mask to keep users out */
					   	/* SWITCH_FLAG( channel->mode, MODE_P, set); */
					   	break;
					   case 'k': /* set a channel key (password) */
					   	SWITCH_FLAG( channel->mode, MODE_K, set);
					   	break;
					   default:
					   	break;
					  } 
					} while ( *mode && *(++mode) != ' ');
				} else {
				    if ( !strcmp( param->nick, param->connection->data.irc->nick)) { 
					char *mode = g_list_nth_data( param->arg_list, 1);
				    	int  set = *mode++ == '+';
				    	do {
				    	  switch ( *mode) {
				            case 'i': /* marks a users as invisible */
				            	SWITCH_FLAG( param->connection->data.irc->user_mode, MODE_I, set);
				            	break;
				            case 's': /* marks a user for receipt of server notices */
				            	SWITCH_FLAG( param->connection->data.irc->user_mode, MODE_S, set);
				            	break;
				            case 'w': /* user receives wallops */
				            	SWITCH_FLAG( param->connection->data.irc->user_mode, MODE_W, set);
				            	break;
				            case 'o': /* operator flag */
				            	SWITCH_FLAG( param->connection->data.irc->user_mode, MODE_O, set);
				            	break;
				            default:
				            	break;
				    	  } 
				    	} while ( *mode && *(++mode) != ' ');
				    }
				   //wallop_nick_message( param->connection->channels, param, param->nick);
	               		}
	               		break;
	               default:
	               		break;
	      }
	      
			
	} else if ( param->message_type == SERVER_NUMERIC_REPLY) {
	               	    	
	       switch ( param->token) {
	               case RPL_ENDOFMOTD:
	               		param->connection->data.irc->connected = TRUE;
	               		break;
	               case RPL_LISTSTART:
	               		gb_gui_clist = gui_clist_new( gb_rpl_list_label, gb_rpl_list_titles);
	               		gtk_signal_connect( GTK_OBJECT( gb_gui_clist->window), "destroy", 
	               			( GtkSignalFunc) destroy_widget_cb, &gb_gui_clist);
	               		gui_clist_start( gb_gui_clist);
	               		break;
	               case RPL_LIST:
	               		if ( gb_gui_clist)
	               			gui_clist_append( gb_gui_clist, g_list_last(param->arg_list));
	               		break;
	               case RPL_LISTEND:
	               		if ( gb_gui_clist)
	               			gui_clist_end( gb_gui_clist);
	               		break;
	               case RPL_TOPIC:
	               		if ( channel)
	               			channel_set_topic( channel, last);
	               		break;
	               case RPL_NAMEREPLY:
	               		if ( param->arg_list && param->arg_list->next) {
	               			channel = channel_find( param->connection->channels, param->arg_list->next->data);
	               			if ( !channel)
	               				channel = param->connection->default_channel;
	               			else if ( !FLAG_IS_SET(channel->flags, CHANNEL_JOINED))
	               				channel_add_users( channel, last);
	               		}	               			
	               		break;
	               case RPL_ENDOFNAMES:
	               		if ( channel && !FLAG_IS_SET(channel->flags, CHANNEL_JOINED) )
	               			SET_FLAG( channel->flags, CHANNEL_JOINED);
	               		break;
	               case ERR_INVITEONLYCHAN:
	               case ERR_BANNEDFROMCHAN:
	               case ERR_BADCHANNELKEY:
	               case ERR_CHANNELISFULL:
	               		if ( channel) {
	               			connection_remove_channel( param->connection, channel);
	               			channel = param->connection->default_channel;
	               		}
	               		break;
	               default:
	               		break;
	       }
	       
	} else if ( param->message_type == USER_COMMAND) {
	       
	       /* Echo privmsgs and notices */
	       switch ( param->token) {
	               case CTCP_ACTION_Q:
	               case IRC_PRIVMSG:
	               		if ( !channel)
	               			channel = connection_new_channel( param->connection, target);
	               		break;
	               case IRC_NOTICE:
	               		break;
	               case IRC_JOIN:
	               		if ( is_channel) 
	               			channel = connection_new_channel( param->connection, target);
	               		show_msg = 0;
	               		break;
	               case IRC_NICK:
	               		if ( !param->connection->data.irc->connected) {
	               			g_free( param->connection->data.irc->nick);
	               			param->connection->data.irc->nick = g_strdup( target);
	               			if ( channel)
	               				gui_window_update_status( channel->window);
	               		}
	               		show_msg = 0;
	               		break;
	               case CTCP_DCC_CHAT_A:
	               case CTCP_DCC_SEND_A:
	               		break;
	               default:
	               		show_msg = 0;
	               		break;
	      }
	      	      
	} else g_warning( "Unknown message type !!!!");	
	
	if ( show_msg) {
		if ( channel) 
			channel_append_message( channel, param);
		else 
			channel_append_message( param->connection->default_channel, param);
	}
	return TRUE;
}

/* parse_server_message : parse the server message : extract the origin of the message	
	( nickname, user and user's address), extract command's arguments		
	and then call parse_server_parameters						
*/
int
parse_server_message( Connection *connection, char *server_command) {
        IRCMessage param;
        
        param.connection   = connection;
        param.message_type = SERVER_COMMAND;
        param.arg_list	   = NULL;
        param.arg_trailing = -1;
        param.def	   = NULL;
        
        if ( *server_command == ':') {
                server_command++;
        	get_next_item( &server_command, &param.address, ' ');
        	get_next_item( &param.address, &param.nick, '!');
        	get_next_item( &param.address, &param.user, '@');
        } else {
                param.nick = NULL;
                param.user = NULL;
                param.address = NULL;
        }
	
	if ( get_server_message_arguments( &param, server_command))
		return 0;
		
	parse_server_parameters( &param);
	if ( param.arg_list)
		g_list_free( param.arg_list);
	return 0;
}	

/* parse_user_parameters : format the command, and eventualy quote it 			
	then send it to the server and finaly pass it to parse_server_parameters 	
	which will echo some commands like PRIVMSG, NOTICE ...				
*/
int
parse_user_parameters( Channel *channel, IRCMessage *param) {
        char buff[512];
        GList *arg_item;
        GList *arg_trailing;
        char *arg_value;
        char *tmp;
        int  i, quote;
        
	g_return_val_if_fail( param != NULL, 0);
	
        if ( param->token < IRC_END) { /* It's a message to the server */
        	if ( channel && 
		     (!param->connection || param->connection->data.irc->connected == FALSE) && 
		     param->token != IRC_NICK) {
        	       gui_window_display_message( channel->window,
        	       				   "Can't send message to the server. You're not connected\n",
        	       				   PROPERTY_ERROR);
        	       return 0;
        	}
        	
        	tmp = buff;
        	quote = param->token > CTCP_BEGIN;
        	arg_trailing = param->arg_trailing == -1 ? NULL : g_list_nth( param->arg_list, param->arg_trailing);
        	
        	/* keep 3 characters to hold the "\r\n\0" sequence */
        	i = 509;
        	
        	/* Write the command to the buffer */
        	arg_value = param->def.ucmd->token;
        	while ( *arg_value) { 
        		*tmp++ = *arg_value++; 
        		i--;
        	}
        	
        	/* Write arguments to the buffer */
        	arg_item = param->arg_list;
        	while ( arg_item != arg_trailing) {
        	       arg_value = (char*) arg_item->data;
        	       arg_item = arg_item->next;
        	       		 
        	       if ( i) { 
        	       		*tmp++ = ' ';
        	       		i--;
        	       }
        	       while ( *arg_value && i) { 
        	       		*tmp++ = *arg_value++; 
        	       		i--; 
        	       }
        	}
        	
        	if ( arg_trailing && (i -= 2)) {        	       
        	       *tmp++ = ' ';
        	       *tmp++ = ':';
        	       param->arg_list = g_list_append( param->arg_list, tmp);
        	       if ( quote) {
        	       		*tmp++ = '\001';
        	       		i--;
        	       }
        	       arg_item = arg_trailing;
        	       while ( arg_item && arg_item->next) {
        	       		arg_value = arg_item->data;
        	       		arg_item  = arg_item->next;
        	       		param->arg_list = g_list_remove_link( param->arg_list, arg_item->prev);
        	       		while( *arg_value && i) { 
        	       			*tmp++ = *arg_value++; 
        	       			i--; 
        	       		};
        	       		if ( i) { 
        	       			*tmp++ = ' ';
        	       			i--;
        	       		}
        	       }
        	       i--;tmp--;
        	       if ( quote) {
        	       		*tmp++ = '\001';
        	       		i--;
        	       }
        	}
        	
        	*tmp++ = '\r';
        	*tmp++ = '\n';
        	*tmp   = '\0';
        	
        	/* send the command to the server */
        	irc_send( param->connection, buff, strlen(buff));
        	
        	tmp -= 2; /* remove the newlines ("\r\n") */
        	*tmp = '\0';
      
		/* and finaly echo the message */
     	
		parse_server_parameters( param);
		
	} else { /* it's an "internal" command */
        	switch ( param->token) {
        	        case CMD_LEVEL:
        	        	cmd_level( param, channel);
        	        	break;
        	        case CMD_SERVER:
        	        	cmd_server( channel, g_list_nth_data( param->arg_list, 0), g_list_nth_data( param->arg_list, 1));
        	        	break;
        		case CMD_ALIAS:
        			if ( param->arg_list) 
       					new_alias( channel, param);
       				else if ( channel)
       					show_aliases( channel);
        			break;
        		case CMD_POPUP_WINDOW:
        			if ( channel && param->connection->default_channel &&
        			     channel->window == param->connection->default_channel->window)
        				channel_popup_window( channel);
        			break;
        		case CMD_POPDOWN_WINDOW:
        			if ( channel && param->connection->default_channel && 
        			     channel->window != param->connection->default_channel->window)
        				channel_popdown_window( param->connection->default_channel, channel);
        			break;
        		case CMD_LEAVE:
        			cmd_leave( param->connection, channel);
        			break;
        		case CMD_SET:
        			cmd_set_variable( param, channel);
        			break;
        		case CMD_LOG:
        			cmd_log( param, channel);
        			break;
        		default:
        			break;
        	}
        	
        	
       }	       
              		
       return 0;
}       

/* parse_user_command : parse the user command: find the message "definition" 	
	( its entry in the "gb_ucmd_messages" table), expands aliases		
	and then call parse_user_parameters	
*/
int
parse_user_command( Channel *channel, char *user_command) {
        IRCMessage param;
        char	   *command;
        char	   *message;
        
        param.connection   = channel ? channel->connection : NULL;
	param.message_type = USER_COMMAND;
	param.def	   = NULL;
	param.nick 	   = NULL;
	param.user 	   = NULL;
	param.address 	   = NULL;
       	while (1) {		
               	get_next_item( &user_command, &message, ';');
               	
               	if ( message == NULL || *message == '\0')
               		break;

		param.arg_list = NULL;
		param.arg_trailing = -1;
			
		get_next_item( &message, &command, ' ');
	   	/* try to tokenize the first string by looking in the command table 
 	 	   if this fails consider the command as an alias and expand it 
 	 	*/
      		if ( get_message_def( &param, command)) {
       		        char *alias;
       			if ( (alias = expand_alias( channel, command, message))) {
       				parse_user_message( channel, alias);
       				g_free( alias);
       				continue;
       			} else {
			    sprintf( gb_tmp_buff, "%s is unknwon command to me\n", command);
			    if ( channel) 
				gui_window_display_message( channel->window, gb_tmp_buff, PROPERTY_ERROR);
			    else
				g_warning( gb_tmp_buff);
       			      return 0;
       		       }
       		}
       		if ( param.token == CMD_ALIAS) {
       		      if ( *user_command)
       		      		message[strlen(message)] = ';';
       		      if ( get_user_message_arguments( &param, message) == 0)
       		      		return parse_user_parameters( channel, &param);
       		      return 0;
       		} else if ( param.token == CTCP_BEGIN) { 
       			char *target;
       			Connection *dcc_con;
       			
       			/* get the nick to which this query will be send */  
       	 		get_next_item( &message, &target, ' ');
			param.arg_list = g_list_append( NULL, target);        	      
			 
			param.arg_trailing = g_list_length( param.arg_list);
			param.arg_list = g_list_append( param.arg_list, message);
			param.def.ucmd = &gb_ucmd_messages[ IRC_PRIVMSG];
			get_ctcp_token_num( &param, message);
			
			switch (param.token) {
			     case CTCP_PING_Q:
			    		sprintf( gb_small_buff, "PING %d", (int) time(NULL));
			    		param.arg_list = g_list_append( param.arg_list, gb_small_buff);
			    		break;
			     case CTCP_DCC_CHAT_A:
			     case CTCP_DCC_SEND_A:
			     		if ( target && (dcc_con = dcc_request_find( param.token, target))) {
			     			dcc_request_accept( dcc_con);
			     			continue;
			     		} else
			     			dcc_answer_new( &param);
			     		break;
			     default:
			     		break;
			}
       		} else if ( get_user_message_arguments( &param, message)) {
       		        if ( param.arg_list);
       		        	g_list_free( param.arg_list);
	       		continue;
	       	}
       		parse_user_parameters( channel, &param);     
	       	
	       	if ( param.arg_list)
	       		g_list_free( param.arg_list);
       	}
        return 0;
}

/* parse_user_message : check to see if the message is escaped. 
	if so it's a command else it's a private message to the current chanel 
*/
int 
parse_user_message( Channel *channel, char *user_message) {
	if ( *user_message == VAR_CHAR( VAR_ESCAPE_CHAR)) {
		parse_user_command( channel, ++user_message);
        } else if ( channel && channel->connection && FLAG_IS_SET( channel->connection->flags, CON_IS_DCC)) {
                /* Handle dcc chat */
                char *msg;
                
                msg = g_malloc( strlen( user_message)+4);
                sprintf( msg, "> %s\n", user_message);
                irc_send( channel->connection, msg+2, strlen(msg+2));
                gui_window_display_message( channel->window, msg, PROPERTY_PRIVMSG);
                g_free( msg);
	} else {
        	IRCMessage param;
        	
                if ( !channel || channel == channel->connection->default_channel) {
                        return 0;
                }
                param.connection   = channel->connection;
		param.message_type = USER_COMMAND;
                param.def = &gb_ucmd_messages[IRC_PRIVMSG];
                param.token = param.def.ucmd->token_num;
                param.arg_list = g_list_append( NULL, channel->name);
                param.arg_trailing = g_list_length( param.arg_list);
                param.arg_list     = g_list_append( param.arg_list, user_message);
		param.nick 	   = NULL;
		param.user 	   = NULL;
		param.address 	   = NULL;
        	parse_user_parameters( channel, &param);
        	g_list_free( param.arg_list);
        }
                
        return TRUE;
} 

/* Misc functions */

IRCMessage*
irc_message_copy( IRCMessage *param) {
        IRCMessage *param_copy;
	GList *item;
	char  *arg;
	
        param_copy = g_malloc( sizeof(IRCMessage));
        param_copy->connection	 = param->connection;
        param_copy->message_type = param->message_type;
        param_copy->token	 = param->token;
        param_copy->def		 = param->def;
	param_copy->arg_list	 = NULL;
	param_copy->arg_trailing = param->arg_trailing;
	param_copy->nick	 = g_strdup(param->nick);
	param_copy->user	 = g_strdup(param->user);
	param_copy->address	 = g_strdup(param->address);
	
	item = param->arg_list;
	while (item) {
                arg = item->data ? g_strdup( (char*) item->data) : NULL;
       		item = item->next;
       		param_copy->arg_list = g_list_append( param_copy->arg_list, arg);
        }
        
        return param_copy;
}

void
irc_message_delete( IRCMessage *param) {
        GList	*item;
        
        if ( param->nick) g_free( param->nick);
        if ( param->user) g_free( param->user);
        if ( param->address) g_free( param->address);
        item = param->arg_list;
        while (item) {
                if ( item->data)
                	g_free( item->data);
                item = item->next;
        }
        if ( param->arg_list) 
        	g_list_free( param->arg_list);
        g_free( param);
}



