#ifndef __IRCDEFS_H__
#define __IRCDEFS_H__

#ifdef __cplusplus
extern "C"
{
#endif /* __cplusplus */

#include "gnirc.h"
#include "enums.h"

/* 
   the following arrays does not describe perfectly the IRC command syntax 		
   they are merely a way to give a general method to parse server & user messages 	
   So most of the time we'll only take care of exceptions to these rules		
   some special commands that can not be fully described by this method are for example 
   - MODE,PRIVMSG,NOTICE which can have a channel or a nick as "target"		
   - RPL_NAMEREPLY which does have as first parameter a channel name but a "=" instead    
   - some "numeric replies" merge text and parameters ( ERR_FILEERROR for example which
   	template is ":File error doing <file op> on <file>"
*/


/* I think it would be better to replace these arrays with strings 
   ( it would be easier to understand) 
   eg :
   replace {  ARG_NICK, 0} with for example "%n" or "n"
*/

unsigned short int arg_one_string[]		= {  ARG_STR, 0};
unsigned short int arg_one_nick[]		= {  ARG_NICK, 0};
unsigned short int arg_one_channel[]		= {  ARG_CHANNEL, 0};
unsigned short int arg_one_server[]		= {  ARG_SERVER, 0};
unsigned short int arg_one_user[]		= {  ARG_USER, 0};
unsigned short int arg_one_trailing[]		= {  ARG_TRAILING, 0};
unsigned short int arg_one_char[]		= {  ARG_CHAR, 0};
unsigned short int arg_one_int[]		= {  ARG_INT, 0};


unsigned short int arg_oper[]		= {  ARG_USER, ARG_STR, 0 };
unsigned short int arg_join[]		= {  ARG_CHANNEL_LIST, ARG_STR_LIST_COMMA, 0 };
unsigned short int arg_part[]		= {  ARG_CHANNEL_LIST, 0 };
unsigned short int arg_cmode[]		= {  ARG_CHANNEL, ARG_MODE_FLAG, ARG_STR, 0 };
unsigned short int arg_umode[] 		= {  ARG_NICK, ARG_MODE_FLAG, 0 };
unsigned short int arg_topic[] 		= {  ARG_CHANNEL, ARG_TOPIC, 0};
unsigned short int arg_names[] 		= {  ARG_CHANNEL_LIST, 0};
unsigned short int arg_list[]  		= {  ARG_CHANNEL_LIST, ARG_SERVER, 0};
unsigned short int arg_invite[]		= {  ARG_NICK, ARG_CHANNEL, 0};
unsigned short int arg_kick[]  		= {  ARG_CHANNEL, ARG_NICK, ARG_TRAILING, 0};
unsigned short int arg_stats[] 		= {  ARG_CHAR, ARG_SERVER, 0};
unsigned short int arg_links[] 		= {  ARG_SERVER, ARG_STR, 0};
unsigned short int arg_connect[]	= {  ARG_SERVER, ARG_INT, ARG_SERVER, 0};
unsigned short int arg_privmsg[]	= {  ARG_CHANNEL_LIST, ARG_TRAILING, 0};
unsigned short int arg_notice[]		= {  ARG_NICK, ARG_TRAILING, 0};
unsigned short int arg_who[]		= {  ARG_STR, ARG_CHAR, 0};
unsigned short int arg_whois[]		= {  ARG_SERVER, ARG_NICK_LIST, 0};
unsigned short int arg_whowas[]		= {  ARG_NICK, ARG_INT, ARG_STR, 0};
unsigned short int arg_kill[]		= {  ARG_SERVER, ARG_TRAILING, 0};
unsigned short int arg_ping[]		= {  ARG_SERVER, ARG_SERVER, 0};
unsigned short int arg_away[]		= {  ARG_TRAILING, 0};
unsigned short int arg_summon[]		= {  ARG_USER, ARG_SERVER, 0};
unsigned short int arg_users[]		= {  ARG_SERVER, 0};
unsigned short int arg_wallops[]	= {  ARG_TRAILING, 0};
unsigned short int arg_userhost[]	= {  ARG_NICK, ARG_NICK, ARG_NICK, ARG_NICK, ARG_NICK, 0};
unsigned short int arg_ison[]		= {  ARG_NICK_LIST, 0};
unsigned short int arg_dcc_chat[]	= {  ARG_NICK, 0};
unsigned short int arg_dcc_send[]	= {  ARG_NICK, ARG_STR, 0};

unsigned short int arg_alias[]		= {  ARG_STR, ARG_TRAILING, 0}; /* syntax alias <alias name> <expanded alias>  */
unsigned short int arg_log[]		= {  ARG_STR, ARG_STR, 0}; 	/* syntax log [[<channel>|<nick>] [<logfile>]] */
unsigned short int arg_server[]		= {  ARG_SERVER, ARG_STR, 0};

unsigned short int arg_err_usernotinchannel[]	= {  ARG_NICK, ARG_CHANNEL, 0};
unsigned short int arg_err_useronchannel[]	= {  ARG_USER, ARG_NICK, 0};
unsigned short int arg_err_unknownmode[]	= {  ARG_CHAR, 0};

unsigned short int arg_rpl_whoisuser[]		= {  ARG_NICK, ARG_USER, ARG_SERVER, ARG_REALNAME, 0};
unsigned short int arg_rpl_whoisserver[]	= {  ARG_NICK, ARG_SERVER, ARG_TRAILING, 0};
unsigned short int arg_rpl_whoisidle[]		= {  ARG_NICK, ARG_TIME, 0};
unsigned short int arg_rpl_whowasuser[]		= {  ARG_NICK, ARG_USER, ARG_SERVER, ARG_REALNAME, 0};
unsigned short int arg_rpl_list[]		= {  ARG_CHANNEL, ARG_CHAR, ARG_TOPIC, 0};
unsigned short int arg_rpl_channelmodeis[]	= {  ARG_CHANNEL, ARG_MODE_FLAG, ARG_STR, 0};
unsigned short int arg_rpl_topic[]		= {  ARG_CHANNEL, ARG_TOPIC, 0};
unsigned short int arg_rpl_inviting[]		= {  ARG_CHANNEL, ARG_NICK, 0};
unsigned short int arg_rpl_version[]		= {  ARG_STR, ARG_SERVER, ARG_TRAILING, 0};
unsigned short int arg_rpl_whoreply[]		= {  ARG_CHANNEL, ARG_USER, ARG_SERVER, ARG_SERVER, ARG_NICK, 
							ARG_CHAR, ARG_CHAR, ARG_CHAR, ARG_TRAILING, 0};
unsigned short int arg_rpl_namereply[]		= {  ARG_CHANNEL, ARG_NICK_LIST, 0};
unsigned short int arg_rpl_links[]		= {  ARG_STR, ARG_SERVER, 0};
unsigned short int arg_rpl_banlist[]		= {  ARG_CHANNEL, ARG_STR, 0};
unsigned short int arg_rpl_time[]		= {  ARG_SERVER, ARG_TIME_STR, 0};

/* 
    the last parameter is subject to the following expansions : 
	$n = nickname					       
	$u = username					      
	$h = address					       
	$0 = first parameter ....			       
	$+ = last parameter				       
	$* = all parameters ( == $0 $1 $2 ... $+)	       
*/

/* 
    Note : the first parameter of numeric replies seems to be  (is always ?)   
		the nickname that made the query. So i've decided to remove   
		it from the argument list. As a consequence the parameter "$0"
		is in reality the second parameter ($1) of the reply		      
*/ 

UCmd gb_ucmd_messages[] = 
{
        { "NICK"	, IRC_NICK	, 1, arg_one_nick	, "$n is now known as $0"					},
        { "OPER"	, IRC_OPER	, 2, arg_oper		, NULL								},
        { "QUIT"	, IRC_QUIT	, 0, arg_one_trailing	, "$n ($u@$h) has quit irc. ($+)"				},
        { "JOIN"	, IRC_JOIN	, 1, arg_join		, "$n ($u@$h) has joined channel $0"				},
        { "PART"	, IRC_PART	, 1, arg_part		, "$n has quit channel $0"					},
        { "MODE"	, IRC_MODE	, 2, arg_cmode		, "$n set mode of $0 to $1 $2"					},
        { "TOPIC"	, IRC_TOPIC	, 1, arg_topic		, "$n sets the topic of channel $0 to $+"			},
        { "NAMES"	, IRC_NAMES	, 1, arg_names		, NULL								},
        { "LIST"	, IRC_LIST	, 0, arg_list		, NULL								},
        { "INVITE"	, IRC_INVITE	, 2, arg_invite		, "$n invite you on channel $0"					},
        { "KICK"	, IRC_KICK	, 2, arg_kick		, "$1 has been kicked of $0 by $n ($+)" 			},
        { "VERSION"	, IRC_VERSION	, 0, arg_one_server	, NULL								},
        { "STATS"	, IRC_STATS	, 0, arg_one_server	, NULL								},
        { "LINKS"	, IRC_LINKS	, 0, arg_links		, NULL								},
        { "TIME"	, IRC_TIME	, 0, arg_one_server	, NULL								},
        { "CONNECT"	, IRC_CONNECT	, 1, arg_connect	, NULL								},
        { "TRACE"	, IRC_TRACE	, 0, arg_one_server	, NULL								},
        { "ADMIN"	, IRC_ADMIN	, 0, arg_one_server	, NULL								},
        { "INFO"	, IRC_INFO	, 0, arg_one_server	, NULL								},
        { "PRIVMSG"	, IRC_PRIVMSG	, 2, arg_privmsg	, NULL								},
        { "NOTICE"	, IRC_NOTICE	, 2, arg_notice		, NULL								},
        { "WHO"		, IRC_WHO	, 0, arg_who		, NULL								},
        { "WHOIS"	, IRC_WHOIS	, 1, arg_whois		, NULL								},
        { "WHOWAS"	, IRC_WHOWAS	, 1, arg_whowas		, NULL								},
        { "KILL"	, IRC_KILL	, 2, arg_kill		, NULL								},
        { "PING"	, IRC_PING	, 1, arg_ping		, NULL								},
        { "PONG"	, IRC_PONG	, 1, arg_one_trailing	, NULL								},
        { "ERROR"	, IRC_ERROR	, 1, arg_one_trailing	, "ERROR : $+"							},
        { "AWAY"	, IRC_AWAY	, 0, arg_away		, "$n is away ($+)"						},
        { "REHASH"	, IRC_REHASH	, 0, NULL		, NULL								},
        { "RESTART"	, IRC_RESTART	, 0, NULL		, NULL								},
        { "SUMMON"	, IRC_SUMMON	, 1, arg_summon		, NULL								},
        { "USERS"	, IRC_USERS	, 0, arg_one_server	, NULL								},
        { "WALLOPS"	, IRC_WALLOPS	, 1, arg_one_trailing	, NULL								},
        { "USERHOST"	, IRC_USERHOST	, 1, arg_userhost	, NULL								},
        { "ISON"	, IRC_ISON	, 1, arg_ison		, NULL								},
        { "CTCP"	, CTCP_BEGIN	, 2, arg_privmsg	, NULL								},
        { "CTCP FINGER QUERY"		, CTCP_FINGER_Q		, 2, arg_privmsg	, "FINGER request from $n to $0"       	},
        { "CTCP USERINFO QUERY"		, CTCP_USERINFO_Q	, 2, arg_privmsg	, "USERINFO request from $n to $0"     	},
        { "CTCP CLIENTINFO QUERY"	, CTCP_CLIENTINFO_Q	, 2, arg_privmsg	, "CLIENTINFO request from $n to $0"   	},
        { "CTCP ERRMSG QUERY"		, CTCP_ERRMSG_Q		, 2, arg_privmsg	, "ERRMSG received from $n :$+"		},
        { "CTCP PING QUERY"		, CTCP_PING_Q		, 2, arg_privmsg	, "PING request from $n to $0"		},
        { "CTCP TIME QUERY"		, CTCP_TIME_Q		, 2, arg_privmsg	, "TIME request from  $n to $0"		},
        { "CTCP ACTION QUERY"		, CTCP_ACTION_Q		, 2, arg_privmsg	, NULL					},
        { "CTCP VERSION QUERY"		, CTCP_VERSION_Q	, 2, arg_privmsg	, "VERSION request from $n ($u@$h) to $0"},
        { "CTCP SOURCE QUERY"		, CTCP_SOURCE_Q		, 2, arg_privmsg	, "SOURCE request from $n to $0"       	},
        { "CTCP UNKNOWN QUERY"		, CTCP_UNKNOWN_Q	, 0, NULL		, "UNKNOWN request from $n ($u@$h) :$+"	},
        { "CTCP FINGER ANSWER"		, CTCP_FINGER_A		, 2, arg_notice		, "FINGER answer from $n : $+"		},
        { "CTCP USERINFO ANSWER"	, CTCP_USERINFO_A	, 2, arg_notice		, "USERINFO answer from $n : $+"	},
        { "CTCP CLIENTINFO ANSWER"	, CTCP_CLIENTINFO_A	, 2, arg_notice		, "CLIENTINFO answer from $n : $+"	},
        { "CTCP ERRMSG ANSWER"		, CTCP_ERRMSG_A		, 2, arg_notice		, "ERRMSG answer from $n :$+"		},
        { "CTCP PING ANSWER"		, CTCP_PING_A		, 2, arg_notice		, "PING answer from $n : $+ second(s)"	},
        { "CTCP TIME ANSWER"		, CTCP_TIME_A		, 2, arg_notice		, "TIME answer from  $n : $+"		},
        { "CTCP ACTION ANSWER"		, CTCP_ACTION_A		, 2, arg_notice		, NULL					},
        { "CTCP VERSION ANSWER"		, CTCP_VERSION_A	, 2, arg_notice		, "VERSION answer from $n : $+"		},
        { "CTCP SOURCE ANSWER"		, CTCP_SOURCE_A		, 2, arg_notice		, "SOURCE answer from $n : $+"		},
        { "CTCP UNKNOWN ANSWER"		, CTCP_UNKNOWN_A	, 0, NULL		, "UNKNOWN answer from $n :$+"		},
        { "CTCP DCC CHAT QUERY"		, CTCP_DCC_CHAT_Q	, 1, arg_notice		, "DCC chat request from $n ($u@$h)"	},
        { "CTCP DCC SEND QUERY"		, CTCP_DCC_SEND_Q	, 2, arg_notice		, "DCC send request from $n ($u@$h)"},
        { "CTCP DCC CHAT ANSWER"	, CTCP_DCC_CHAT_A	, 1, arg_dcc_chat	, "Sending dcc chat request to $0"	},
        { "CTCP DCC SEND ANSWER"	, CTCP_DCC_SEND_A	, 2, arg_dcc_send	, "Sending dcc send request to $0"	},
        
        /* Program specific commands */
        { "LEVEL"			, CMD_LEVEL		, 0, arg_one_trailing	, NULL					},
        { "SERVER"			, CMD_SERVER		, 1, arg_server		, "Connecting to $n"			},
        { "ALIAS"			, CMD_ALIAS		, 0, arg_alias		, "$n is now an alias to $u"		},
        { "POPUP_WINDOW"		, CMD_POPUP_WINDOW	, 0, NULL		, NULL					},
        { "POPDOWN_WINDOW"		, CMD_POPDOWN_WINDOW	, 0, NULL		, NULL					},
        { "SET"				, CMD_SET		, 0, arg_alias		, "$n value is now $u"			},
        { "LOG"				, CMD_LOG		, 0, arg_log		, NULL					},
        { "WARANTY"			, CMD_WARANTY		, 0, NULL		, NULL					},
        { "LEAVE"			, CMD_LEAVE		, 0, NULL		, NULL					},
        { NULL		, 0		, 0, NULL		, NULL								}
};

URply	gb_urply_messages[] =
{
        { RPL_UNKNOWN		, NULL			, "$*"									},
        { RPL_USERHOST		, NULL			, "$*"									},
        { RPL_ISON		, NULL			, "Currently online : $+"						},
        { RPL_AWAY		, arg_one_nick		, "$0 is away ($+)"							},
        { RPL_UNAWAY		, NULL			, "You are no longer marked as being away"				},
        { RPL_NOWAWAY		, NULL			, "You have been marked as being away"					},
        { RPL_WHOISUSER		, arg_rpl_whoisuser	, "$0 is $1@$2 ($+)"							},
        { RPL_WHOISSERVER	, arg_rpl_whoisserver	, "$0 uses server $1 ($+)"						},
        { RPL_WHOISOPERATOR	, NULL			, "$0 s an IRC operator"						},
        { RPL_WHOISIDLE		, arg_rpl_whoisidle	, "$0 has been idled $1 seconds"					},
        { RPL_ENDOFWHOIS	, NULL			, NULL									},
        { RPL_WHOISCHANNELS	, NULL			, "$0 is on channel(s) : $+"						},
        { RPL_WHOWASUSER	, arg_rpl_whowasuser	, "$0 was $1@$2 ($+)"							},
        { RPL_LISTSTART		, NULL			, NULL									},
        { RPL_LIST		, arg_rpl_list		, NULL									},
        { RPL_LISTEND		, NULL			, NULL									},
        { RPL_CHANNELMODEIS	, arg_rpl_channelmodeis, "mode of $0 is $1 $2"							},
        { RPL_NOTOPIC		, NULL			, "channel $0 has no topic set"						},
        { RPL_TOPIC		, arg_rpl_topic		, "topic of $0 is $+"							},
        { RPL_INVITING		, arg_rpl_inviting	, "$1 has been invited on $0"						},
        { RPL_SUMMONING		, arg_one_user		, "Summoning user $0 to IRC"						},
        { RPL_VERSION		, arg_rpl_version	, "server version of $1 is $0 ($+)"					},
        { RPL_WHOREPLY		, arg_rpl_whoreply	, "$*"									},
        { RPL_ENDOFWHO		, NULL			, NULL									},
        { RPL_NAMEREPLY		, arg_rpl_namereply	, "users on $1 : $+"							},
        { RPL_ENDOFNAMES	, NULL			, NULL									},
        { RPL_LINKS		, arg_rpl_links		, "$*"									},
        { RPL_ENDOFLINKS	, arg_one_string	, NULL									},
        { RPL_BANLIST		, arg_rpl_banlist	, "$*"									},
        { RPL_ENDOFBANLIST	, arg_one_channel	, NULL									},
        { RPL_INFO		, arg_one_trailing	, "$*"									},
        { RPL_ENDOFINFO		, NULL			, NULL									},
        { RPL_MOTDSTART		, NULL			, NULL									},
        { RPL_MOTD		, arg_one_trailing	, "$+"									},
        { RPL_ENDOFMOTD		, NULL			, NULL									},
        { RPL_YOUREOPER		, NULL			, "You are now an IRC operator"						},
        { RPL_REHASHING		, arg_one_string	, "Rehashing $0"							},
        { RPL_TIME		, arg_rpl_time		, "Time on server $0 is $+"						},
        { RPL_USERSSTART	, NULL			, NULL									},
        { RPL_USERS		, NULL			, "$*"									},
        { RPL_ENDOFUSERS	, NULL			, NULL									},
        { RPL_NOUSERS		, NULL			, "Nobody logged in"							},
        { RPL_TRACELINK		, NULL			, "$*"									}, 
        
        /* i need to understand what all of that means first ;-)								*/
        { RPL_TRACECONNECTING	, NULL			, "$*"									},
        { RPL_TRACEHANDSHAKE	, NULL			, "$*"									},
        { RPL_TRACEUNKNOWN	, NULL			, "$*"									},
        { RPL_TRACEOPERATOR	, NULL			, "$*"									},
        { RPL_TRACEUSER		, NULL			, "$*"									},
        { RPL_TRACESERVER	, NULL			, "$*"									},
        { RPL_TRACENEWTYPE	, NULL			, "$*"									},
        { RPL_TRACELOG		, NULL			, "$*"									},
        { RPL_STATSLINKINFO	, NULL			, "$*"									},
        { RPL_STATSCOMMANDS	, NULL			, "$*"									},
        { RPL_STATSCLINE	, NULL			, "$*"									},
        { RPL_STATSNLINE	, NULL			, "$*"									},
        { RPL_STATSILINE	, NULL			, "$*"									},
        { RPL_STATSKLINE	, NULL			, "$*"									},
        { RPL_STATSYLINE	, NULL			, "$*"									},
        { RPL_ENDOFSTATS	, NULL			, NULL									},
        { RPL_STATSLLINE	, NULL			, "$*"									},
        { RPL_STATSUPTIME	, NULL			, "$*"									},
        { RPL_STATSOLINE	, NULL			, "$*"									},
        { RPL_STATSHLINE	, NULL			, "$*"									},
        { RPL_UMODEIS		, NULL			, "$*"									},
        
        { RPL_LUSERCLIENT	, arg_one_trailing	, "$+"									},
        { RPL_LUSEROP		, arg_one_int		, "$0 operator(s) online"						},
        { RPL_LUSERUNKNOWN	, arg_one_int		, "$0 unknown connection(s)"						},
        { RPL_LUSERCHANNELS	, arg_one_int		, "$0 channel(s) has been formed"					},
        { RPL_LUSERME		, arg_one_trailing	, "$+"									},
        { RPL_ADMINME		, arg_one_server	, "$*"									},
        { RPL_ADMINLOC1		, arg_one_trailing	, "$*"									},
        { RPL_ADMINLOC2		, arg_one_trailing	, "$*"									},
        { RPL_ADMINEMAIL	, arg_one_trailing	, "$*"									},
        
        
	{ ERR_NOSUCHNICK	, arg_one_nick		, "user $0 doesn't exist"						},
	{ ERR_NOSUCHSERVER	, arg_one_server	, "server $0 doesn't exist"						},
	{ ERR_NOSUCHCHANNEL	, arg_one_channel	, "there's no  channel named $0"					},
	{ ERR_CANNOTSENDTOCHAN  , arg_one_channel	, "Can't send to channel $0"						},
	{ ERR_TOOMANYCHANNELS	, arg_one_channel	, "Can't join $0, you have joined to many channels"			},
	{ ERR_WASNOSUCHNICK	, arg_one_nick		, "there was no user named $0"						},
	{ ERR_TOOMANYTARGETS	, arg_one_string	, "$0 :Duplicate recipients. No message delivered"			},
	{ ERR_NOORIGIN		, NULL			, "No origin specified"							},
	{ ERR_NORECIPIENT	, NULL			, NULL									},  
	{ ERR_NOTEXTTOSEND	, NULL			, "No text to send"							},
	{ ERR_NOTOPLEVEL	, arg_one_string	, "$0 :No toplevel domain specified"					},
	{ ERR_WILDTOPLEVEL	, arg_one_string	, "$0 :Wildcard in toplevel domain"					},
	{ ERR_UNKNOWNCOMMAND	, arg_one_string	, "$0 :Unknown command"							},
	{ ERR_NOMOTD		, NULL			, "MOTD File is missing"						},
	{ ERR_NOADMININFO	, arg_one_server	, "$0 :No administrative info available"				},
	{ ERR_FILEERROR		, NULL			, "$*"								},  
	{ ERR_NONICKNAMEGIVEN	, NULL			, "No nickname given"							},
	{ ERR_ERRONEUSNICKNAME	, arg_one_nick		, "$0 :Erroneus nickname"						},
	{ ERR_NICKNAMEINUSE	, arg_one_nick		, "$0 :Nickname is already in use"					},
	{ ERR_NICKCOLLISION	, arg_one_nick		, "$0 :Nickname collision KILL"						},
	{ ERR_USERNOTINCHANNEL	, arg_err_usernotinchannel, "$0 $1 :They aren't on that channel"				},
	{ ERR_NOTONCHANNEL	, arg_one_channel	, "$0 :You're not on that channel"					},
	{ ERR_USERONCHANNEL	, arg_err_useronchannel	, "user $1 is already on channel $0"					},
	{ ERR_NOLOGIN		, arg_one_user		, "$0 :User not logged in"						},
	{ ERR_SUMMONDISABLED	, NULL			, "SUMMON has been disabled"						},
	{ ERR_USERSDISABLED	, NULL			, "USERS has been disabled"						},
	{ ERR_NOTREGISTERED	, NULL			, "You have not registered"						},
	{ ERR_NEEDMOREPARAMS	, arg_one_string	, "$0 :Not enough parameters"						},
	{ ERR_ALREADYREGISTRED	, NULL			, "You may not reregister"						},
	{ ERR_NOPERMFORHOST	, NULL			, "Your host isn't among the privileged"				},
	{ ERR_PASSWDMISMATCH	, NULL			, "Password incorrect"							},
	{ ERR_YOUREBANNEDCREEP	, NULL			, "You are banned from this server"					},
	{ ERR_KEYSET		, arg_one_channel	, "$0 :Channel key already set"						},
	{ ERR_CHANNELISFULL	, arg_one_channel	, "Cannot join channel $0 ( there's too much users)"			},
	{ ERR_UNKNOWNMODE	, arg_one_char		, "$0 :is unknown mode char to me"					},
	{ ERR_INVITEONLYCHAN	, arg_one_channel	, "Cannot join channel $0 ( you haven't been invited to join)"		},
	{ ERR_BANNEDFROMCHAN	, arg_one_channel	, "Cannot join channel $0 ( you're banned from that channel)"		},
	{ ERR_BADCHANNELKEY	, arg_one_channel	, "Cannot join channel $0 ( you've got to supply the right password)"	},
	{ ERR_NOPRIVILEGES	, NULL			, "Permission Denied- You're not an IRC operator"			},
	{ ERR_CHANOPRIVSNEEDED	, arg_one_channel	, "You're not channel operator on $0"					},
	{ ERR_CANTKILLSERVER	, NULL			, "You cant kill a server!"						},
	{ ERR_NOOPERHOST	, NULL			, "No O-lines for your host"						},
	{ ERR_UMODEUNKNOWNFLAG	, NULL			, "Unknown MODE flag"							},
	{ ERR_USERSDONTMATCH	, NULL			, "Cant change mode for other users"					},
	{ 0			, NULL			, NULL									}
	
};
        
UVariable
gb_uvariables[] = 
{
        { "config_dir"		, ARG_STR	, NULL  },
        { "nickname"		, ARG_STR	, NULL	},
        { "realname"		, ARG_TRAILING	, NULL	},
        { "user_mode"		, ARG_INT	, NULL  },
        { "userinfo"		, ARG_TRAILING	, NULL	},
        { "email"		, ARG_STR	, NULL	},
        { "default_quit_reason" , ARG_STR	, NULL	},
        { "default_server"	, ARG_STR	, NULL	},
        { "default_port"	, ARG_UINT	, NULL	},
        { "default_user_mode"	, ARG_STR	, NULL	},
        { "input_timeout"	, ARG_UINT	, NULL	},
        { "window_width"	, ARG_UINT	, NULL	},
        { "window_height"	, ARG_UINT	, NULL	},
        { "show_user_list"	, ARG_INT	, NULL  },
        { "show_toolbar"	, ARG_INT	, NULL	},
        { "escape_character"	, ARG_CHAR	, NULL  },
        { "history_size"	, ARG_INT	, NULL	},
        { "logfiles_dir"	, ARG_STR	, NULL	},
        { "system_type"		, ARG_STR	, NULL  },
        { "system_version"	, ARG_STR	, NULL	},
        { "program_name"	, ARG_STR	, NULL	},
        { "program_version"	, ARG_STR	, NULL	},
        { "program_source_host"	, ARG_STR	, NULL	},
        { "program_source_dir"	, ARG_STR	, NULL	},
        { "program_source_files", ARG_TRAILING	, NULL	},
        { "program_pixmaps_path", ARG_STR	, NULL  },
        { "dcc_send_directory"	, ARG_STR	, NULL	},
        { "end of builtin vars"	, -1		, NULL	},
        { NULL			, 0		, NULL  }
};

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif
