/*
 *    gnirc - Another IRC client for X 
 *    Copyright (C) 1998  david Ramboz <dramboz@imaginet.fr>
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <netdb.h>
#include <fcntl.h>
#include <string.h>
#include <ctype.h>
#include <sys/time.h> 
#include <sys/types.h> 
#include <sys/socket.h>
#include <netinet/in.h>
#include <gtk/gtk.h>

#include "gnirc.h"
#include "enums.h"

extern  char		*gb_tmp_buff;
extern  UVariable 	gb_uvariables[];

GList	*gb_connections = NULL;

/* set after the first call to socket_listen ( in network byte order)*/
unsigned long int gb_address = 0;
        
/****************************/
/*** High level functions ***/
/*--------------------------*/

Connection*
server_connection_new( char *server, int port) {
        Connection *connection = NULL;
        
	if ( server == NULL || port < 1)
		return FALSE;
		        
        connection = connection_new();
        connection->input_func = irc_input;
        connection->data.irc = g_malloc( sizeof(IRCParam));
        connection->data.irc->connected = FALSE;
        connection->data.irc->nick = g_strdup(VAR_STR(VAR_NICKNAME));
        connection->data.irc->user_mode = 0;
        connection->address = g_strdup( server);
        connection->port = g_malloc( 8);
        sprintf( connection->port, "%d", port);
        
        if ( (connection->sock = socket_new()) && 
              socket_connect_by_name( connection->sock, server, port) &&
              connection_register( connection) ) {
        	connection->flags = CON_CONNECTED;
                irc_connect( connection, VAR_STR(VAR_NICKNAME), VAR_TRAILING(VAR_REALNAME), NULL);
             	return connection;
        }
        
        connection_free( connection);
        return FALSE;
}
             
int
server_connection_reuse( Connection *connection, char *server, int port) {
        int flag;
        g_return_val_if_fail( connection != NULL, FALSE);
        
        if ( FLAG_IS_SET( connection->flags, CON_CONNECTED) && connection->sock > 0)
        	close( connection->sock);
	if ( !FLAG_IS_SET( connection->flags, CON_IS_DCC))
		connection->data.irc->connected = FALSE;
		
        if ( (connection->sock = socket_new())) {
                if ( socket_connect_by_name( connection->sock, server, port)) {
			SET_FLAG( connection->flags, CON_CONNECTED);
			/* Initialize the input buffer */
               		connection->buffer.next_line = connection->buffer.buff;
               		connection->buffer.next_byte = connection->buffer.buff;
        		flag = fcntl(connection->sock, F_GETFL);
        		if ( fcntl(connection->sock, F_SETFL, flag | O_NONBLOCK ) == -1 )
        			return FALSE;
        		irc_connect( connection, VAR_STR(VAR_NICKNAME), VAR_TRAILING(VAR_REALNAME), NULL);
        		return TRUE;
        	}
        }
        
        return FALSE;
                        
}
       
        
            
/***************/
/*** sockets ***/
/*-------------*/

int
socket_new() {
        int sock;
        
	sock = socket( PF_INET, SOCK_STREAM, 0);
	if ( sock < 0) {
	       g_warning( "socket : %s", strerror( errno));
	       return FALSE;
	}
	return sock;
}

int
socket_listen( int sock, int port) {
	struct sockaddr_in name;
	int size;
	
	g_return_val_if_fail( sock > 0, FALSE);
	g_return_val_if_fail( port >= 0, FALSE);
	
	size = sizeof(name);
 	name.sin_family = AF_INET;
	name.sin_port  = port;
	name.sin_addr.s_addr = htonl (INADDR_ANY);
	
	if ( bind( sock, (struct sockaddr*) &name, size) == -1) {
	       g_warning( "bind : %s", strerror( errno));
	       return FALSE;
	}
	
	if ( listen( sock, 1) == -1) {
	       g_warning( "listen : %s", strerror( errno));
	       return FALSE;
	}
	
	if ( getsockname( sock, (struct sockaddr*) &name, &size) == -1) {
	       g_warning( "listen : %s", strerror( errno));
	       return FALSE;
	}
	gb_address = htonl(name.sin_addr.s_addr);
	
	return name.sin_port;
}

int
socket_connect( int sock, unsigned long int host, int port) {
	struct sockaddr_in name;
	
	g_return_val_if_fail( sock > 0, FALSE);
	
	name.sin_family = AF_INET;
	name.sin_port  = htons(port);
	name.sin_addr.s_addr = htonl(host);
	
	if ( connect( sock, (struct sockaddr *) &name, sizeof( name)) &&
	      errno != EINPROGRESS) {
	       g_warning( "connect : %s", strerror(errno));
	       return FALSE;
	}
	
	return TRUE;
}

int
socket_connect_by_name( int sock, char *host, int port) {     
	struct sockaddr_in name;
        struct hostent *hostinfo;
	
	g_return_val_if_fail( sock > 0, FALSE);
	g_return_val_if_fail( host != NULL, FALSE);
	g_return_val_if_fail( port > 0, FALSE);
	
	name.sin_family = AF_INET;
	name.sin_port  = htons(port);
		
        hostinfo = gethostbyname (host);
        if (hostinfo == NULL) {
           g_warning("%s : unknown host", host);
           return FALSE;
        }
        name.sin_addr = *(struct in_addr *) hostinfo->h_addr;
        
	if ( connect( sock, (struct sockaddr *) &name, sizeof( name)) &&
	      errno != EINPROGRESS) {
	       g_warning( "connect : %s", strerror(errno));
	       return FALSE;
	}
	
	return TRUE;
}
	
/**************************/
/***     Connections    ***/
/*------------------------*/

Connection*
connection_new() {
        Connection *connection;
        
        connection = g_malloc( sizeof( Connection));
        connection->sock = -1;
        connection->address = NULL;
        connection->port = NULL;
        connection->input_func = NULL;
        connection->buffer.next_line = connection->buffer.buff;
        connection->buffer.next_byte = connection->buffer.buff;
        *connection->buffer.buff = '\0';
        connection->channels = NULL;
        connection->default_channel = NULL;
        connection->data = NULL;
        connection->flags = 0;
        
        return connection;
}

void
connection_free( Connection *connection) {
	GList *channel_item;
	Channel *channel;
                
        g_return_if_fail( connection != NULL);
        
        if ( FLAG_IS_SET( connection->flags, CON_IS_DCC))
        	dcc_free( connection->data.dcc);
        else if ( FLAG_IS_SET( connection->flags, CON_CONNECTED)) {
	        if ( connection->data.irc->connected)
		    irc_disconnect( connection, VAR_STR( VAR_DEFAULT_QUIT_REASON));
                g_free( connection->data.irc->nick);
        	g_free( connection->data.irc);
        }
        	
        gb_connections = g_list_remove( gb_connections, connection);
        if ( connection->sock > 0)
        	close( connection->sock);
        	
	channel_item = connection->channels;
	while ( channel_item) {
	      channel = channel_item->data;
	      channel_item = channel_item->next;
	      
	      if ( FLAG_IS_SET( channel->flags, CHANNEL_MASTER))
		      channel->connection = NULL;
	      else
		      channel_free( channel);
	}
        if ( connection->channels)
                g_list_free( connection->channels);
        
        if ( connection->address)
        	g_free( connection->address);
        	
        if ( connection->port)
        	g_free( connection->port);
        	
	g_free( connection);
}

void
connection_free_all() {
        
        while ( gb_connections)
        	connection_free( (Connection*) gb_connections->data);
}

int
connection_register( Connection *connection) {
	int flag;
	
        g_return_val_if_fail( connection != NULL, FALSE);
        g_return_val_if_fail( connection->sock > 0, FALSE);
        
        /* */
        flag = fcntl(connection->sock, F_GETFL);
        if ( fcntl(connection->sock, F_SETFL, flag | O_NONBLOCK ) == -1 )
        	return FALSE;
        
        gb_connections = g_list_prepend( gb_connections, connection);
        return TRUE;
}

void
connection_unregister( Connection *connection) {
        g_return_if_fail( connection != NULL);
        g_return_if_fail( connection->sock > 0);
        
        gb_connections = g_list_remove( gb_connections, connection);
}

Channel*
connection_new_channel( Connection *connection, char *name) {
        Channel *channel;
        g_return_val_if_fail( connection != NULL, FALSE);
        g_return_val_if_fail( name != NULL, FALSE);
        
	if ( connection->channels && 
	     ( channel = channel_find( connection->channels, name)))
	            return channel;

        channel = channel_new( connection, connection->default_channel, name);
        
        connection_append_channel( connection, channel);
        
        return channel;
}

void
connection_append_channel( Connection *connection, Channel *channel) {
        g_return_if_fail( connection != NULL);
        g_return_if_fail( channel != NULL);
        
        connection->channels = g_list_append( connection->channels, channel);
        
        if ( connection->default_channel == NULL) 
        	connection->default_channel = channel;
}
        
void
connection_remove_channel( Connection *connection, Channel *channel) {
        g_return_if_fail( connection != NULL);
        g_return_if_fail( channel != NULL);
        
	if ( !FLAG_IS_SET( channel->flags, CHANNEL_MASTER)) {
	    if ( channel == connection->default_channel)
        	connection->default_channel = NULL;
	    connection->channels = g_list_remove( connection->channels, channel);
	    channel_free( channel);
	}
}
        
Connection*
connection_find( int sock) {
        Connection *connection;
        GList *connection_item;
        
        connection_item = gb_connections;
        while ( connection_item) {
                connection = connection_item->data;
                connection_item = connection_item->next;
                
                if ( connection->sock == sock)
                	return connection;
        }
        
        return FALSE;
}

gint
connection_input_timeout() {
	struct timeval timeout = { 0, 0};
        Connection *connection;
        GList *connection_item;
	fd_set read_fd_set;
	
	if ( gb_connections == NULL)
		return TRUE;
		
	FD_ZERO( &read_fd_set);
	
        connection_item = gb_connections;
        while ( connection_item) {
                connection = connection_item->data;
                connection_item = connection_item->next;
        	FD_SET( connection->sock, &read_fd_set);
        }
        
        if ( select( FD_SETSIZE, &read_fd_set, NULL, NULL, &timeout) < 0) {
                g_warning( "select : %s", strerror( errno));
                return FALSE;
        }
        
        connection_item = gb_connections;
        while ( connection_item) {
                connection = connection_item->data;
                connection_item = connection_item->next;
        	if ( FD_ISSET( connection->sock, &read_fd_set))
        	       (*connection->input_func) ( connection); 
        }
        
        return TRUE;
}

void
connection_error( Connection *connection, int error) {
	
	g_return_if_fail( connection != NULL);
	
	switch ( error) {
	       case PER_CONNECTION_LOST:
	       		if ( connection->default_channel)
	       			gui_window_display_message( connection->default_channel->window,
	       				"Remote end has closed connection.\n", PROPERTY_NUMERIC_REPLY);
	       		connection->sock = 0;
			UNSET_FLAG( connection->flags, CON_CONNECTED);
	       		connection_free( connection);
	       		break;
	       /* I don't know wich of those errors are likely to occur */
	       case EBADF:
	       case ENOTSOCK:
	       case EFAULT:
	       case EINTR:
	       case ENOTCONN:
	       /* these errors won't happen IMHO since we always send small packets of data in irc & dcc "modes".
		  ( even for "dcc send" because we always wait for the other side of the connection to acknowledge
		  the last sent packet before sending a new one 
	       */
	       case EMSGSIZE:
	       case ENOBUFS: 
		        sprintf( gb_tmp_buff, "connection error : %s (%d). Please mail this message to <dramboz@imaginet.fr>\n",
				 strerror( error), error);
			//if ( connection->default_channel)
			    //gui_window_display_message( connection->default_channel->window, gb_tmp_buff, PROPERTY_ERROR);
			//else
			    g_warning( gb_tmp_buff);
			/* We can't continue from here. */
			connection_free( connection);
			break;
	       default:
	       		if ( error > 0)
	       			g_warning( "connection error : %s", strerror( error));
	       		break;
	}
}

char*
connection_get_line( Connection *connection, char *separator) {
	int  byte_read = 0;
	char *end_of_msg;
	char *line = NULL;
        
	end_of_msg = strstr( connection->buffer.next_line, separator);
	if ( end_of_msg == NULL) {
       		/* There's nothing to parse in the buffer. We've got to read some more data */
       		
		/* check to see if we are not already at the begining of the buffer */
		if ( connection->buffer.next_line != connection->buffer.buff ) {
		       	int i;
		       
			i = connection->buffer.next_byte - connection->buffer.next_line;
			memmove(&connection->buffer.buff, connection->buffer.next_line, i);
			connection->buffer.next_line = connection->buffer.buff;
			connection->buffer.next_byte = connection->buffer.next_line + i;
	    	} 
	    	
       		byte_read = recv( connection->sock, connection->buffer.next_byte, 512, 0);
       		if ( byte_read == -1) {
       			/* an error occurred */    
       			if ( errno != EWOULDBLOCK)
       				connection_error( connection, errno);
       			return NULL;
       		} else if ( byte_read == 0)
       		       connection_error( connection, PER_CONNECTION_LOST);
       			
#ifdef DEBUG_PROTO
		g_print("input buffer :\n%s(EOB)\n",    connection->buffer.buff);
#endif    	
       		connection->buffer.next_byte +=  byte_read;
       		*connection->buffer.next_byte = '\0';
       		
       		end_of_msg = strstr( connection->buffer.next_line, separator);
	}
	
	if ( end_of_msg) {
		*end_of_msg = '\0';
#ifdef DEBUG_PROTO
		g_print( "in %% %s\n", connection->buffer.next_line);
#endif
		line = connection->buffer.next_line;
    		connection->buffer.next_line = end_of_msg + strlen( separator);
    		
    		return line;
      	} else if ( connection->buffer.next_byte - connection->buffer.buff >= 512) {
      	 	/*
		if it comes here then the message size is either bigger 
		than 512 bytes or the server don't use separator as message 'separator'
		which should never happens. 
  	     	*/
		g_warning( "Message is bigger than 512 bytes !!!");
		close( connection->sock);
  	}

      		
	return NULL;
}
void
irc_send( Connection *connection, void *data, unsigned int size) {
        /* Do i need to make an output queue here ? */
        int i;
        i = send( connection->sock, data, size , 0);
        if ( i != size )
        	connection_error( connection, errno);
#ifdef DEBUG_PROTO
	g_print( "out %% %s", (char*) data);
#endif
}
        	
void
irc_connect( Connection *connection, char *nick, char *realname, char *password) {    

        if ( password != NULL) {
                snprintf(gb_tmp_buff, TMP_BUFF_SIZE, "PASS %s\r\n", password);
                irc_send( connection, gb_tmp_buff, strlen(gb_tmp_buff));

        }
        	
        snprintf( gb_tmp_buff, TMP_BUFF_SIZE, "NICK %s\r\n", nick);
        irc_send( connection, gb_tmp_buff, strlen(gb_tmp_buff));
        
        snprintf(gb_tmp_buff, TMP_BUFF_SIZE, "USER %s spoof spoof :%s\r\n", 
        	getlogin(), realname ? realname : "*Unknown*");
        irc_send( connection, gb_tmp_buff, strlen(gb_tmp_buff));
        
}

void
irc_disconnect( Connection *connection, char *reason) {
        g_return_if_fail( connection->data.irc->connected);
        
        snprintf( gb_tmp_buff, TMP_BUFF_SIZE, "QUIT :%s\r\n", reason );
        irc_send( connection, gb_tmp_buff, strlen(gb_tmp_buff));
}


void
irc_input( Connection *connection) {
        char *line;
        
        if ( FLAG_IS_SET( connection->flags, CON_CONNECTED)) {
        	while ( (line = connection_get_line( connection, "\r\n")))
        		parse_server_message( connection, line);
        } else {
                SET_FLAG( connection->flags, CON_CONNECTED);
		irc_connect( connection, VAR_STR(VAR_NICKNAME), VAR_TRAILING(VAR_REALNAME), NULL);
	}
}

Connection*
dcc_find( int token, char *arg) {
        GList *connection_item;
        Connection *connection;
        DCCParam *dcc;
        
        connection_item = gb_connections;
        while ( connection_item) {
                connection = connection_item->data;
                connection_item = connection_item->next;
                
                if ( FLAG_IS_SET( connection->flags, CON_IS_DCC)) {
                        dcc = connection->data.dcc;
                	switch ( token) {
                	       case CTCP_DCC_CHAT_Q:
                	       case CTCP_DCC_CHAT_A:
                	       		if ( !strcmp( dcc->nick, arg))
                	       			return connection;
                	       		break;
                	       case CTCP_DCC_SEND_Q:
                	       case CTCP_DCC_SEND_A:
                	       		if ( !strcmp( dcc->send->filename, arg))
                	       			return connection;
                	       		break;
                	       	default:
                	       		return NULL;
                	}
                }
	}
	return NULL;
}
        
void
dcc_input( Connection *connection) {
        DCCParam *dcc;
        unsigned int byte_read = 1;
        
        dcc = connection->data.dcc;
        
        if ( FLAG_IS_SET(connection->flags, CON_LISTEN)) {
		struct sockaddr_in clientname;
                int new, size, flag;
                                
                size = sizeof (clientname);
                new = accept( connection->sock, (struct sockaddr*) &clientname, &size);
                
        	flag = fcntl(new, F_GETFL);
        	if ( fcntl(new, F_SETFL, flag | O_NONBLOCK ) == -1 )
        		new = -1;
        		
                if ( new == -1) {
                        connection_error( connection, errno);
                        return;
                }
                
                close( connection->sock);
                connection->sock = new;
                connection->address = g_malloc( 16);
                sprintf( connection->address, "%d", clientname.sin_addr.s_addr);
                
                UNSET_FLAG( connection->flags, CON_LISTEN);
                SET_FLAG( connection->flags, CON_CONNECTED);
                        			
                if ( dcc->token == CTCP_DCC_SEND_A) {
                        unsigned int size;
                        /* send the first packet */
			size = dcc_send_read_packet( connection, 0);
			       		        
 	      	        if ( send( connection->sock, connection->buffer.buff, size, 0) !=
       	       			size) {
				/* an error occured */
				connection_error( connection, errno);
				return;
    	   	        }
                	snprintf( gb_tmp_buff, TMP_BUFF_SIZE, "DCC send connection with %s (%s:%s) established\n",
                		dcc->nick, long_ip_to_str(atol(connection->address)), connection->port);
                	gui_window_display_message( connection->default_channel->window, gb_tmp_buff, PROPERTY_NUMERIC_REPLY);
    	   	} else if ( dcc->token == CTCP_DCC_CHAT_A) {
    	   	        Channel *channel;
			channel = channel_new( connection, 
					       connection->default_channel, 
			  		       dcc->nick);
			connection->default_channel = channel;
			connection_append_channel( connection, channel);
                	snprintf( gb_tmp_buff, TMP_BUFF_SIZE, "DCC chat connection with %s (%s:%s) established\n",
                		dcc->nick, long_ip_to_str(atol(connection->address)), connection->port);
                	gui_window_display_message( channel->window, gb_tmp_buff, PROPERTY_NUMERIC_REPLY);
    	   	}
      	} else {
		if ( dcc->token == CTCP_DCC_CHAT_Q || dcc->token == CTCP_DCC_CHAT_A) {
  	     		char *line;
        		
   	    		while( (line = connection_get_line( connection, "\n")))
   			    	dcc_chat_parse_line( connection, line);
   		    	
     	 	} else if ( dcc->token == CTCP_DCC_SEND_A) {
      		        unsigned int ack, size;
       		        
       		        while ( (byte_read = recv( connection->sock, &ack, sizeof(unsigned int)  , 0)) && byte_read != -1) {
				ack = ntohl( ack);
#ifdef DEBUG_PROTO
				g_print( "dcc send ack :%d\n", ack);
#endif	        
				size = dcc_send_read_packet( connection, ack);
			        if ( size == 0) {
			             connection_free( connection);
			             return;
			        }
			        
 	      	        	if ( send( connection->sock, connection->buffer.buff, size, 0) !=
       	       				size) {
					/* an error occured */
					connection_error( connection, errno);
					return;
    	   	        	}
       		       }
       		       if ( byte_read == 0)
       		       		connection_error( connection, PER_CONNECTION_LOST);
       		       		
     	  	} else if ( dcc->token == CTCP_DCC_SEND_Q) {
      	 	       unsigned int ack;
       	       
       		       while( (byte_read = recv( connection->sock, connection->buffer.buff, INPUT_BUFFER_SIZE, 0)) && byte_read != -1) {
       	       
	  	       		ack = dcc_send_write_packet( connection, byte_read);
				if ( ack == 0) {
				        connection_free( connection);
					return;
				}
#ifdef DEBUG_PROTO
				g_print( "dcc send byte read :%d\n", ack);
#endif	        

		       }
    	  	       /* Acknowlegde the received data */
     	  	       ack = htonl( ack);
      	 	       if ( send( connection->sock, &ack, sizeof(unsigned int)  , 0) != sizeof(unsigned int)) {
			        /* an error occured */
				connection_error( connection, errno);
				return;
   	    	       }
       		       if ( byte_read == 0)
       		       		connection_error( connection, PER_CONNECTION_LOST);
    	   	}
	}
}

        
