/*
 *    gnirc - Another IRC client for X 
 *    Copyright (C) 1998  david Ramboz <dramboz@imaginet.fr>
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 */


#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/utsname.h> 
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <gtk/gtk.h>

#include "gnirc.h"
#include "enums.h"

void		display_notice		( Channel *channel);
void		syntax			();
void    	init_program		( int argc, char **argv);

gint		gb_connection_timeout_tag = 0;
char		*gb_default_channel_name = "Gnirc ";
char		*gb_tmp_buff = NULL;

extern UVariable 	gb_uvariables[];
extern TextProperty 	gb_text_properties[];
        
void
exit_program (int err) {
        gtk_exit(err);
}

void
display_notice( Channel *channel) {
        g_return_if_fail( channel != NULL);
        
        sprintf( gb_tmp_buff, "%s version %s, Copyright (C) 1998 david Ramboz\n", 
        		PROGRAM_NAME, PROGRAM_VERSION);
        gui_window_display_message( channel->window, gb_tmp_buff, PROPERTY_OTHER);
        sprintf( gb_tmp_buff, "%s comes with ABSOLUTELY NO WARRANTY\n", PROGRAM_NAME);
        gui_window_display_message( channel->window, gb_tmp_buff, PROPERTY_OTHER);
        gui_window_display_message( channel->window, 
        	"This is free software, and you are welcome to redistribute it\n",
        	PROPERTY_OTHER);
        gui_window_display_message( channel->window, 
        	"under certain conditions ( see the file ./COPYING for more info).\n",
        	PROPERTY_OTHER);
}
        
        
void
syntax() {
        fprintf( stderr, 
		 "syntax: %s <nick name> <server> -r <real name> [options]\n"
		 "\t -p <port> : irc server connection port\n"
		 "\t -r <name> : user's realname ( necessary)\n"
		 "\t -u <info> : user information\n"
		 "\t -e <email>: user email\n"
		 "\t -d <directory> : change the configuration directory\n",
		 PROGRAM_NAME);
        exit_program(0);
}


/* Parse the command line and initialize user preferences */
void
init_program(int argc, char **argv) {
	GdkColormap *cmap;
        struct utsname unamebuf;
        char opt;
        int i, k;
        
        initialize_parser();
        /* Parse command line	*/
        for ( i=1; i < argc; i++) {
                if ( *argv[i] == '-') {
                        /* parse options */
                        opt = tolower( *(argv[i] + 1));
                        switch (opt) {
                                case 'p' : 
                                	k = atoi(argv[++i]);
                                	SET_VAR( VAR_DEFAULT_PORT, &i);
                                	break;
                                case 'd' : SET_VAR( VAR_CONFIG_DIR, argv[++i]);
                                	break;
                                case 'r' : SET_VAR( VAR_REALNAME, argv[++i]);
                                	break;
                                case 'e' : SET_VAR( VAR_EMAIL, argv[++i]);
                                	break;
                                case 'u' : SET_VAR( VAR_USERINFO, argv[++i]);
                                	break;
                                case '-' : /* long options */
                                	break;
                                default: syntax();
                                	break;
                        }
                } else {
                        if ( VAR_STR( VAR_NICKNAME) == NULL)
                        	SET_VAR( VAR_NICKNAME, argv[i]);
                        else if ( VAR_STR( VAR_DEFAULT_SERVER) == NULL)
                        	SET_VAR( VAR_DEFAULT_SERVER, argv[i]);
                        else syntax();
               }
        }
        
        if ( !VAR_VALUE( VAR_CONFIG_DIR)) {
        	char *config_dir = NULL;
        	char *home_dir;
        	
                home_dir = getenv("HOME");
        	config_dir = g_malloc( strlen(home_dir) + strlen(PROGRAM_NAME) + 1);
        	sprintf( config_dir, "%s/.%s", home_dir, PROGRAM_NAME);
        	if ( access( config_dir, F_OK) /* && mkdir( config_dir, S_IRWXU | S_IRWXG | S_IRWXO) */) {
        	        // g_warning( "Can't create directory %s", config_dir);
        	} else 
        		SET_VAR( VAR_CONFIG_DIR, config_dir);
        	g_free( config_dir);
        }
        	
        SET_VAR( VAR_PROGRAM_NAME	, PROGRAM_NAME);
        SET_VAR( VAR_PROGRAM_VERSION	, PROGRAM_VERSION);
        SET_VAR( VAR_PROGRAM_SOURCE_HOST, PROGRAM_SOURCE_HOST);
        SET_VAR( VAR_PROGRAM_SOURCE_DIR	, PROGRAM_SOURCE_DIR);
        SET_VAR( VAR_PROGRAM_SOURCE_FILES, PROGRAM_SOURCE_FILES);
        k = 0; 	SET_VAR( VAR_USER_MODE, &k);
        
        /* Identify the system type and find its version */
        if ( uname( &unamebuf)) {
                SET_VAR( VAR_SYSTEM_TYPE	, "unknown");
                SET_VAR( VAR_SYSTEM_VERSI0N	, "unknown");
        } else {
                SET_VAR( VAR_SYSTEM_TYPE	, unamebuf.sysname);
                SET_VAR( VAR_SYSTEM_VERSI0N	, unamebuf.release);
        }
        
        /* Initialize user preferences */
        irc_prefs_init();
        
	cmap = gtk_widget_get_default_colormap();
	for ( i = 0; i < PROPERTY_LAST; i++)
		gdk_color_alloc( cmap, &gb_text_properties[i].color);
	
}


int 
main(int argc, char **argv) {        
       	Channel *channel;
       
  	gtk_set_locale ();
  	gtk_init (&argc, &argv);
        
        //init_messages_handlers();
        gb_tmp_buff = g_malloc( TMP_BUFF_SIZE);
        init_program( argc, argv);
        
        channel = channel_new( NULL, NULL, gb_default_channel_name);
	SET_FLAG( channel->flags, CHANNEL_MASTER);
        display_notice(channel);
	if ( VAR_VALUE( VAR_DEFAULT_SERVER) && VAR_VALUE( VAR_NICKNAME) && VAR_VALUE( VAR_REALNAME)) {
        	sprintf( gb_tmp_buff, "%d", VAR_UINT( VAR_DEFAULT_PORT));
        	cmd_server( channel, VAR_STR( VAR_DEFAULT_SERVER), gb_tmp_buff);
	} else {
        	gui_window_display_message( channel->window,
        		"No server specified\n", PROPERTY_ERROR );
        	gui_window_display_message( channel->window,
        		"Use Server <server address> [<port>] to connect\n", PROPERTY_ERROR );
        }
        gb_connection_timeout_tag = gtk_timeout_add( VAR_UINT( VAR_INPUT_TIMEOUT), 
        				(GtkFunction) connection_input_timeout, NULL);
       	gtk_main ();

  	return 0;
}


