/*
 *    gnirc - Another IRC client for X 
 *    Copyright (C) 1998  david Ramboz <dramboz@imaginet.fr>
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
 
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <ctype.h>
#include <gtk/gtk.h>

#include "gnirc.h"

char gb_str_ip_buff[16];

/* get_next_item : */
int
get_next_item( char **src, char **dst, char sep) {
        
        if ( *src) {
                *src += strspn( *src, " \t");
        
        	*dst = *src;
        	while( **src && **src != sep) (*src)++;
        	if ( **src) {
        		**src = '\0';
        		(*src)++;
        		return TRUE;
        	} else return FALSE;
        } 
        *dst = NULL;
        return FALSE;
}

char*
skip_line( char *src) {
        while ( *src && *src != '\n') src++;
        return src;
}

int
get_next_string( char **src, char **dst) {
        char	*tmp;
        
        g_return_val_if_fail( *src != NULL, 0);
        
        *src += strspn( *src, " \t\n\r");
        
        while ( **src && **src == '#') {
                if (!(tmp = strchr( *src, '\n')))
		    return FALSE;
		*src = tmp;
                *src += strspn( tmp, " \t\n\r");
        }
        
	if ( !**src)
	        return FALSE;

        /* quoted strings */
        if ( **src == '\"') {
                *dst = ++(*src);
		tmp  = strchr( *src, '\"');
		if ( tmp) {
		        *tmp = '\0';
			*src = ++tmp;
			return TRUE;
		}
		return FALSE;
        }           
        
        /* non quoted strings */    
        tmp = *dst = *src;
        while ( *tmp) {
	        if ( *tmp == '#') {
                	if (!(tmp = strchr( tmp, '\n')))
			    return FALSE;
		}
                if ( *tmp == ' ' || *tmp == '\t' || *tmp == '\n') {
                        *tmp++ = '\0';
                        *src = tmp;
                        return TRUE;
                }else tmp++;
     	}
     	return FALSE;
}

int
get_next_line( char **src, char **dst) {
        char *tmp;
        g_return_val_if_fail( *src != NULL, 0);
        
        *src += strspn( *src, " \t\n");
        while ( **src && **src == '#') {
                if (!(tmp = strchr( *src, '\n')))
		        return FALSE;
		*src = tmp;
        	*src += strspn ( tmp, " \t\n");
        }
        	
        *dst = tmp = *src;
        while ( *tmp) {
                if ( *tmp == '#') {
		        *tmp = '\0';
                        if ((tmp = strchr( tmp, '\n')))
			        *src = tmp;
                        return TRUE;
                } else if ( *tmp == '\n') {
                        *tmp++ = '\0';
                        *src = tmp;
                        return TRUE;
                }
                tmp++;
        }
        
        return FALSE;
} 
                
void
lower_string( char *src) {
        while ( *src) 
                *src++ = tolower(*src);
}

void
upper_string( char *src) {
        while ( *src)
                *src++ = toupper(*src);
}

void
loup_string( char *src) {
        
        while( *src) {
                *src++ = toupper(*src);
                if ( *src)
                	*src++ = tolower(*src);
        }
}

int
compare_users( char *user1, char *user2) {
        if ( *user1 == '@')
        	return *user2 == '@' ? strcmp( user1, user2) : -1;
        else if ( *user2 == '@')
        	return 1;
        else if ( *user1 == '+')
        	return *user2 == '+' ? strcmp( user1, user2) : -1;
        else if ( *user2 == '+')
        	return 1;
        
        return strcmp( user1, user2);
}

int 
compare_nicks( char *nick1, char *nick2) {
        if ( *nick1 == '@' || *nick1 == '+')
        	nick1++;
        if ( *nick2 == '@' || *nick2 == '+')
        	nick2++;
        return strcmp( nick1, nick2);
}

char*	
get_local_time() {
        time_t timeval;
        char   *timestr;
        
        timeval = time(&timeval);
        timestr = ctime(&timeval);
        
        /* remove the '/n' char */
        timestr[strlen(timestr)-1] = '\0';
        return timestr;
}

guint		
g_num_hash( const gpointer pointer) {
        guint *h;
        
        h = (guint*) pointer;
        return *h;
}

gint
g_num_compare( gpointer v1, gpointer v2) {
        gint *x;
        gint *y;
        
        x = (gint*) v1;
        y = (gint*) v2;
        
        return *x == *y;
}

gint		
g_list_find_pos( GList *list, gpointer data) {
        gint pos = 0;
        
        while ( list) {
                if ( list->data == data)
                	return pos;
                list = list->next;
                pos++;
        }
        return 0; 
}
gpointer	
g_list_find_data( GList *list, gpointer data) {
	while (list) {
		if (list->data == data)
			break;
		list = list->next;
	}
  
	if ( list)
		return list->data;
		
	return NULL;
}

gpointer	
g_list_nth_data( GList *list, gint pos) {
	while ((pos-- > 0) && list)
		list = list->next;
  
	return list ? list->data : NULL;
}

char*
long_ip_to_str( unsigned long int ip) {
        int n1, n2, n3, n4;
        char *ipptr;
        
        ipptr = (char*) &ip;
        n1 = *ipptr++;
        n2 = *ipptr++;
        n3 = *ipptr++;
        n4 = *ipptr;
	
	sprintf( gb_str_ip_buff, "%d.%d.%d.%d", n1, n2, n3, n4);
	return gb_str_ip_buff;
}

  /***************/
 /*** History ***/
/***************/

History*
history_new( gint size, FreeFunc func) {
        History *history;
        
        g_return_val_if_fail( size > 0, NULL);
        g_return_val_if_fail( func != NULL, NULL);
        
        history = g_malloc( sizeof( History));
        history->pos_max = size - 1;
        history->pos_current = 0;
        history->vpos = 0;
        history->data = g_malloc( size * sizeof( void*));
        history->free = func;
        
        bzero( history->data, size * sizeof( void*));
        
        return history;
}

void
history_free( History *history) {
        int i;
        g_return_if_fail( history != NULL);
        
        for ( i = 0; i < history->pos_max; i++)
        	if ( history->data[i])
        		(*history->free) ( history->data[i]);
        		
        g_free( history->data);
        g_free( history);
}

void*
history_next( History *history) {
        g_return_val_if_fail( history != NULL, FALSE);
        
        if ( history->vpos == history->pos_max)
        	history->vpos = 0;
        	
       	return history->data[ history->vpos++];
}

void*
history_previous( History *history) {
        g_return_val_if_fail( history != NULL, FALSE);
        
        if ( history->vpos == 0)
        	history->vpos = history->pos_max;
        	
        return history->data[ history->vpos--];
}

void*
history_first( History *history) {
        g_return_val_if_fail( history != NULL, FALSE);
        
	if ( history->data[ history->pos_current])
		return history->data[ history->pos_current];
	else
		return history->data[ 0];
}

void*
history_last( History *history) {
        g_return_val_if_fail( history != NULL, FALSE);
        
        if ( history->pos_current > 0)
       		return history->data[ history->pos_current - 1];
       	else
       		return history->data[ history->pos_max];
}

void
history_rewind( History *history) {
        g_return_if_fail( history != NULL);
        
	if ( history->data[ history->pos_current])
		history->vpos =  history->pos_current;
	else
		history->vpos = 0;

}
        
void
history_append( History *history, void *data) {
        g_return_if_fail( history != NULL);
        g_return_if_fail( data != NULL);
        
        if ( history->data[ history->pos_current])
        	(*history->free) ( history->data[ history->pos_current]);
        
        history->data[ history->pos_current] = data;
        history->pos_current = history->pos_current == history->pos_max ? 0 : history->pos_current + 1;
        history->vpos = history->pos_current;
}

void
history_dump_strings( History *history) {
        int i;
        
        g_print( "History size : %d\n", history->pos_max);
        g_print( "History current pos : %d\n", history->pos_current);
        for ( i = 0; i < history->pos_max; i++)
                if ( history->data[i])
			g_print( "%d - %s\n", i, (gchar*) history->data[i]);
}

GtkWidget*
new_pixmap (char      *filename,
	    GdkWindow *window,
	    GdkColor  *background)
{
	GtkWidget *wpixmap;
	GdkPixmap *pixmap;
	GdkBitmap *mask;

	if ( !filename) 
		return NULL;
		
	pixmap = gdk_pixmap_create_from_xpm (window, &mask,
				       background,
				       filename);
	wpixmap = pixmap ? gtk_pixmap_new (pixmap, mask) : NULL;

	return wpixmap;
}


void
destroy_widget_cb( GtkWidget *w, gpointer *data) {
        *data = NULL;
}

void
gtk_text_scroll_down_to_end( GtkText	*text) {
        g_return_if_fail( GTK_IS_TEXT( text));

	/* I don't know what this means but if i don't put this 
           here, the program will crash	
	*/
        if ( text->line_start_cache && text->line_start_cache->next) {
                text->vadj->value = text->vadj->upper - text->vadj->page_size;
		gtk_signal_emit_by_name (GTK_OBJECT (text->vadj), "value_changed");
		
	}
	
}      

void
gtk_text_delete_first_line( GtkText *text) {
	guint line_size = 0;
        
        g_return_if_fail( GTK_IS_TEXT( text));
             
	while ( GTK_TEXT_INDEX( text, line_size) != '\n')
	     	line_size++;
	gtk_text_set_point( text, 0);
	gtk_text_forward_delete( text, line_size + 1);
	gtk_text_set_point( text, gtk_text_get_length( text));
}
      
