/*
 *    gnirc - Another IRC client for X 
 *    Copyright (C) 1998  david Ramboz <dramboz@imaginet.fr>
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 */
 
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <gtk/gtk.h>
#include <gdk/gdk.h>

#include "gnirc.h"
#include "enums.h"

#define IRC_SERVERS_FILE "servers"
#define GTK_RC_FILE  	 "ircrcx"
#define IRC_RC_FILE	 "ircrc"
#define IRC_MENU_FILE	 "menus"

char*		read_next_item		( FILE *file, int type, char **pos);
char*		pref_file_find	( char **paths, char *default_dir, char *filename);

GdkColor	property_privmsg 	=  {0, 0x0   , 0x0   , 0x0   };
GdkColor	property_notice 	=  {0, 0x4103, 0x69a6, 0xe79d};
GdkColor	property_action		=  {0, 0x28a2, 0x8a28, 0x5144};
GdkColor	property_numeric_reply 	=  {0, 0x8e38, 0x4514, 0x1040};
GdkColor	property_error		=  {0, 0xb6da, 0x2081, 0x2081};
GdkColor	property_ctcp_query	=  {0, 0xcf3c, 0x0   , 0xcf3c};

static char *properties_label[] = {
        "text_privmsg",
        "text_notice",
        "text_action",
        "text_reply",
        "text_ctcp_query",
        "text_ctcp_reply",
        "text_error",
        "text_other",
        NULL
};

extern UVariable gb_uvariables[];
TextProperty	gb_text_properties[PROPERTY_LAST];
GList 		*gb_toolbar_list = NULL;
GList 		*gb_menu_list = NULL;
GList		*gb_irc_networks = NULL;

char 		*gb_pref_buff = NULL;
int   		gb_pref_buff_size = 4096;

char		*gb_search_path[] = {
        		"/usr/lib",
        		"/usr/local/lib",
        		NULL
        	};
enum {
        A_STRING,
        A_LINE
};

void
irc_prefs_init() {
	GtkWidget *trick;
	GtkStyle  *style;
	int	  i;
	
	
        gb_pref_buff	= g_malloc( gb_pref_buff_size);
        
        gb_text_properties[PROPERTY_PRIVMSG].color	= property_privmsg;
        gb_text_properties[PROPERTY_NOTICE].color	= property_notice;
        gb_text_properties[PROPERTY_ACTION].color	= property_action;
        gb_text_properties[PROPERTY_NUMERIC_REPLY].color = property_numeric_reply;
        gb_text_properties[PROPERTY_ERROR].color	= property_error;
        gb_text_properties[PROPERTY_CTCP_QUERY].color	= property_ctcp_query;
        gb_text_properties[PROPERTY_CTCP_REPLY].color	= property_ctcp_query;
        gb_text_properties[PROPERTY_OTHER].color		= property_numeric_reply;
        
        i = 850		; SET_VAR( VAR_WINDOW_WIDTH	, &i);
        i = 400		; SET_VAR( VAR_WINDOW_HEIGHT	, &i);
        i = 6667	; SET_VAR( VAR_DEFAULT_PORT	, &i);
        i = 100		; SET_VAR( VAR_HISTORY_SIZE	, &i);
        i = 1		; SET_VAR( VAR_SHOW_USER_LIST	, &i);
        i = DEFAULT_INPUT_TIMEOUT; SET_VAR( VAR_INPUT_TIMEOUT	, &i);
        SET_VAR( VAR_SHOW_TOOLBAR	, &i);
        SET_VAR( VAR_ESCAPE_CHAR	, "/");
        SET_IF_NULL( VAR_NICKNAME		, "SpaceMan");
        SET_IF_NULL( VAR_REALNAME		, "Unknown");
        
        parse_ircrc( IRC_RC_FILE);
        parse_menus( IRC_MENU_FILE);
        parse_server_list( IRC_SERVERS_FILE);
        parse_gtkrc( GTK_RC_FILE);
                        
        trick = gtk_hbox_new( FALSE, 0);
        for ( i = 0; properties_label[i]; i++) {
        	       gtk_widget_set_name( trick, properties_label[i]);
        	       style = gtk_rc_get_style( trick);
        	       if ( style) {
        	       		gb_text_properties[i].color = style->fg[GTK_STATE_NORMAL]; 
        	       		gb_text_properties[i].font  = style->font; 
        	       }
       	}
	
	gb_pref_buff = NULL;
        g_free( gb_pref_buff);
}

int
parse_menus( char *file) {
        FILE	*menu_file;
        GList	*list = NULL;
        char	*label, *command, *icon, *help;
        char    *tmp = NULL;
        IRCMenu 	*menu_item;
        IRCToolbar 	*toolbar_item;
        
        file = pref_file_find( gb_search_path, VAR_STR( VAR_CONFIG_DIR), file);
        if ( !file)
        	return FALSE;
        	
        menu_file = fopen( file, "r");
        if ( !menu_file) {
                g_warning( "Can't open configuration file %s for reading (%s)", file, strerror( errno));
                return FALSE;
        }
        g_free( file);
        while ( (label = read_next_item( menu_file, A_STRING, &tmp))) { 
        	if ( !strcmp( label, "begin_toolbar")) {
        	       list = NULL;
        	       label = read_next_item( menu_file, A_STRING, &tmp);
        	       while ( (label = read_next_item( menu_file, A_STRING, &tmp)) &&
        	       	   strcmp( label, "end_toolbar")) {
        	       		command	= read_next_item(  menu_file, A_STRING, &tmp);
        	       		icon	= read_next_item(  menu_file, A_STRING, &tmp);
        	       		help	= read_next_item(  menu_file, A_STRING, &tmp);
        	       		
        	       		if ( command == NULL || icon == NULL || help == NULL) {
        	       		      	fclose( menu_file);
        	       			return FALSE;
        	       		}
        	       			
        	       		toolbar_item = g_malloc( sizeof( IRCToolbar));
        	       		toolbar_item->label = *label ? label: NULL;
        	       		toolbar_item->command = *command ? command : NULL;
        	       		toolbar_item->icon = *icon ? icon : NULL;
        	       		toolbar_item->help = *help ? help : NULL;
        	       		list = g_list_append( list, toolbar_item);
        	      }
        	      gb_toolbar_list = g_list_append( gb_toolbar_list, list);
        	      
        	} else if ( !strcmp( label, "begin_menu")) {
        	       list  = NULL;
        	       
        	       while ( (label = read_next_item( menu_file, A_STRING, &tmp)) &&
        	       	   strcmp( label, "end_menu")) {
        	       		command = read_next_item(  menu_file, A_STRING, &tmp);
        	       		
        	       		if ( command == NULL) {
        	       		        fclose( menu_file);
        	       			return FALSE;
        	       		}
                       		menu_item = g_malloc( sizeof( IRCMenu));
                       		menu_item -> label    = label;
                      		menu_item -> command  = command;
		       		list = g_list_append( list, menu_item);
		       }
		       
		       gb_menu_list = g_list_append( gb_menu_list, list);
		}
	}
	
	fclose( menu_file);
	return TRUE;
}                

int 
parse_ircrc( char *file) {
        FILE	*rcfile;
        char	*line;
        char	*tmp = NULL;
        
        file = pref_file_find( gb_search_path, VAR_STR( VAR_CONFIG_DIR), file);
        if ( !file)
        	return FALSE;
        	
        rcfile = fopen( file, "r");
        if ( !rcfile) {
                g_warning( "Can't open configuration file %s for reading (%s)", file, strerror( errno));
                return FALSE;
        }
        g_free( file);
        
        while ( (line = read_next_item( rcfile, A_LINE, &tmp))) 
        	parse_user_command( NULL, line);
	
	fclose( rcfile);
	return TRUE;
}

int
parse_server_list( char *file) {
        FILE 	*server_file;
        IRCNetwork *irc_network;
        IRCServer *server;
        char	*label, *address, *port, *desc;
        char	*tmp = NULL;
        
        file = pref_file_find( gb_search_path, VAR_STR( VAR_CONFIG_DIR), file);
        if ( !file)
        	return FALSE;
        	
        server_file = fopen( file, "r");	
        if ( !server_file) {
                g_warning( "Can't open configuration file %s for reading (%s)", file, strerror( errno));
                return FALSE;
        }
        g_free( file);
        while ( (label = read_next_item( server_file, A_STRING, &tmp)))
                
	        if ( !strcmp( label, "begin_irc_network")) {
			      label = read_next_item( server_file, A_STRING, &tmp);
			      desc  = read_next_item( server_file, A_STRING, &tmp);
			      
			      if ( label == NULL || desc == NULL) {
			             	fclose( server_file);
			      		return FALSE;
			      }
			      irc_network = g_malloc( sizeof( IRCNetwork));
	                      irc_network->label = g_strdup( label);
	                      irc_network->description = g_strdup( desc);
	                      irc_network->servers = NULL;
	                      gb_irc_networks = g_list_append( gb_irc_networks, irc_network);
	                      
	                      while ( (label = read_next_item( server_file, A_STRING, &tmp)) && 
	                      	strcmp( label, "end_irc_network")) {
	                       		address = read_next_item( server_file, A_STRING, &tmp);
	                       		port	= read_next_item( server_file, A_STRING, &tmp);
	                       		
					if ( address == NULL || port == NULL) {
					   	fclose( server_file);
						return FALSE;
					}
					server = g_malloc( sizeof( IRCServer));
					server->label = label;
					server->address = address;
					server->ports = g_list_append( NULL, port);
					//while( get_next_string( &tmp, &port)) 
					//	server->ports = g_list_append( server->ports, g_strdup(port));
					irc_network->servers = g_list_append( irc_network->servers, server);
			     }
			     
		}
	
	fclose( server_file);
	return TRUE;
}
		
int
parse_gtkrc( char *file) {        
        file = pref_file_find( gb_search_path, VAR_STR( VAR_CONFIG_DIR), file);
        if ( !file)
        	return FALSE;
        gtk_rc_parse( file);
        g_free( file);
        return TRUE;
}
        
char*
read_next_item( FILE *file, int type, char **pos) {
	char *string;
	int  i = 0;
	int  byte_read;
	
	if ( *pos) {
	       	switch ( type) {
	       	       case A_STRING:
				if ( get_next_string( pos, &string))
					return g_strdup(string);
				break;
		       case A_LINE:
				if ( get_next_line( pos, &string))
					return g_strdup(string);
				break;
		       default:
		       		return NULL;
		}
		i = strlen( *pos);
		memmove( gb_pref_buff, *pos, i);
	}
	byte_read = fread( gb_pref_buff + i, 1, gb_pref_buff_size-i-1, file);
	if ( !byte_read) {
	        if ( feof( file) && i) {
	               *pos = NULL;
	               gb_pref_buff[i] = '\0';
	               return gb_pref_buff;
	        }
		return NULL;
	}
	gb_pref_buff[byte_read+i] = '\0';
	*pos = gb_pref_buff;
	switch ( type) {
      	       case A_STRING:
			if ( get_next_string( pos, &string))
				return g_strdup(string);
			break;
	       case A_LINE:
			if ( get_next_line( pos, &string))
				return g_strdup(string);
			break;
	       default:
	       		return NULL;
	}
       	g_warning( "The buffer is to small to hold this huge string");
       	return NULL;
}

char*
pref_file_find( char **paths, char *default_dir, char *filename) {
        char *wd, *result;
        char *dir = NULL;
        
        g_return_val_if_fail( paths != NULL, FALSE);
        g_return_val_if_fail( filename != NULL, FALSE);
        
        wd = getcwd( NULL, 0);
        
        if ( default_dir != NULL &&
             !chdir( default_dir) && 
             !access( filename, R_OK)) {
        	result = g_malloc( strlen( default_dir) + strlen( filename) + 2);
        	sprintf( result, "%s/%s", default_dir, filename);
        	if ( wd) {
		    chdir( wd);
		    g_free( wd);
		}
        	return result;
         }
        
        while ( *paths) {
                if ( !chdir( *paths) && 
                     !chdir( PROGRAM_NAME) &&
                     !access( filename, R_OK)) {
                	dir = *paths;
                	break;
                }
                
                paths++;
       }
       
       if ( wd) {
       		chdir( wd);
       		g_free( wd);
       }
       
       if ( dir) {        	
        	result = g_malloc( strlen( dir) + strlen( PROGRAM_NAME) + strlen( filename) + 3);
        	sprintf( result, "%s/%s/%s", dir, PROGRAM_NAME, filename);
        	
        	return result;
       }
       
       return NULL;
}
                
        
        



