/*
 *    gnirc - Another IRC client for X 
 *    Copyright (C) 1998  david Ramboz <dramboz@imaginet.fr>
 *
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 */

#include <stdio.h> 
#include <unistd.h>
#include <string.h>
#include <gdk/gdkkeysyms.h>
#include <gtk/gtk.h>

#include "gnirc.h"
#include "enums.h"

/* some icons */

#include "icons.h"

GtkWidget*	gui_window_init		( GUIWindow *window);
GtkWidget*	gui_window_init_toolbar	( GUIWindow *window);
void	gui_window_toolbar_cb		( GtkWidget *button, GUIWindow *window);
void	gui_window_select_channel_cb	( GtkWidget *w, gint row, gint column, GdkEventButton * bevent, GUIWindow *window);
void	gui_window_select_channel_	( GUIWindow *window, Channel *channel, int select);
gint	gui_window_handle_key_press_cb  ( GtkWidget *entry, 
				          GdkEventKey *key, 
	      		 	          GUIWindow *window);
void	gui_window_build_channel_list	( GUIWindow *window);
void	gui_window_destroy_cb 		( GtkWidget *w, GUIWindow *window);
void	gui_window_add_user		( GUIWindow *window, gchar *nickname, gint pos);
void	gui_window_delete_user		( GUIWindow *window, gint pos);
void	gui_window_build_user_list	( GUIWindow *window);


GtkWidget*	gui_clist_init		( GUICList *list, char **titles);
void		gui_clist_select_cb	( GtkWidget *widget,
			 		  gint row, 
	     				  gint column, 
	     				  GdkEventButton *beven);
void		gui_clist_destroy_cb	( GtkWidget *w, GUICList *list);
void		gui_clist_make_filesel_cb( GtkWidget *w, GUICList *list);
void		gui_clist_save_cb	( GtkWidget *w, GUICList *list);

GtkWidget*	gui_server_list_init	( GUIServerList *slist);
void		gui_server_list_select_server_cb( GtkWidget *list, GtkWidget *list_item, GUIServerList *slist);
void		gui_server_list_select_network_cb( GtkWidget *menuitem, GUIServerList *slist);

void		gui_notebook_append_frame	( GtkWidget *notebook, GtkWidget *frame, char *name);

extern	 char	  *gb_tmp_buff;
extern   TextProperty gb_text_properties[];
extern	 UVariable gb_uvariables[];
extern	 GList	  *gb_irc_networks;
extern   GList	  *gb_toolbar_list;

gint	 gb_number_of_windows = 0;
GList    *gb_gui_windows_list = NULL;

void
handle_delete_window_event( GtkWidget *window, Channel *channel) {
        gb_number_of_windows--;
        gtk_widget_destroy( window);
        if ( gb_number_of_windows < 1)
        	exit_program(0);
}

void
register_new_window( GtkWidget *window, Channel *channel) {
        
        gtk_signal_connect( GTK_OBJECT( window), "delete_event",
        	(GtkSignalFunc) handle_delete_window_event, channel);
        gb_number_of_windows++;
}

/**********************/
/*** Channel window ***/

/* Channels windows contain all PRIVMSG either to a channel or a nickname */

GtkWidget*
gui_window_init( GUIWindow *window) {
        GtkWidget *mainbox;
	GtkWidget *table;
	GtkWidget *vscroll;
        GtkWidget *hbox;
        
        mainbox = gtk_vbox_new( FALSE, 0);
        gtk_container_border_width( GTK_CONTAINER(mainbox), 3);
        window->channels	= NULL;
        window->current_channel = NULL;
        window->dont_freeze 	= 0;
        window->num_lines	= 0;
        window->max_lines	= VAR_INT( VAR_HISTORY_SIZE) * 3; 
        window->history 	= history_new( VAR_INT( VAR_HISTORY_SIZE), (FreeFunc) g_free);
        window->levels		= LEVEL_ALL;
	                
        hbox = gtk_hbox_new( FALSE, 0);
        gtk_box_pack_start( GTK_BOX( mainbox), hbox, TRUE, TRUE, 0);
        gtk_widget_show( hbox);
        
        window->channel_list = gtk_clist_new( 3);
        gtk_box_pack_start( GTK_BOX( hbox), window->channel_list, FALSE, TRUE, 0);
	gtk_clist_set_policy( GTK_CLIST( window->channel_list), GTK_POLICY_AUTOMATIC,
	       		GTK_POLICY_AUTOMATIC);
        gtk_widget_set_usize( window->channel_list, 100, -1);
	//gtk_clist_column_titles_hide( GTK_CLIST( window->channel_list));
	gtk_signal_connect( GTK_OBJECT( window->channel_list), "select_row", 
		(GtkSignalFunc) gui_window_select_channel_cb, window);
        gtk_widget_show( window->channel_list);
        
        table = gtk_table_new (2, 2, FALSE);
        gtk_table_set_row_spacing (GTK_TABLE (table), 0, 0);
        gtk_table_set_col_spacing (GTK_TABLE (table), 0, 2);
        gtk_box_pack_start (GTK_BOX (hbox), table, TRUE, TRUE, 5);
        gtk_widget_show (table);

	window->text = gtk_text_new (NULL, NULL);
	gtk_text_set_word_wrap( GTK_TEXT( window->text), TRUE);
        gtk_table_attach_defaults (GTK_TABLE (table), window->text, 0, 1, 0, 1);
        GTK_WIDGET_UNSET_FLAGS( window->text, GTK_CAN_FOCUS);
	gtk_widget_show(window->text);
	
        vscroll = gtk_vscrollbar_new (GTK_TEXT (window->text)->vadj);
        gtk_table_attach (GTK_TABLE (table), vscroll, 1, 2, 0, 1,
			GTK_FILL, GTK_EXPAND | GTK_FILL, 0, 0);
        GTK_WIDGET_UNSET_FLAGS( vscroll, GTK_CAN_FOCUS);
        gtk_widget_show (vscroll);
        	
	window->status = gtk_label_new("");
	gtk_widget_set_name( window->status, "Status");
	gtk_misc_set_alignment( GTK_MISC( window->status), 0.1, 0.5);
	gtk_box_pack_start(GTK_BOX(mainbox), window->status, FALSE, TRUE, 2);
	gtk_widget_show(window->status);
	
	window->cmdline = gtk_entry_new();
	gtk_signal_connect( GTK_OBJECT(window->cmdline), "key_press_event",
		(GtkSignalFunc) gui_window_handle_key_press_cb, window);
	gtk_box_pack_start(GTK_BOX(mainbox), window->cmdline, FALSE, TRUE, 2);
	gtk_widget_show(window->cmdline);
	
	if ( VAR_INT( VAR_SHOW_USER_LIST)) {
	       	       
	       window->user_list = gtk_clist_new( 1);
	       gtk_widget_set_usize( window->user_list, 100, -1);
	       gtk_box_pack_end( GTK_BOX( hbox), window->user_list, FALSE, TRUE, 0);
	       gtk_clist_set_policy( GTK_CLIST( window->user_list), GTK_POLICY_AUTOMATIC,
	       		GTK_POLICY_AUTOMATIC);
	       //gtk_clist_column_titles_hide( GTK_CLIST( window->user_list));
	       gtk_widget_show( window->user_list);
	       
	       return mainbox;
	       
	} 
	window->user_list = NULL;
	return mainbox;
}

GtkWidget*
gui_window_init_toolbar( GUIWindow *window) {
        GtkWidget *toolbar;
        GtkWidget *button;
        GList	  *item;
        IRCToolbar *toolbar_item;
        char	  *pix_path, *wd;
        
        g_return_val_if_fail( window != NULL, FALSE);
        g_return_val_if_fail( GTK_IS_BOX( window->channel_list->parent->parent), FALSE);
        
        toolbar = gtk_toolbar_new (GTK_ORIENTATION_HORIZONTAL, GTK_TOOLBAR_BOTH);
        gtk_box_pack_start( GTK_BOX(window->channel_list->parent->parent), toolbar, FALSE, FALSE, 5);
        gtk_box_reorder_child( GTK_BOX(window->channel_list->parent->parent), toolbar, 0);
        gtk_widget_realize( toolbar);
        
        wd = getcwd( NULL, 0);
        if (  ( pix_path = VAR_STR( VAR_PROGRAM_PIXMAPS_PATH)) && chdir( pix_path))
                g_warning( "Can't cd to %s : %s", pix_path, strerror( errno));
        
        item = gb_toolbar_list ? gb_toolbar_list->data : NULL;
        while ( item) {
                toolbar_item = (IRCToolbar*) item->data;
                item = item->next;
                
        	button = gtk_toolbar_append_item ( GTK_TOOLBAR (toolbar), toolbar_item->label, toolbar_item->help, NULL,
        			new_pixmap( toolbar_item->icon, toolbar->window, &toolbar->style->bg[GTK_STATE_NORMAL]),
			        (GtkSignalFunc) gui_window_toolbar_cb, window);
		gtk_object_set_user_data( GTK_OBJECT( button), toolbar_item);
		gtk_container_border_width( GTK_CONTAINER( button), 3);
		
		//gtk_toolbar_append_space( GTK_TOOLBAR (toolbar));
		
        }
        
        if ( wd) {
                chdir( wd);
                g_free( wd);
        }
          
        gtk_widget_show( toolbar);
        return toolbar;
}
        
GUIWindow*
gui_window_new( ) {
	static GdkPixmap *app_icon_pix = NULL;
	static GdkPixmap *app_icon_mask = NULL;
	GUIWindow *window;
	GtkWidget *mainbox;
	
	window = g_malloc(sizeof(GUIWindow));
	gb_gui_windows_list = g_list_append( gb_gui_windows_list, window);

	window->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	register_new_window( window->window, NULL);
	gtk_signal_connect( GTK_OBJECT( window->window), "destroy",
		(GtkSignalFunc) gui_window_destroy_cb, window);
	gtk_widget_set_usize(window->window, VAR_UINT( VAR_WINDOW_WIDTH),
					     VAR_UINT( VAR_WINDOW_HEIGHT));
		
	gtk_window_set_policy( GTK_WINDOW(window->window), 1, 1, 0);
	gtk_container_border_width( GTK_CONTAINER(window->window), 2);
	gtk_widget_realize( window->window);
	if ( app_icon_pix == NULL) {
		app_icon_pix = gdk_pixmap_create_from_xpm_d( window->window->window, 
			&app_icon_mask, &window->window->style->bg[GTK_STATE_NORMAL],
			feet_xpm);
	}
	gdk_window_set_icon( window->window->window, NULL, app_icon_pix, app_icon_mask);
	
	mainbox = gui_window_init( window);
	gtk_container_add(GTK_CONTAINER(window->window), mainbox);
	gtk_signal_connect_object( GTK_OBJECT( window->window), "key_press_event", 
		(GtkSignalFunc) gtk_widget_grab_focus, GTK_OBJECT( window->cmdline));
	gtk_widget_show(mainbox);
        
        if ( VAR_INT( VAR_SHOW_TOOLBAR))
        	gui_window_init_toolbar( window);
        
	return window;
}

void
gui_window_destroy_cb( GtkWidget *w, GUIWindow *window) {
        GList   *item;
        Channel *channel;
        
	gb_gui_windows_list = g_list_remove( gb_gui_windows_list, window);

        item = window->channels;
        while ( item) {
                channel = item->data;
                item = item->next;
                channel->window = NULL;
       		cmd_leave( channel->connection, channel);
        }
        if ( window->channels)
        	g_list_free( window->channels);
        	
        g_free( window);
}

void
gui_window_select_channel_cb( GtkWidget *w, gint row, gint column, GdkEventButton * bevent, GUIWindow *window) {
        Channel *channel;
        
        if ( bevent) 
        	switch ( bevent->button) {
                	case 1: /* the selected channel becomes the current channel */
        			channel = g_list_nth_data( window->channels, row);
        			gui_window_select_channel_( window, channel, 0);
        			break;
        		case 3: /* popup the channel menu */
        		default:
        			break;
        	}
        		
}

void
gui_window_select_channel_( GUIWindow *window, Channel *channel, int select) {
        gint i;
        char *msg;
        
        if ( channel->window->current_channel == NULL) {
                channel->window->current_channel = channel;
                gtk_clist_select_row( GTK_CLIST( window->channel_list), 0, 2);
        } else if ( channel != channel->window->current_channel) {
                
                /* 
                
                i = g_list_find_pos( window->channels, channel->window->current_channel);
                gtk_clist_unselect_row( GTK_CLIST( window->channel_list), i, 2);
                */
                
                channel->window->current_channel = channel;
                if ( select) {
                	i = g_list_find_pos( window->channels, channel);
                	gtk_clist_select_row( GTK_CLIST( window->channel_list), i, 2);
                }
                
      		gui_window_build_user_list( channel->window);
       		gui_window_update_status( channel->window);
       		
	       	msg = g_malloc( strlen( channel->name) + 32);
	       	sprintf( msg, "%s is now the current channel\n", channel->name);
	       	gui_window_display_message( channel->window, msg, PROPERTY_NUMERIC_REPLY);
	       	g_free( msg);
	}
}
        
void
gui_window_select_channel( GUIWindow *window, Channel *channel) {        
        g_return_if_fail( window != NULL);
        g_return_if_fail( channel != NULL);
        
        gui_window_select_channel_( window, channel, 1);
        
}
         
void
gui_window_append_channel_item( GUIWindow *window, Channel *channel) {
        gchar *row[3] = { "", "", NULL};
        
        row[2] = channel->name;
        gtk_clist_append( GTK_CLIST( window->channel_list), row);
        
}
        
void
gui_window_build_channel_list( GUIWindow *window) {
        Channel *channel;
        GList *item;
        
        gtk_clist_freeze( GTK_CLIST( window->channel_list));
        gtk_clist_clear( GTK_CLIST( window->channel_list));
        
        item = window->channels;
        while( item) {
                channel = (Channel*) item->data;
                item = item->next;
                gui_window_append_channel_item( window, channel);
        }
          
        gtk_clist_thaw( GTK_CLIST( window->channel_list));
                 
}
     
void
gui_window_new_channel( GUIWindow *window, Channel *channel) {
        
        g_return_if_fail( window != NULL);
        g_return_if_fail( channel != NULL);
        
        window->channels = g_list_append( window->channels, channel);
        
        gui_window_append_channel_item( window, channel);
        gui_window_select_channel( window, channel);
        gui_window_build_user_list( window);
        gui_window_update_status( window);
        
        if ( g_list_length( window->channels) == 1)
        	gtk_clist_select_row( GTK_CLIST( window->channel_list), 0, 2);
}

void
gui_window_delete_channel( GUIWindow *window, Channel *channel) {
        
        g_return_if_fail( window != NULL);
        g_return_if_fail( channel != NULL);
        
        gtk_clist_remove( GTK_CLIST( window->channel_list), g_list_find_pos( window->channels, channel));
        window->channels = g_list_remove( window->channels, channel);
        if ( window->channels) {
        	if ( window->current_channel == channel) {
                	window->current_channel = (Channel*) window->channels->data;
                	gui_window_build_user_list( window);
                }
        	gui_window_update_status( window);
        } else
        	gtk_widget_destroy( window->window);
}

void
gui_window_build_user_list( GUIWindow *window) {
        GList *item;
        char  *nick;
        gint  row = 0;
        
        if ( window->user_list) {
                gtk_clist_freeze( GTK_CLIST( window->user_list));
                
 	        gtk_clist_clear( GTK_CLIST( window->user_list));
 	        
 	        row = 0;
                item = window->current_channel->users;
                while ( item) {
                        nick = (gchar*) item->data;
                        gui_window_add_user( window, nick, row++);
                        item = item->next;
                }
                gtk_clist_thaw( GTK_CLIST( window->user_list));
        }
}

                
void	
gui_window_add_user( GUIWindow *window, gchar *nickname, gint pos) {
        static Icon icons[3] = {
                { NULL, NULL},
                { NULL, NULL},
                { NULL, NULL}
        };
                
        if ( window->user_list) {
	        if ( icons[0].pixmap == NULL) {
			if ( !GTK_WIDGET_REALIZED( window->user_list))
  		              	gtk_widget_realize( window->user_list);
                
			icons[0].pixmap = gdk_pixmap_create_from_xpm_d( window->user_list->window, 
                			&icons[0].mask,
                			&window->user_list->style->fg[GTK_STATE_NORMAL],
                			ball_blue_xpm);
			icons[1].pixmap = gdk_pixmap_create_from_xpm_d( window->user_list->window, 
     	           			&icons[1].mask,
                			&window->user_list->style->fg[GTK_STATE_NORMAL],
                			ball_orange_xpm);
	                icons[2].pixmap = gdk_pixmap_create_from_xpm_d( window->user_list->window, 
                			&icons[2].mask,
                			&window->user_list->style->fg[GTK_STATE_NORMAL],
                			ball_white_xpm);
	        }
                
                switch ( *nickname) {
                        case '@' :
                        	nickname++;
               			gtk_clist_insert( GTK_CLIST( window->user_list), pos, &nickname);
                        	gtk_clist_set_pixtext( GTK_CLIST( window->user_list),
                        		pos, 0, nickname, 5, icons[0].pixmap, icons[0].mask);
                        	break;
                        case '+' :
                        	nickname++;
               			gtk_clist_insert( GTK_CLIST( window->user_list), pos, &nickname);
                        	gtk_clist_set_pixtext( GTK_CLIST( window->user_list),
                        		pos, 0, nickname, 5, icons[1].pixmap, icons[1].mask);
                        	break;
                        default :
               			gtk_clist_insert( GTK_CLIST( window->user_list), pos, &nickname);
                        	gtk_clist_set_pixtext( GTK_CLIST( window->user_list),
                        		pos, 0, nickname, 5, icons[2].pixmap, icons[2].mask);
                        	break;
                }
        }
}

void 
gui_window_delete_user( GUIWindow *window, gint pos) {
        if ( window->user_list) {
                gtk_clist_remove( GTK_CLIST( window->user_list), pos);
        }
}

void	
gui_window_toolbar_cb ( GtkWidget *button, GUIWindow *window) {
        IRCToolbar *toolbar_item;
        
        toolbar_item = gtk_object_get_user_data( GTK_OBJECT( button));
        g_return_if_fail( toolbar_item != NULL);
        
        parse_user_command( window->current_channel, toolbar_item->command);
}


gint
gui_window_handle_key_press_cb ( GtkWidget *entry, GdkEventKey *key, 
				GUIWindow *window) {
	char *cmd;
	cmd = gtk_entry_get_text(GTK_ENTRY(entry));
			
	switch ( key->keyval) {
		case GDK_Return :
			if ( cmd && *cmd != '\0') {
			        history_append( window->history, g_strdup( cmd));
                        	parse_user_message( window->current_channel, cmd);
                		gtk_entry_set_text(GTK_ENTRY(entry), "");
                	}
                	break;
                case GDK_Up:
                	cmd = (char*) history_previous( window->history);
                	if ( cmd)
                		gtk_entry_set_text(GTK_ENTRY(entry), cmd);
                	break;
                case GDK_Down:
                	cmd = (char*) history_next( window->history);
                	if ( cmd)
                		gtk_entry_set_text(GTK_ENTRY(entry), cmd);
                	break;
                default:
                	break;
	}
        return 1;  
}

        
void
gui_window_display_message( GUIWindow *window, char *text, int property) {
        gint scroll;
        
        g_return_if_fail( window != NULL);
        g_return_if_fail( text != NULL);
        g_return_if_fail( property < PROPERTY_LAST && property >= 0);
        
        scroll = GTK_TEXT( window->text)->vadj->upper - GTK_TEXT( window->text)->vadj->value 
        		== GTK_TEXT( window->text)->vadj->page_size;
        
        window->num_lines++;		
        if ( window->dont_freeze) {
       		if ( window->num_lines > window->max_lines)
       			gtk_text_delete_first_line( GTK_TEXT( window->text));
       			
        	gtk_text_insert ( GTK_TEXT( window->text),
        		  gb_text_properties[property].font,
        		  &gb_text_properties[property].color,
        		  NULL,
        		  text, -1 );    
        } else {    
        	gtk_text_freeze ( GTK_TEXT( window->text));
        	
       		if ( window->num_lines > window->max_lines)
       			gtk_text_delete_first_line( GTK_TEXT( window->text));
       			
        	gtk_text_insert ( GTK_TEXT( window->text),
        		  gb_text_properties[property].font,
        		  &gb_text_properties[property].color,
        		  NULL,
        		  text, -1 );    
        
       			
        	gtk_text_thaw( GTK_TEXT( window->text));
        }

        if ( scroll)
		gtk_text_scroll_down_to_end( GTK_TEXT( window->text));    
}

/* depreciated */
void
gui_window_display_ircmessage( GUIWindow *window, IRCMessage *param) {
        char *message = NULL;
        int  property = 0;
        
        g_return_if_fail( window != NULL);
        g_return_if_fail( param != NULL);
        
        format_message( param, NULL, &message, &property);
        if ( message) {
        	gui_window_display_message( window, message, property);
        	g_free( message);
        }
        
}
                	
void
gui_window_update_status( GUIWindow *window) {
       
       g_return_if_fail( window != NULL);
       
       if ( window->current_channel && window->current_channel->connection) {
        	if ( FLAG_IS_SET( window->current_channel->connection->flags, CON_IS_DCC))
        		sprintf( gb_tmp_buff, "DCC chat with %s", window->current_channel->name);
        	else
        		sprintf( gb_tmp_buff, "%s on %s", 
        			window->current_channel->connection->data.irc->nick, 
        			window->current_channel->name);
        		
       		gtk_label_set( GTK_LABEL(window->status), gb_tmp_buff);
       		
		/* I don't know why the text->vadj value is messed here 
		   purging the event queue is the only substitute i'v found to make 
		   the text widget scroll down 
		*/
               	while ( gtk_events_pending())
               		gtk_main_iteration();
               	gtk_text_scroll_down_to_end( GTK_TEXT( window->text));
       }
}
	                

/*****************/
/*** GUICList  ***/



GtkWidget*
gui_clist_init( GUICList *list, char **titles) {
        GtkWidget *mainbox;
        GtkWidget *hbox;
        GtkWidget *hsep;
        GtkWidget *button;
        int	i = 0;
	
	while (titles[i]) i++;
	list->items = NULL;
	list->filesel = NULL;
	
	mainbox = gtk_vbox_new( FALSE, 5);
	gtk_container_border_width( GTK_CONTAINER( mainbox), 10);
	
	hbox = gtk_hbox_new( FALSE, 5);
	gtk_box_pack_start( GTK_BOX( mainbox), hbox, FALSE, FALSE, 5);
	gtk_widget_show( hbox);
		
        list->clist = gtk_clist_new_with_titles( i, titles);
        gtk_clist_set_policy( GTK_CLIST(list->clist), GTK_POLICY_AUTOMATIC, 
        			GTK_POLICY_AUTOMATIC);
        gtk_box_pack_start( GTK_BOX( mainbox), list->clist, TRUE, TRUE, 15);
        gtk_clist_set_column_width( GTK_CLIST( list->clist), 0, 100);
        gtk_signal_connect ( GTK_OBJECT (list->clist), "select_row",
			    (GtkSignalFunc) gui_clist_select_cb, NULL);
        
        gtk_widget_show( list->clist);
        
        hsep = gtk_hseparator_new();
        gtk_box_pack_start( GTK_BOX(mainbox), hsep, FALSE, FALSE, 0);
        gtk_widget_show(hsep);
        
        hbox = gtk_hbox_new( FALSE, 5);
        gtk_box_pack_start( GTK_BOX(mainbox), hbox, FALSE, FALSE, 5);
        gtk_widget_show( hbox);
        
        button = gtk_button_new_with_label( "Save list ...");
        gtk_box_pack_start( GTK_BOX(hbox), button, FALSE, FALSE, 5);
        gtk_signal_connect( GTK_OBJECT( button), "clicked", 
        	(GtkSignalFunc) gui_clist_make_filesel_cb, list);
        gtk_widget_show( button);
        
        return mainbox;
}
        
GUICList*
gui_clist_new( char *label, char **titles) {
        GUICList  *list;
        GtkWidget *frame;
        
        g_return_val_if_fail( label != NULL, FALSE);
        g_return_val_if_fail( titles != NULL, FALSE);
        
        list = g_malloc( sizeof( GUICList));
        
        list->window = gtk_window_new( GTK_WINDOW_TOPLEVEL);
        register_new_window( list->window, NULL);
        gtk_window_set_title( GTK_WINDOW(list->window), label);
        gtk_window_set_policy( GTK_WINDOW(list->window), 1, 1, 0);
        gtk_widget_set_usize( list->window, 400, 200);
        gtk_signal_connect( GTK_OBJECT( list->window), "destroy",
        	(GtkSignalFunc) gui_clist_destroy_cb, list);
        	
	frame = gui_clist_init( list, titles);
	gtk_container_add( GTK_CONTAINER(list->window), frame);
	gtk_widget_show( frame);
	
        gtk_widget_show(list->window);
        
        return list;
}

void
gui_clist_free( GUICList *list) {
        g_return_if_fail( list != NULL);
        
	gtk_widget_destroy( list->window);
}
        
void
gui_clist_destroy_cb( GtkWidget *w, GUICList *list) {
        
        if ( list->timeout_tag)
        	gtk_timeout_remove( list->timeout_tag);
        if ( list->filesel)
        	gtk_widget_destroy( list->filesel);
        
        g_free( list);
        
}


void
gui_clist_start( GUICList *list) {
        g_return_if_fail( list != NULL);
        
	gtk_clist_clear( GTK_CLIST(list->clist));
	if ( list->timeout_tag)
		gtk_timeout_remove( list->timeout_tag);
	list->timeout_tag = gtk_timeout_add( GUI_CLIST_UPDATE_TIMEOUT, 
		(GtkFunction) gui_clist_update , list);
}

void
gui_clist_end( GUICList *list) {
        g_return_if_fail( list != NULL);
        
        if ( list->timeout_tag) {
		gtk_timeout_remove( list->timeout_tag);
		list->timeout_tag = 0;
	}
	gui_clist_update( list);
}

void
gui_clist_append( GUICList *list, GList *items) {
        GList *item;
        char **saved_line;
        int    i;
        
        g_return_if_fail( list != NULL);
        g_return_if_fail( items != NULL);
        
       	saved_line = g_malloc( GTK_CLIST(list->clist)->columns*sizeof(char*));
       	
       	item = items; i = 0;
       	while (item && i < GTK_CLIST(list->clist)->columns) {
       	        if ( item->data)
       	        	saved_line[i] = g_strdup( (char*) item->data);
       	        item = item->next;
       	        i++;
       	}
       	list->items = g_list_append( list->items, saved_line);
}

gint			
gui_clist_update ( GUICList *list) {
        char **line;
        GList *item;
        int    i;
        
        g_return_val_if_fail( list != NULL, FALSE);
        
        if ( list->items != NULL) {
        	gtk_clist_freeze( GTK_CLIST(list->clist));
        
        	item = list->items;
        	while (item) {
        		line = (char**) item->data;
        		item = item->next;
			gtk_clist_append( GTK_CLIST(list->clist), line);
			for ( i = 0; i < GTK_CLIST(list->clist)->columns; i++) 
				g_free( line[i]);
		}
	
		g_list_free( list->items);
		list->items = NULL;
	
		gtk_clist_thaw( GTK_CLIST(list->clist));
	}
	return 1;
}

void		
gui_clist_select_cb ( GtkWidget *widget,
 		   gint row, 
		   gint column, 
		   GdkEventButton *beven){
	      		 
}

void
gui_clist_make_filesel_cb( GtkWidget *w, GUICList *list) {
        GtkWidget *filesel;
        
        filesel = gtk_file_selection_new( "Save list");
        gtk_signal_connect( GTK_OBJECT( GTK_FILE_SELECTION(filesel)->ok_button),"clicked",
        	 (GtkSignalFunc) gui_clist_save_cb, list);
        gtk_signal_connect_object( GTK_OBJECT( GTK_FILE_SELECTION(filesel)->cancel_button),"clicked",
        	 (GtkSignalFunc) gtk_widget_destroy, GTK_OBJECT(filesel));
        
        list->filesel = filesel;
        gtk_widget_show( filesel);
}
        
void
gui_clist_save_cb( GtkWidget *w, GUICList *list) {
        FILE	*file;
        GList	*item;
        GtkCListRow *row;
        char	*filename;
        int	i;
        
        filename = gtk_file_selection_get_filename( GTK_FILE_SELECTION(list->filesel));
        file = fopen( filename, "w");
        if ( file) {
        	item = GTK_CLIST(list->clist)->row_list;
        	while (item) {
        	        row = (GtkCListRow*) item->data;
        	        item = item->next;
        		for ( i = 0; i < GTK_CLIST(list->clist)->columns; i++) 
        			fprintf( file, "%s \t", GTK_CELL_TEXT (row->cell[i])->text);
        		fprintf( file, "\n");
        	       
        	}
        	fclose( file);
        } else g_warning( "Cannot open %s for writing", filename);
        gtk_widget_destroy( list->filesel);
}


/*******************/
/*** Server list ***/

/* TODO : use GtkCList instead of GtkList 
 */
GtkWidget*
gui_server_list_init( GUIServerList *slist) {
        GtkWidget *mainbox;
        GtkWidget *omenu;
        GtkWidget *menu;
        GtkWidget *menu_item;
        GtkWidget *box1;
        GtkWidget *box2;
        GtkWidget *label;
        GtkWidget *sw;
        
        GList	  *irc_networks;
        IRCNetwork *irc_network;
        
        mainbox = gtk_vbox_new( FALSE, 8);
        
        box1 = gtk_hbox_new( FALSE, 10);
        gtk_box_pack_start( GTK_BOX( mainbox), box1, FALSE, FALSE, 0);
        gtk_widget_show( box1);
        
        label = gtk_label_new( "Irc network :");
        gtk_box_pack_start( GTK_BOX( box1), label, FALSE, FALSE, 0);
        gtk_widget_show( label);
        
        menu = gtk_menu_new();
        
        omenu = gtk_option_menu_new();
        gtk_box_pack_start( GTK_BOX(box1), omenu, FALSE, TRUE, 5);
        gtk_option_menu_set_menu( GTK_OPTION_MENU( omenu), menu);
        gtk_widget_set_usize( omenu, 150, -1);
        gtk_widget_show( omenu);
        
        slist->net_desc = gtk_text_new( NULL, NULL);
        gtk_box_pack_start( GTK_BOX( mainbox), slist->net_desc, FALSE, FALSE, 5);
        //gtk_widget_show( slist->net_desc);
        
        box1 = gtk_hbox_new( FALSE, 5);
        gtk_container_add( GTK_CONTAINER( mainbox), box1);
        gtk_widget_show( box1);
        
        sw = gtk_scrolled_window_new( NULL, NULL);
        gtk_container_add( GTK_CONTAINER(box1), sw);
        gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW(sw), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
        gtk_widget_show( sw);
        
        slist->list = gtk_list_new();
        gtk_container_add( GTK_CONTAINER(sw), slist->list);
        gtk_signal_connect( GTK_OBJECT( slist->list), "select_child", 
        	(GtkSignalFunc) gui_server_list_select_server_cb, slist);
        gtk_widget_show( slist->list);
        
        irc_networks = gb_irc_networks;
        while ( irc_networks) {
                irc_network = (IRCNetwork*) irc_networks->data;
                irc_networks = irc_networks->next;
                
                menu_item  = gtk_menu_item_new_with_label( irc_network->label);
                gtk_menu_append( GTK_MENU(menu), menu_item);
                gtk_signal_connect( GTK_OBJECT( menu_item), "activate",
                	(GtkSignalFunc) gui_server_list_select_network_cb, slist);
                gtk_object_set_user_data( GTK_OBJECT( menu_item), irc_network);
                gtk_widget_show( menu_item);
                
        }
        
        slist->current_network = (IRCNetwork*) gb_irc_networks->data;
        gtk_option_menu_set_history( GTK_OPTION_MENU( omenu), 0);
        
        box2 = gtk_vbox_new( FALSE, 10);
        gtk_box_pack_end( GTK_BOX(box1), box2, FALSE, FALSE, 5);
        gtk_widget_set_usize( box2, 250, -1);
        gtk_widget_show( box2);
        
        label = gtk_label_new( "Server address :");
        gtk_box_pack_start( GTK_BOX( box2), label, FALSE, FALSE, 0);
        gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
        gtk_widget_show( label);
        
        slist->server = gtk_entry_new();
        gtk_box_pack_start( GTK_BOX( box2), slist->server, FALSE, FALSE, 0);
        gtk_entry_set_editable( GTK_ENTRY(slist->server), 0);
        gtk_widget_show( slist->server);
        
        box1 = gtk_hbox_new( FALSE, 5);
        gtk_box_pack_start( GTK_BOX( box2), box1, FALSE, FALSE, 5);
        gtk_widget_show( box1);
         
        label = gtk_label_new( "Server Ports :");
        gtk_box_pack_start( GTK_BOX( box1), label, FALSE, FALSE, 0);
        gtk_misc_set_alignment (GTK_MISC (label), 0.0, 0.5);
        gtk_widget_show( label);
        
        slist->ports = gtk_option_menu_new();
        gtk_box_pack_end( GTK_BOX( box1), slist->ports, FALSE, FALSE, 0);
        gtk_widget_set_usize( slist->ports, 100, -1);
        gtk_widget_show( slist->ports);
        
        return mainbox;
}

GUIServerList*
gui_server_list_new( ) {
        GUIServerList *slist;
        GtkWidget *window;
        GtkWidget *mainbox;
        
        window = gtk_window_new( GTK_WINDOW_TOPLEVEL);
        gtk_container_border_width( GTK_CONTAINER(window), 10);
        gtk_window_set_title( GTK_WINDOW( window), "Server list");
        gtk_widget_set_usize( window, 500, 400);
        gtk_window_set_policy( GTK_WINDOW( window), 1, 1, 0);
        register_new_window( window, NULL);
        
        slist = g_malloc( sizeof( GUIServerList));
        mainbox = gui_server_list_init( slist);
        gtk_container_add( GTK_CONTAINER( window), mainbox);
        gtk_widget_show( mainbox);
        
        gtk_widget_show( window);
        
        return slist;
}

void
gui_server_list_select_network_cb( GtkWidget *menuitem, GUIServerList *slist) {
        GList 		*item;
        IRCNetwork	*irc_network;
        IRCServer	*server;
        GtkWidget	*list_item;
        
        gtk_list_remove_items( GTK_LIST( slist->list), gtk_container_children( GTK_CONTAINER( slist->list)));
	irc_network = gtk_object_get_user_data( GTK_OBJECT( menuitem));
	slist->current_network = irc_network;
	
        item = irc_network->servers;
        while (item) {
                server	= (IRCServer*) item->data;
                item = item->next;
                list_item = gtk_list_item_new_with_label( server->label);
                gtk_container_add( GTK_CONTAINER(slist->list), list_item);
                gtk_widget_show( list_item);
        }
}

void
gui_server_list_select_server_cb( GtkWidget *list, GtkWidget *list_item, GUIServerList *slist) {
        GtkWidget 	*menu;
        GtkWidget 	*menu_item;
        GList		*item;
        IRCServer 	*server;
        int		i;
        
        i = gtk_list_child_position( GTK_LIST(list), list_item);
        item = g_list_nth( slist->current_network->servers, i);
        if ( item) {
                server = (IRCServer*) item->data;
                gtk_entry_set_text( GTK_ENTRY( slist->server), server->address);
                
                menu = gtk_option_menu_get_menu( GTK_OPTION_MENU( slist->ports));
                if ( menu)
                	gtk_widget_destroy( menu);
                	
		menu = gtk_menu_new();
		item = server->ports;
		while ( item) {
		      menu_item = gtk_menu_item_new_with_label( (char*) item->data);
		      gtk_menu_append( GTK_MENU( menu), menu_item);
		      gtk_widget_show( menu_item);
		      
		      item = item->next;
		}
                        
                gtk_option_menu_set_menu( GTK_OPTION_MENU( slist->ports), menu);
        }
}

        
/***********************/
/*** Misc functions  ***/

  
